/*
 * This file is part of the AdvanceMAME project
 *
 * Copyright (C) 1999-2000 Andrea Mazzoleni
 *
 * The contents of this file are free for non-commercial use.
 * Any commercial use is forbidden.
 */

#include "game.h"
#include "text.h"
#include "menu.h"

#include <unistd.h>

#include <strstream>
#include <iomanip>
#include <algorithm>

// -------------------------------------------------------------------------
// Draw

int draw_string(int x, int y, int dx, const string& s, int color) {
        if (dx>0) {
                if (s.length() >= dx) {
	        	text_put(x,y,string(s,0,dx),color);
                } else {
	                text_put(x,y,s,color);
                        text_fill(x+s.length(),y,dx-s.length(),' ',color);
                }
                return dx;
        } else {
	        return 0;
        }
}

int draw_string(int x, int y, const string& s, int color) {
	return draw_string(x,y,s.length(),s,color);
}

int draw_char(int x, int y, int dx, char c, int color) {
	if (dx>0) {
		text_fill(x,y,dx,c,color);
        	return dx;
        } else {
        	return 0;
        }
}

void draw_string_special(int x, int y, int len, const string& s, int c0, int c1, int c2) {
	bool in = false;
	for(int i=0;i<len;++i) {
		if (i==0) {
			draw_char(x,y,1,s[i],c0);
		} else {
			if (s[i]==')')
				in = false;
			draw_char(x,y,1,s[i], in ? c1 : c2);
			if (s[i]=='(')
				in = true;
		}
		++x;
	}
}

// ------------------------------------------------------------------------
// Choice menu

#define CHOICE_COLOR_TITLE  0x7F
#define CHOICE_COLOR_ITEM  0x70
#define CHOICE_COLOR_ITEM_TAG 0x70
#define CHOICE_COLOR_ITEM_SELECT 0x20
#define CHOICE_COLOR_ITEM_SELECT_TAG 0x20

struct choice {
	choice(const string& Adesc);
	choice(const string& Adesc, bool Abistate);
	choice(const string& Adesc, tristate_t Atristate);

	string desc;
	int state;
	bool bistate;
	tristate_t tristate;
};

choice::choice(const string& Adesc) :
	desc(Adesc), state(1) {
}

choice::choice(const string& Adesc, bool Abistate) :
	desc(Adesc),state(2),bistate(Abistate) {
}

choice::choice(const string& Adesc, tristate_t Atristate) :
	desc(Adesc),state(3),tristate(Atristate) {
}

typedef list<choice> choice_container;

void draw_choice_window(const string& title, int x, int y, int dx, choice_container& ch, choice_container::iterator& pos) {
	draw_string(x,y,dx,title, CHOICE_COLOR_TITLE);
	dx -= 6;
	++y;
	for(choice_container::iterator i = ch.begin();i!=ch.end();++i) {
		int color;
		if (i==pos) {
			color = CHOICE_COLOR_ITEM_SELECT;
		} else {
			color = CHOICE_COLOR_ITEM;
		}
		const char* sstate;
		if (i->state==1) {
			if (i==pos) {
				sstate = " * ";
			} else {
				sstate = "   ";
			}
		} else if (i->state==2) {
			if (i->bistate) {
				sstate = " + ";
			} else {
				sstate = " - ";
			}
		} else {
			if (i->tristate == exclude) {
				sstate = " - ";
			} else if (i->tristate == exclude_not) {
				sstate = " * ";
			} else {
				sstate = " + ";
			}
		}
		draw_string(x,y,sstate, color);
		draw_string(x+3,y,dx,i->desc, color);
		draw_string(x+3+dx,y,"   ", color);
		++y;
	}
}

int run_choice(const string& title, int x, int y, int dx, choice_container& ch, choice_container::iterator& pos) {
	int key = TEXT_KEY_ESC;
	int done = 0;

	while (!done) {
		draw_choice_window(title,x,y,dx,ch,pos);
		key = text_getkey();

		switch (key) {
			case '-' :
				for(choice_container::iterator i=ch.begin();i!=ch.end();++i) {
					switch (i->state) {
						case 2:	i->bistate = false; break;
						case 3:	i->tristate = exclude; break;
					}
				}
			break;
			case '+' :
				for(choice_container::iterator i=ch.begin();i!=ch.end();++i) {
					switch (i->state) {
						case 2:	i->bistate = true; break;
						case 3:	i->tristate = include; break;
					}
				}
			break;
			case TEXT_KEY_HOME :
				pos = ch.begin();;
				break;
			case TEXT_KEY_END :
				if (pos != ch.end()) {
					do {
						++pos;
					} while (pos != ch.end());
					--pos;
				}
				break;
			case TEXT_KEY_UP :
				if (pos != ch.begin())
					--pos;
				break;
			case TEXT_KEY_DOWN :
				if (pos != ch.end()) {
					if (++pos == ch.end())
						--pos;
				}
				break;
			case TEXT_KEY_SPACE :
				switch (pos->state) {
					case 2:	pos->bistate = !pos->bistate; break;
					case 3:
						switch (pos->tristate) {
							case include : pos->tristate = exclude; break;
							case exclude : pos->tristate = exclude_not; break;
							case exclude_not : pos->tristate = include; break;
						}
				}
				break;
			case TEXT_KEY_ENTER :
			case TEXT_KEY_ESC :
				done = 1;
				break;
		}
	}

	return key;
}

// ------------------------------------------------------------------------
// Menu

#define COLOR_MENU_NORMAL 0x0B
#define COLOR_MENU_SELECT 0x70
#define COLOR_MENU_HIDDEN 0x03
#define COLOR_MENU_HIDDEN_SELECT 0x70
#define COLOR_MENU_IN 0x03
#define COLOR_MENU_IN_SELECT 0x70
#define COLOR_MENU_FIRST 0x0F
#define COLOR_MENU_SELECT_FIRST	0x70

#define COLOR_MENU_BAR 0x70
#define COLOR_MENU_BAR_TAG 0x7C
#define COLOR_MENU_BORDER 0x07

void draw_menu_game(const game_set& gar, const game* g, int x, int y, int dx, bool selected, bool ident_if_clone, merge_t merge) {
	string s;
	if (g) {
		s = g->description_get();
	} else
	 	s = "";

	int len = s.length();
        if (len > dx)
        	len = dx;

	unsigned space_0;
	if (ident_if_clone && g && g->cloneof_get().length())
		space_0 = 2;
	else
		space_0 = 0;
	if (len + space_0 > dx)
		len = dx - space_0;
        unsigned space_1 = dx - len - space_0;

        int color;
	int color_first;
	int color_in;
        if (g && g->working_get() && gar.is_tree_rom_of_present(g->name_get(),merge)) {
	        color = selected ? COLOR_MENU_SELECT : COLOR_MENU_NORMAL;
		color_first = selected ? COLOR_MENU_SELECT_FIRST: COLOR_MENU_FIRST;
		color_in = selected ? COLOR_MENU_IN_SELECT : COLOR_MENU_IN;
        } else {
	        color = selected ? COLOR_MENU_HIDDEN_SELECT : COLOR_MENU_HIDDEN;
		color_first = color;
		color_in = color;
        }

        if (space_0)
        	draw_char(x,y,space_0,' ',color);
        if (len)
        	draw_string_special(x+space_0,y,len,s,color_first,color_in,color);
        if (space_1)
        	draw_char(x+len+space_0,y,space_1,' ',color);
}

void draw_menu_bar(const game_set& gar, const game* g, int g1, int g2, int x, int y, int dx) {
	draw_char(x,y,dx,' ',COLOR_MENU_BAR);

	int separator =	dx > 40 ? 1 : 0;
	int in_separator =  dx > 40 ? 2 : 1;
	// int num = dx - separator;
	int xr = x + dx - separator;
	int xl = x + separator;

	if (g) {
		ostrstream os;
		os << setw(4) << setfill(' ') << g1 << "/" << setw(4) << setfill(' ') << g2 << ends;
		int len = strlen(os.str());
		if (len < xr-xl) {
			draw_string(xr-len,y,os.str(),COLOR_MENU_BAR);
			xr -= len;
			xr -= in_separator;
		}
		os.freeze(0);
	}

	if (g) {
		ostrstream os;
		os << setw(3) << setfill('0') << (g->time_get()/3600) << ":" << setw(2) << setfill('0') << ((g->time_get() / 60) % 60) << ends;
		int len = strlen(os.str());
		if (len < xr-xl) {
			draw_string(xr-len,y,os.str(),COLOR_MENU_BAR);
			xr -= len;
			xr -= in_separator;
		}
		os.freeze(0);
	}

	if (g) {
		ostrstream os;
		if (g->size_get() >= 10*1000*1000)
			os << setw(4) << g->size_get()/1000/1000 << "M"  << ends;
		else
			os << setw(4) << g->size_get()/1000 << "k"  << ends;
		int len = strlen(os.str());
		if (len < xr-xl) {
			draw_string(xr-len,y,os.str(),COLOR_MENU_BAR);
			xr -= len;
			xr -= in_separator;
		}
		os.freeze(0);
	}

	if (g && g->sizex_get() && g->sizey_get()) {
		ostrstream os;
		os << setw(3) << g->sizex_get() << "x" << g->sizey_get() << ends;
		int len = strlen(os.str());
		if (len < xr-xl) {
			draw_string(xr-len,y,os.str(),COLOR_MENU_BAR);
			xr -= len;
			xr -= in_separator;
		}
		os.freeze(0);
	}

	if (g && g->group_get() != GROUP_UNDEFINED) {
		ostrstream os;
		os << g->group_get() << ends;
		int len = strlen(os.str());
		if (len < xr-xl) {
			draw_string(xr-len,y,os.str(),COLOR_MENU_BAR);
			xr -= len;
			xr -= in_separator;
		}
		os.freeze(0);
	}

	if (g && g->group_get() != TYPE_UNDEFINED) {
		ostrstream os;
		os << g->type_get()  << ends;
		int len = strlen(os.str());
		if (len < xr-xl) {
			draw_string(xr-len,y,os.str(),COLOR_MENU_BAR);
			xr -= len;
			xr -= in_separator;
		}
		os.freeze(0);
	}

	if (g) {
		ostrstream os;
		os << g->name_get()  << ends;
		int len = separator + strlen(os.str());
		if (len < xr-xl) {
			draw_string(xl,y,os.str(),COLOR_MENU_BAR);
			xl += len;
			xl += in_separator;
			xl -= separator;
		}
		os.freeze(0);
        }
}

void draw_menu_info(const game_set& gar, const game* g, int x, int y, int dx, merge_t merge) {
	draw_char(x,y,dx,' ',COLOR_MENU_BAR);

	int separator =	dx > 40 ? 1 : 0;
	int d = separator;

        if (g) {
		d += draw_string(x+d,y,g->manufacturer_get(),COLOR_MENU_BAR);
	        d += draw_string(x+d,y,", ",COLOR_MENU_BAR);
	        d += draw_string(x+d,y,g->year_get(),COLOR_MENU_BAR);

		if (g->clone_get() > 0) {
			char buf[32];
			sprintf(buf,", %d clone",g->clone_get());
			d += draw_string(x+d,y,buf,COLOR_MENU_BAR);
		}

	        if (g->cloneof_get().length()) {
                        game_set::const_iterator i = gar.find( g->cloneof_get() );
			if (i!=gar.end()) {
	                        d += draw_string(x+d,y,", cloneof ",COLOR_MENU_BAR);
		 	        d += draw_string(x+d,y,i->description_get(),COLOR_MENU_BAR);
			}
		}

		/*
	        if (g->cloneof_get().length()) {
                        game_set::const_iterator i = gar.find( g->cloneof_get() );
			if (i!=gar.end()) {
	                        d += draw_string(x+d,y,", cloneof ",COLOR_MENU_BAR);
		 	        d += draw_string(x+d,y,i->description_get(),COLOR_MENU_BAR);
                        }
	        } else if (g->romof_get().length()) {
			game_set::const_iterator i = gar.find( g->romof_get() );
			if (i!=gar.end()) {
	                        d += draw_string(x+d,y,", romof ",COLOR_MENU_BAR);
		 	        d += draw_string(x+d,y,i->description_get(),COLOR_MENU_BAR);
                        }
		}
		*/

                // status
	        if (!gar.is_tree_rom_of_present(g->name_get(),merge))
			draw_string(x+dx-separator-7,y,"MISSING",COLOR_MENU_BAR_TAG);
		else if (!g->working_get())
		       draw_string(x+dx-separator-5,y,"ALPHA",COLOR_MENU_BAR_TAG);
		else if (!g->color_get())
		        draw_string(x+dx-separator-7,y,"NOCOLOR",COLOR_MENU_BAR_TAG);
		else if (!g->sound_get())
		        draw_string(x+dx-separator-7,y,"NOSOUND",COLOR_MENU_BAR_TAG);
        }
}

void draw_menu_window(const game_set& gar, const game_container& gc, int x, int y, int dx, int dy, int coln, int start, int pos, bool ident_if_clone, merge_t merge) {
	int entry_dx = (dx-1) / coln;
        for(int r=0;r<dy;++r) {
		int c;
	        for(c=0;c<coln;++c) {
                	const game* g = start<gc.size() ? gc[start] : 0;
		        draw_char(x+c*entry_dx,y+r,1,' ',COLOR_MENU_BORDER);
	                draw_menu_game(gar,g,x+1+c*entry_dx,y+r,entry_dx-1, start == pos, ident_if_clone, merge );
			++start;
                }
		draw_char(x+c*entry_dx,y+r,dx-c*entry_dx,' ',COLOR_MENU_BORDER);
        }
}

// ------------------------------------------------------------------------

void menu_pos(int pos, menu_state& rs, int pos_max, int base_max, int coln) {
	while (pos >= base_max)
		pos -= coln;
	while (pos < 0)
		pos += coln;
	if (pos >= rs.pos_base && pos < rs.pos_base + pos_max) {
		rs.pos_rel = pos - rs.pos_base;
	} else if (pos < rs.pos_base) {
		rs.pos_rel = pos % coln;
		rs.pos_base = pos - rs.pos_rel;
	} else {
		rs.pos_rel = pos % coln + pos_max - coln;
		rs.pos_base = pos - rs.pos_rel;
		if (rs.pos_base < 0) {
			rs.pos_base = 0;
			rs.pos_rel = pos;
		}
	}
}

void menu_center(int pos, menu_state& rs, int pos_max, int base_max, int coln) {
	while (pos >= base_max)
		pos -= coln;
	while (pos < 0)
		pos += coln;
	if (pos >= rs.pos_base && pos < rs.pos_base + pos_max) {
		rs.pos_rel = pos - rs.pos_base;
	} else {
		rs.pos_rel = pos % coln + (pos_max / coln) / 2 * coln;
		rs.pos_base = pos - rs.pos_rel;
		if (rs.pos_base < 0) {
			rs.pos_base = 0;
			rs.pos_rel = pos;
		}
	}
}

bool menu_fast_compare(const string& game, const string& fast) {
	unsigned igame = 0;
	unsigned ifast = 0;
	while (igame < game.length() && ifast < fast.length()) {
		while (igame < game.length() && !isalnum(game[igame]))
		++igame;
		if (igame < game.length()) {
			if (toupper(game[igame]) != toupper(fast[ifast]))
				return false;
			++igame;
			++ifast;
		}
	}
	return ifast == fast.length();
}

//--------------------------------------------------------------------------

bool backdrop_find(string& path, const string& snap, const string& name) {
	unsigned i = 0;
	while (i<snap.length()) {
		int end = snap.find(';',i);
                if (end==string::npos) {
       			path = string( snap, i );
                        i = snap.size();
                } else {
       		        path = string( snap, i, end-i );
                        i = end + 1;
                }
		if (path.length()) {
			if (path[path.length()-1]!='\\' && path[path.length()-1]!='/') {
				path += "/";
			}
       		}

		path += name + ".png";

		// test read access
		if (access(path.c_str(),F_OK | R_OK)==0) {
			return true;
		}
	}
	return false;
}

bool backdrop_get(const game_set& gar, string& path, menu_state& rs) {

	if (rs.pgame) {
		game_set::const_iterator i = gar.find( rs.pgame->name_get() );
		while (i != gar.end()) {
		 	if (backdrop_find(path,rs.snap,i->name_get()))
				return true;
			i =  gar.find( i->cloneof_get() );
		}
	}

	return false;
}

#define BACKDROP_FOREGROUND 1.0
#define BACKDROP_BACKGROUND 0.6

int run_menu(const game_set& gar, menu_state& rs, bool flipxy) {
	game_container gc;
        int game_good = 0;

        for(game_set::const_iterator i=gar.begin();i!=gar.end();++i) {
		bool accept = true;
		bool accept_group = rs.include_group.size() == 0;
		bool accept_type = rs.include_type.size() == 0;

		// group
			for(group_container::iterator j=rs.include_group.begin();j!=rs.include_group.end();++j) {
				if (*j == i->group_get()) {
					accept_group = true;
					break;
				}
			}

		// type
			for(type_container::iterator j=rs.include_type.begin();j!=rs.include_type.end();++j) {
				if (*j == i->type_get()) {
					accept_type = true;
					break;
				}
			}

		if  (accept_group && accept_type) {
			bool present = false;

			// exclude
	        	if (accept && rs.exclude_bad == exclude && !i->working_get())
				accept = false;
		       	if (accept && rs.exclude_bad == exclude_not && i->working_get())
				accept = false;
		       	if (accept && rs.exclude_clone == exclude && i->cloneof_get().length()!=0)
				accept = false;
        		if (accept && rs.exclude_clone == exclude_not && i->cloneof_get().length()==0)
				accept = false;
	        	if (accept && rs.exclude_neogeo == exclude && gar.is_game_neogeo(i->name_get()))
				accept = false;
			if (accept && rs.exclude_neogeo == exclude_not && !gar.is_game_neogeo(i->name_get()))
				accept = false;
			if (accept)
				present = gar.is_tree_rom_of_present(i->name_get(),rs.merge);
			if (accept && rs.exclude_missing == exclude && !present)
				accept = false;
			if (accept && rs.exclude_missing == exclude_not && present)
				accept = false;
			if (accept && rs.exclude_vector == exclude && i->vector_get())
				accept = false;
			if (accept && rs.exclude_vector == exclude_not && !i->vector_get())
				accept = false;
			if (accept && i->resource_get())
				accept = false;
			if (accept) {
				// count game runnable
				if (i->working_get() && present)
		                	++game_good;
				// insert
		        	gc.insert( gc.end(), &*i);
			}
		}
        }

	// sort
	switch (rs.sort_mode) {
		case sort_by_root_name :
			sort(gc.begin(),gc.end(), combine_less<const game*,pgame_by_rootdesc_less,pgame_by_clone_less>(gar));
			break;
		case sort_by_name :
			sort(gc.begin(),gc.end(), pgame_by_desc_less(gar));
			break;
		case sort_by_manufacturer :
			sort(gc.begin(),gc.end(), combine_less<const game*,pgame_by_manufacturer_less,pgame_by_rootdesc_less>(gar));
			break;
		case sort_by_year :
			sort(gc.begin(),gc.end(), combine_less<const game*,pgame_by_year_less,pgame_by_rootdesc_less>(gar));
			break;
		case sort_by_time :
			sort(gc.begin(),gc.end(), combine_less<const game*,pgame_by_time_less,pgame_by_rootdesc_less>(gar));
			break;
		case sort_by_group :
			sort(gc.begin(),gc.end(), combine_less<const game*,pgame_by_group_less,pgame_by_rootdesc_less>(gar));
			break;
		case sort_by_type :
			sort(gc.begin(),gc.end(), combine_less<const game*,pgame_by_type_less,pgame_by_rootdesc_less>(gar));
			break;
		case sort_by_size :
			sort(gc.begin(),gc.end(), combine_less<const game*,pgame_by_size_less,pgame_by_rootdesc_less>(gar));
			break;
	}

	// fill background
	for(int y=0;y<text_dy_get();++y) {
		text_fill(0,y,text_dx_get(),' ',COLOR_MENU_NORMAL);
	}

	int coln;
        int rown;
	int win_dx;

	int backdrop_x;
	int backdrop_y;
	int backdrop_dx;
	int backdrop_dy;

	if (rs.backdrop_merge) {
		coln = 1;
		rown = text_dy_get() - 2;
		win_dx = text_dx_get();
		backdrop_x = 0;
		backdrop_y = 1;
		backdrop_dx = text_dx_get();
		backdrop_dy = text_dy_get()-2;
	} else {
		if (!flipxy) {
        		coln = 1;
	        	rown = text_dy_get() - 2;
			win_dx = text_dx_get() / 3;
			backdrop_x = win_dx;
			backdrop_y = 1;
			backdrop_dx = text_dx_get()-win_dx;
			backdrop_dy = text_dy_get()-2;
		} else {
			if (text_dx_get() > 60)
				coln = 3;
			else
				coln = 2;
		        rown = (text_dy_get() - 2) / 3;
			win_dx = text_dx_get();
			backdrop_x = 0;
			backdrop_y = 1+rown;
			backdrop_dx = text_dx_get();
			backdrop_dy = text_dy_get()-2-rown;
		}
	}
	text_backdrop_set(backdrop_x,backdrop_y,backdrop_dx,backdrop_dy,COLOR_MENU_NORMAL >> 4, BACKDROP_BACKGROUND);

        int pos_max = coln*rown;
        int base_max = (gc.size() + coln - 1) / coln * coln;

        if (rs.pos_base + rs.pos_rel >= base_max) {
        	rs.pos_base = base_max - pos_max;
                if (rs.pos_base < 0) {
                	rs.pos_base = 0;
                }
                if (rs.pos_base + rs.pos_rel >= base_max) {
                	rs.pos_base = 0;
                        rs.pos_rel = 0;
                }
        }

	if (rs.pgame) {
		unsigned i = 0;
		while (i<gc.size() && rs.pgame != gc[i])
			++i;
		if (i<gc.size())
			menu_pos(i,rs,pos_max,base_max,coln);
	} else if (rs.last_game.length()) {
		unsigned i = 0;
		while (i<gc.size() && rs.last_game != gc[i]->name_get())
			++i;
		if (i<gc.size())
			menu_pos(i,rs,pos_max,base_max,coln);
	}

      	if (rs.pos_rel + rs.pos_base < gc.size()) {
               	rs.pgame = gc[rs.pos_rel+rs.pos_base];
		rs.pclone = 0;
        } else {
               	rs.pgame = 0;
		rs.pclone = 0;
        }

        bool done = false;

        int key = 0;
	bool idle = false;

        while (!done) {
                const game* g = rs.pos_base + rs.pos_rel < gc.size() ? gc[rs.pos_base + rs.pos_rel] : 0;
		if (!idle) {
			draw_menu_window(gar,gc,0,1,win_dx,rown,coln,rs.pos_base, rs.pos_base + rs.pos_rel, rs.sort_mode == sort_by_root_name, rs.merge );
		        draw_menu_bar(gar,g,game_good,gc.size(),0,0,text_dx_get());
			draw_menu_info(gar,g,0,text_dy_get()-1,text_dx_get(),rs.merge);
		}

		string backdrop;
		if (backdrop_get(gar,backdrop,rs)) {
			text_backdrop( backdrop.c_str() );
		} else {
			text_backdrop_clear();
		}

                key = text_getkey();

		string oldfast = rs.fast;
		rs.fast.erase();

		if (idle && key != TEXT_KEY_IDLE_1) {
			text_backdrop_set(backdrop_x,backdrop_y,backdrop_dx,backdrop_dy,COLOR_MENU_NORMAL >> 4, BACKDROP_BACKGROUND);
			idle = false;
		}

                switch (key) {
			case TEXT_KEY_HOME :
				menu_pos(0,rs,pos_max,base_max,coln);
                        	break;
                        case TEXT_KEY_END :
				menu_pos(gc.size()-1,rs,pos_max,base_max,coln);
                        	break;
                        case TEXT_KEY_PGUP :
                        	if (rs.pos_base >= pos_max) {
                                	rs.pos_base -= pos_max;
                                } else if (rs.pos_base>0) {
                                	rs.pos_base = 0;
                                } else {
                                	rs.pos_rel = 0;
                                }
                        	break;
                        case TEXT_KEY_PGDN :
                        	if (rs.pos_base + 2*pos_max <= base_max) {
                                	rs.pos_base += pos_max;
                                } else if (rs.pos_base < base_max - pos_max) {
	                                rs.pos_base = base_max - pos_max;
                                } else {
                                	rs.pos_rel = gc.size() - rs.pos_base - 1;
                                }
                        	break;
                	case TEXT_KEY_UP :
				menu_pos(rs.pos_base + rs.pos_rel - coln,rs,pos_max,base_max,coln);
				break;
			case TEXT_KEY_DOWN :
				menu_pos(rs.pos_base + rs.pos_rel + coln,rs,pos_max,base_max,coln);
                                break;
			case TEXT_KEY_LEFT :
				menu_pos(rs.pos_base + rs.pos_rel -1,rs,pos_max,base_max,coln);
                                break;
			case TEXT_KEY_RIGHT :
				menu_pos(rs.pos_base + rs.pos_rel + 1,rs,pos_max,base_max,coln);
                                break;
			case TEXT_KEY_IDLE_0 :
				menu_center(rand() % gc.size(),rs,pos_max,base_max,coln);
	                       	done = true;
				break;
			case TEXT_KEY_IDLE_1 :
				text_backdrop_set(0,0,text_dx_get(),text_dy_get(),COLOR_MENU_NORMAL >> 4, BACKDROP_BACKGROUND);
				idle = true;
				menu_center(rand() % gc.size(),rs,pos_max,base_max,coln);
				break;
			case TEXT_KEY_HELP :
			case TEXT_KEY_GROUP :
			case TEXT_KEY_TYPE :
			case TEXT_KEY_EXCLUDE :
			case TEXT_KEY_SORT :
			case TEXT_KEY_SETGROUP :
			case TEXT_KEY_SETTYPE :
			case TEXT_KEY_ESC :
                        case TEXT_KEY_ENTER :
			case TEXT_KEY_MODE :
			case TEXT_KEY_RUN_CLONE :
                        	done = true;
                                break;
			default:
				if (isalnum(key)) {
					oldfast.insert(oldfast.length(),1,key);
					game_container::const_iterator i;
					for(i=gc.begin();i!=gc.end();++i) {
						if (menu_fast_compare((*i)->description_get(),oldfast))
							break;
					}
					if (i!=gc.end()) {
						int pos = i - gc.begin();
						menu_pos(pos,rs,pos_max,base_max,coln);
						rs.fast = oldfast;
					}
				}
                }

              	if (rs.pos_rel + rs.pos_base < gc.size()) {
                       	rs.pgame = gc[rs.pos_rel+rs.pos_base];
			rs.pclone = 0;
                } else {
                      	rs.pgame = 0;
			rs.pclone = 0;
                }
        }

	if (rs.pgame)
		rs.last_game = rs.pgame->name_get();
	else
		rs.last_game = "";

        return key;
}

// ------------------------------------------------------------------------
// Sort menu

#define SORT_CHOICE_X text_dx_get()/8
#define SORT_CHOICE_Y text_dy_get()/5
#define SORT_CHOICE_DX 20

void run_sort(menu_state& rs) {
	choice_container ch;
	choice_container::iterator pos[8];

	pos[0] = ch.insert( ch.end(), choice("Parent name") );
	pos[1] = ch.insert( ch.end(), choice("Name") );
	pos[2] = ch.insert( ch.end(), choice("Time played") );
	pos[3] = ch.insert( ch.end(), choice("Group") );
	pos[4] = ch.insert( ch.end(), choice("Category") );
	pos[5] = ch.insert( ch.end(), choice("Manufacturer") );
	pos[6] = ch.insert( ch.end(), choice("Year") );
	pos[7] = ch.insert( ch.end(), choice("Size") );

	choice_container::iterator i;
	switch (rs.sort_mode) {
		case sort_by_root_name: i = pos[0]; break;
		case sort_by_name: i = pos[1]; break;
		case sort_by_time: i = pos[2]; break;
		case sort_by_group: i = pos[3]; break;
		case sort_by_type: i = pos[4]; break;
		case sort_by_manufacturer: i = pos[5]; break;
		case sort_by_year: i = pos[6]; break;
		case sort_by_size: i = pos[7]; break;
	}

	int key = run_choice(" Sort mode",SORT_CHOICE_X,SORT_CHOICE_Y,SORT_CHOICE_DX,ch,i);
	if (key == TEXT_KEY_ENTER) {
		if (0) ;
		else if (i == pos[0]) rs.sort_mode = sort_by_root_name;
		else if (i == pos[1]) rs.sort_mode = sort_by_name;
		else if (i == pos[2]) rs.sort_mode = sort_by_time;
		else if (i == pos[3]) rs.sort_mode = sort_by_group;
		else if (i == pos[4]) rs.sort_mode = sort_by_type;
		else if (i == pos[5]) rs.sort_mode = sort_by_manufacturer;
		else if (i == pos[6]) rs.sort_mode = sort_by_year;
		else if (i == pos[7]) rs.sort_mode = sort_by_size;
	}
}

// ------------------------------------------------------------------------
// Exclude menu

#define EXCLUDE_CHOICE_X text_dx_get()/8
#define EXCLUDE_CHOICE_Y text_dy_get()/5
#define EXCLUDE_CHOICE_DX 30

void run_exclude(menu_state& rs) {
	choice_container ch;
	choice_container::iterator pos[16];

	pos[0] = ch.insert( ch.end(), choice("Not Working", rs.exclude_bad) );
	pos[1] = ch.insert( ch.end(), choice("Missing", rs.exclude_missing) );
	pos[2] = ch.insert( ch.end(), choice("Clone", rs.exclude_clone) );
	pos[3] = ch.insert( ch.end(), choice("Neogeo", rs.exclude_neogeo) );
	pos[4] = ch.insert( ch.end(), choice("Vector", rs.exclude_vector) );

	choice_container::iterator i = ch.begin();;
	int key = run_choice(" Select attrib set",EXCLUDE_CHOICE_X,EXCLUDE_CHOICE_Y,EXCLUDE_CHOICE_DX,ch,i);

	if (key == TEXT_KEY_ENTER) {
		rs.exclude_bad = pos[0]->tristate;
		rs.exclude_missing = pos[1]->tristate;
		rs.exclude_clone = pos[2]->tristate;
		rs.exclude_neogeo = pos[3]->tristate;
		rs.exclude_vector = pos[4]->tristate;
	}
}

// ------------------------------------------------------------------------
// Group menu

#define GROUP_CHOICE_X text_dx_get()/8
#define GROUP_CHOICE_Y text_dy_get()/5
#define GROUP_CHOICE_DX	20

void run_group(menu_state& rs, const group_container& grp) {
	choice_container ch;
	choice_container::iterator pos[32];

	int jn = 0;
	for(group_container::const_iterator j = grp.begin();j!=grp.end() && jn<32;++j) {
		bool tag = false;
		for(group_container::const_iterator k = rs.include_group.begin();k!=rs.include_group.end();++k) {
			if (*j == *k) {
				tag = true;
				break;
			}
		}
		pos[jn] = ch.insert( ch.end(), choice(*j, tag) );
		++jn;
	}

	choice_container::iterator i = ch.begin();;
	int key = run_choice(" Select group set",GROUP_CHOICE_X,GROUP_CHOICE_Y,GROUP_CHOICE_DX,ch,i);

	if (key == TEXT_KEY_ENTER) {
		rs.include_group.erase(rs.include_group.begin(), rs.include_group.end());
		for(choice_container::const_iterator j=ch.begin();j!=ch.end();++j) {
			if (j->bistate)
				rs.include_group.insert( rs.include_group.end(), j->desc );
		}
	}
}

// ------------------------------------------------------------------------
// Type menu

#define TYPE_CHOICE_X text_dx_get()/8
#define TYPE_CHOICE_Y 2
#define TYPE_CHOICE_DX 30

void run_type(menu_state& rs, const type_container& typ) {
	choice_container ch;
	choice_container::iterator pos[32];

	int jn = 0;
	for(type_container::const_iterator j = typ.begin();j!=typ.end() && jn<32;++j) {
		bool tag = false;
		for(type_container::const_iterator k = rs.include_type.begin();k!=rs.include_type.end();++k) {
			if (*j == *k) {
				tag = true;
				break;
			}
		}
		pos[jn] = ch.insert( ch.end(), choice(*j, tag) );
		++jn;
	}

	choice_container::iterator i = ch.begin();;
	int key = run_choice(" Select category set",TYPE_CHOICE_X,TYPE_CHOICE_Y,TYPE_CHOICE_DX,ch,i);

	if (key == TEXT_KEY_ENTER) {
		rs.include_type.erase(rs.include_type.begin(), rs.include_type.end());
		for(choice_container::const_iterator j=ch.begin();j!=ch.end();++j) {
			if (j->bistate)
				rs.include_type.insert( rs.include_type.end(), j->desc );
		}
	}
}

// ------------------------------------------------------------------------
// Move menu

void run_group_move(menu_state& rs, const group_container& grp) {
	choice_container ch;
	choice_container::iterator i = ch.begin();

	for(group_container::const_iterator j = grp.begin();j!=grp.end();++j) {
		choice_container::iterator k = ch.insert( ch.end(), choice(*j) );
		if (rs.pgame->group_get() == *j)
			i = k;
	}

	int key = run_choice(" Select group",GROUP_CHOICE_X,GROUP_CHOICE_Y,GROUP_CHOICE_DX,ch,i);

	if (key == TEXT_KEY_ENTER) {
		rs.pgame->group_set( i->desc );
	}
}

void run_type_move(menu_state& rs, const type_container& typ) {
	choice_container ch;
	choice_container::iterator i = ch.begin();

	for(type_container::const_iterator j = typ.begin();j!=typ.end();++j) {
		choice_container::iterator k = ch.insert( ch.end(), choice(*j) );
		if (rs.pgame->type_get() == *j)
			i = k;
	}

	int key = run_choice(" Select type",TYPE_CHOICE_X,TYPE_CHOICE_Y,TYPE_CHOICE_DX,ch,i);

	if (key == TEXT_KEY_ENTER) {
		rs.pgame->type_set( i->desc );
	}
}

// ------------------------------------------------------------------------
// Clone menu

#define CLONE_CHOICE_X text_dx_get()/10
#define CLONE_CHOICE_Y text_dy_get()/5
#define CLONE_CHOICE_DX	text_dx_get()*4/5

void run_clone(menu_state& rs, const game_set& gar) {
	choice_container ch;
	choice_container::iterator pos[32];
	const game* game[32];
	int jn = 0;
	choice_container::iterator i;

	for(game_set::const_iterator j = gar.begin();j!=gar.end();++j) {
		if (gar.is_game_clone_of(j->name_get(),rs.pgame->name_get())) {
			if (jn < 32) {
				ostrstream s;
				s << j->description_get() << ", " << j->manufacturer_get() << ", " << j->year_get() << ends;
				pos[jn] = ch.insert( ch.end(), choice( s.str() ) );
				s.freeze(0);
				game[jn] = &*j;
				if (j->name_get() == rs.pgame->name_get())
					i = pos[jn];
				++jn;
			}
		}
	}

	int key = run_choice(" Select clone",CLONE_CHOICE_X,CLONE_CHOICE_Y,CLONE_CHOICE_DX,ch,i);

	if (key == TEXT_KEY_ENTER) {
		int j;
		for(j=0;j<jn;++j) {
			if (pos[j] == i) {
				rs.pclone = game[j];
				return;
			}
		}
	}

	rs.pclone = 0;
}

