/*
 * This file is part of the AdvanceMAME project
 *
 * Copyright (C) 1999-2000 Andrea Mazzoleni
 *
 * The contents of this file are free for non-commercial use.
 * Any commercial use is forbidden.
 */

#include "game.h"
#include "menu.h"
#include "text.h"
#include "bitmap.h"
#include "info.h"

#include <pc.h>
#include <allegro.h>
#include <stdlib.h>
#include <dos.h>
#include <dirent.h>
#include <stdio.h>
#include <time.h>
#include <unistd.h>
#include <sys/exceptn.h>
#include <sys/stat.h>

#include <strstream>
#include <fstream>

#define FILE_FONT "mm.fnt"
#define FILE_MAME_INFO "info.txt"
#define FILE_CFG "mm.cfg"
#define FILE_MAME "mame.exe"

extern "C" void video_log(const char* format, ...) {
	(void)format;
}

void sound_signal(void) {
	unsigned i;
	for(i=0;i<10;++i) {
		sound(800);
		delay(10);
		nosound();
		delay(10);
	}
}

// ------------------------------------------------------------------------
// Conversion

string tristate2string(tristate_t v) {
	switch (v) {
		case include : return "include";
		case exclude : return "exclude";
		case exclude_not : return "exclude_not";
	}
	return "";
}

tristate_t string2tristate(const string& v) {
	if (v=="include") {
		return include;
	} else if (v=="exclude") {
		return exclude;
	} else if (v=="exclude_not") {
		return exclude_not;
	}
	return (tristate_t)-1;
}

string sort2string(game_sort_t v) {
	switch (v) {
		case sort_by_group : return "group";
		case sort_by_root_name : return "parent";
		case sort_by_name : return "name";
		case sort_by_time : return "time";
		case sort_by_year : return "year";
		case sort_by_manufacturer : return "manufacturer";
		case sort_by_type : return "type";
		case sort_by_size : return "size";
	}
	return "";
}

game_sort_t string2sort(const string& v) {
	if (v=="group") {
		return sort_by_group;
	} else if (v=="name") {
		return sort_by_name;
	} else if (v=="parent") {
		return sort_by_root_name;
	} else if (v=="time") {
		return sort_by_time;
	} else if (v=="year") {
		return sort_by_year;
	} else if (v=="manufacturer") {
		return sort_by_manufacturer;
	} else if (v=="type") {
		return sort_by_type;
	} else if (v=="size") {
		return sort_by_size;
	}
	return (game_sort_t)-1;
}

// ------------------------------------------------------------------------									   // ------------------------------------------------------------------------
// Load

void scan_game(const game_set& gar, const string& path, const string& name) {
	game_set::const_iterator i = gar.find( game(name) );
        if (i!=gar.end()) {
        	i->rom_zip_set_add(path);
        }
}

void scan_dir(const game_set& gar, const string& dir) {
	DIR* d = opendir(dir.c_str());
        if (!d) {
        	cerr << "Error opening directory " << dir << endl;
        	return;
        }

        struct dirent* dd;

        while ((dd = readdir(d))!=0) {
        	if (strstr(dd->d_name,".ZIP")!=0 || strstr(dd->d_name,".zip")!=0) {
                        const char* file = dd->d_name;
                        const char* dot = strchr(dd->d_name,'.');
                       	string path = string(dir) + "/" + string(dd->d_name);
                        string name = string(file,dot-file);
			scan_game(gar,path,name);
                }
        }

        closedir(d);
}

void scan_dirlist(const game_set& gar, const string& dirlist) {
	unsigned i = 0;
	while (i<dirlist.length()) {
		unsigned end = dirlist.find(';',i);
                if (end==string::npos) {
                	scan_dir(gar, string( dirlist, i ));
                        i = dirlist.size();
                } else {
	                scan_dir(gar, string( dirlist, i, end-i ));
                        i = end + 1;
                }
        }
}

void scan_cfg(string& snap, const game_set& gar, unsigned& orientation, double& gamma, double& brightness) {
	char* s;

	orientation = 0;
	s = get_config_string("config", "ror", "no");
	if (stricmp(s,"yes")==0) {
		orientation ^= ORIENTATION_FLIP_XY | ORIENTATION_MIRROR_X;
	}

	s = get_config_string("config", "rol", "no");
	if (stricmp(s,"yes")==0) {
		orientation ^= ORIENTATION_FLIP_XY | ORIENTATION_MIRROR_Y;
	}

	s = get_config_string("config", "flipx", "no");
	if (stricmp(s,"yes")==0) {
		orientation ^= ORIENTATION_MIRROR_X;
	}

	s = get_config_string("config", "flipy", "no");
	if (stricmp(s,"yes")==0) {
		orientation ^= ORIENTATION_MIRROR_Y;
	}

	gamma = get_config_float("config", "gamma", 1.0);
	brightness = get_config_float("config", "brightness", 1.0);

	s = get_config_string("directory", "rompath", "roms");
	scan_dirlist( gar, s);

	s = get_config_string("directory", "snap", "snap");
	snap = s;

	s = get_config_string("directory", "altss", "");
	string altss = s;

	if (altss.length()) {
		if (snap.length())
			snap += ";";
		snap += altss;
	}
}

void scan_info(game_set& gar, const string& info, const string& mame) {
	struct stat st_info;
	struct stat st_mame;
	int err_info;
	int err_mame;

	err_info = stat(info.c_str(), &st_info);
	err_mame = stat(mame.c_str(), &st_mame);

	if (err_mame==0 && (err_info!=0 || st_info.st_mtime < st_mame.st_mtime)) {
		cerr << "Creating the file " << info << endl;
		char cmd[256];
		sprintf(cmd,"%s -listinfo > %s",mame.c_str(),info.c_str());
		__djgpp_exception_toggle();
	        int r = system(cmd);
		__djgpp_exception_toggle();

	        if (r == -1) {
	        	cerr << "DOS error " << -r << endl;
	        } else if (r & 0xFF != 0) {
        		cerr << "MAME error " << (r & 0xFF) << endl;
	        } else if ((r >> 8) & 0xFF != 0) {
	        	cerr << "MAME signal " << ((r >> 8) & 0xFF) << endl;
		}
	}

        FILE* f = fopen(info.c_str(),"rt");
	if (!f) {
		cerr << "File " << info << " not found!" << endl;
		cerr << "Run the command: MAME -listinfo > " << info << endl;
		exit(EXIT_FAILURE);
	}
        if (!gar.load(f)) {
        	cerr << "Error loading game informations from file " << info << " at row " << info_row_get()+1 << " column " << info_col_get()+1 << endl;
		exit(EXIT_FAILURE);
        }
        if (!gar.test()) {
		cerr << "Error in informations from file " << info << endl;
		exit(EXIT_FAILURE);
	}
        fclose(f);
}

void group_insert(group_container& grp, const string& name) {
	group_container::iterator j;
	for(j=grp.begin();j!=grp.end();++j) {
		if (*j == name)
			break;
	}

	if (j==grp.end())
		grp.insert(grp.end(), name);
}

void type_insert(type_container& typ, const string& name) {
	type_container::iterator j;
	for(j=typ.begin();j!=typ.end();++j) {
		if (*j == name)
			break;
	}

	if (j==typ.end())
		typ.insert(typ.end(), name);
}

bool scan_game(game_set& gar, group_container& grp, type_container& typ, const string& name, const string& group, const string& type, const string& time) {
       	game_set::const_iterator i = gar.find( game( name ) );
	if (i==gar.end())
		return false;

	i->group_set(group);
	i->type_set(type);
	i->time_set(atoi(time.c_str()));

	group_insert(grp,group);
	type_insert(typ,type);

	return true;
}

void scan_grp(menu_state& rs, game_set& gar, group_container& grp, type_container& typ, const string& file) {
	ifstream f(file.c_str(), ios::in );
        if (!f) {
#if 0
        	cerr << "Error opening configuration file " << file << endl;
                exit(EXIT_FAILURE);
#else
		group_insert(grp,"Very Good");
		group_insert(grp,"Good");
		group_insert(grp,"Bad");
		type_insert(typ,"Arcade");
		type_insert(typ,"Shot'em Up");
		type_insert(typ,"Bet'em Up");
		type_insert(typ,"Fighting");
		type_insert(typ,"Gun");
		type_insert(typ,"Puzzle");
		type_insert(typ,"RPG");
		type_insert(typ,"Sport");
		type_insert(typ,"Bowling");
		type_insert(typ,"Breakout");
		type_insert(typ,"Filler");
		type_insert(typ,"Racing");
		type_insert(typ,"Flipper");
#endif
        } else {

	        while (!f.eof()) {
	        	string s;
	                getline(f,s,'\n');

			string cmd;

			unsigned i = 0;

			while (i < s.length() && isspace(s[i]))
				++i;

			if (i<s.length() && s[0]!='#') {
				while (i < s.length() && !isspace(s[i])) {
					cmd += s[i];
					++i;
				}

				while (i < s.length() && isspace(s[i]))
					++i;

				string value = string(s,i);
				while (value.length() && isspace(value[value.length()-1]))
					value.erase(value.length()-1,1);

				if (cmd == "group" || cmd == "type" || cmd == "select_group" || cmd == "select_type") {
					string name;

         				while (i < s.length() && isspace(s[i]))
         					++i;

         				if (i >= s.length() || s[i]!='"') {
         		        		cerr << "Error in configuration file " << file << endl;
	         				cerr << "At line " << s << endl;
        	 		                exit(EXIT_FAILURE);
					}
       					++i;

	       				while (i < s.length() && s[i]!='"') {
	       					name += s[i];
	       					++i;
	       				}
         
	       				if (name.length()==0 || i >= s.length() || s[i]!='"') {
	         		        	cerr << "Error in configuration file " << file << endl;
	         				cerr << "At line " << s << endl;
	        		                exit(EXIT_FAILURE);
					}
					++i;

					while (i < s.length() && isspace(s[i]))
	         				++i;

	         			if (i<s.length()) {
	         		        	cerr << "Error in configuration file " << file << endl;
	         				cerr << "At line " << s << endl;
	         		                exit(EXIT_FAILURE);
	         			}

					if (cmd == "group") {
						group_insert(grp,name);
					} else if (cmd == "type") {
						type_insert(typ,name);
					} else if (cmd == "select_group") {
						group_insert(rs.include_group,name);
					} else if (cmd == "select_type") {
						type_insert(rs.include_type,name);
					}
				} else if (cmd == "game") {
					string game;
					string group;
					string type;
					string time;

	         			while (i < s.length() && !isspace(s[i])) {
	         				game += s[i];
	         				++i;
	         			}
         
	         			while (i < s.length() && isspace(s[i]))
	         				++i;

	         			if (i >= s.length() || s[i]!='"') {
	         		        	cerr << "Error in configuration file " << file << endl;
	         				cerr << "At line " << s << endl;
	         		                exit(EXIT_FAILURE);
					}
	       				++i;
         
	       				while (i < s.length() && s[i]!='"') {
	       					group += s[i];
	       					++i;
	       				}

					if (group.length()==0)
						group = GROUP_UNDEFINED;

	       				if (i >= s.length() || s[i]!='"') {
	         		        	cerr << "Error in configuration file " << file << endl;
	         				cerr << "At line " << s << endl;
	        		                exit(EXIT_FAILURE);
					}
					++i;

	        			while (i < s.length() && isspace(s[i]))
	         				++i;

	         			if (i >= s.length() || s[i]!='"') {
	         		        	cerr << "Error in configuration file " << file << endl;
	         				cerr << "At line " << s << endl;
	         		                exit(EXIT_FAILURE);
					}
	       				++i;
         
	       				while (i < s.length() && s[i]!='"') {
	       					type += s[i];
	       					++i;
	       				}

					if (type.length()==0)
						type = TYPE_UNDEFINED;

	       				if (i >= s.length() || s[i]!='"') {
	         		        	cerr << "Error in configuration file " << file << endl;
	         				cerr << "At line " << s << endl;
	        		                exit(EXIT_FAILURE);
					}
					++i;

	         			while (i < s.length() && isspace(s[i]))
	         				++i;

	         			while (i < s.length() && !isspace(s[i])) {
	         				time += s[i];
	         				++i;
	         			}

		      			while (i < s.length() && isspace(s[i]))
	         				++i;

	         			if (time.length() == 0 || i<s.length()) {
	         		        	cerr << "Error in configuration file " << file << endl;
	         				cerr << "At line " << s << endl;
	         		                exit(EXIT_FAILURE);
	         			}

	         			if (!scan_game(gar,grp,typ,game,group,type,time)) {
#if 0 /* ignore unknow */
         				cerr << "Error in configuration file " << file << endl;
         				cerr << "At line " << s << endl;
         				cerr << "Unknow game " << game << endl;
         		                exit(EXIT_FAILURE);
#endif
	         			}
				} else if (cmd == "select_neogeo") {
					rs.exclude_neogeo = string2tristate(value);
				} else if (cmd == "select_clone") {
					rs.exclude_clone = string2tristate(value);
				} else if (cmd == "select_bad") {
					rs.exclude_bad = string2tristate(value);
				} else if (cmd == "select_missing") {
					rs.exclude_missing = string2tristate(value);
				} else if (cmd == "select_vector") {
					rs.exclude_vector = string2tristate(value);
				} else if (cmd == "sort") {
					rs.sort_mode = string2sort(value);
					if (rs.sort_mode < 0) {
	         		        	cerr << "Error in configuration file " << file << endl;
	         				cerr << "At line " << s << endl;
						cerr << "Unknow sort mode " << value << endl;
	        		                exit(EXIT_FAILURE);
					}
				} else if (cmd == "event") {
					if (!text_key_in(value)) {
	         		        	cerr << "Error in configuration file " << file << endl;
	         				cerr << "At line " << s << endl;
	        		                exit(EXIT_FAILURE);
					}
				} else if (cmd == "idle_start") {
					sscanf(value.c_str(),"%d %d",&rs.idle_start,&rs.idle_start_rep);
				} else if (cmd == "last_game") {
					rs.last_game = value;
				} else if (cmd == "idle_move") {
					sscanf(value.c_str(),"%d %d",&rs.idle_move,&rs.idle_move_rep);
				} else if (cmd == "merge") {
					if (value == "no")
						rs.merge = merge_no;
					else if (value == "differential")
						rs.merge = merge_differential;
					else if (value == "parent")
						rs.merge = merge_parent;
					else if (value == "any")
						rs.merge = merge_any;
					else if (value == "disable")
						rs.merge = merge_disable;
					else {
	         		        	cerr << "Error in configuration file " << file << endl;
	         				cerr << "At line " << s << endl;
						cerr << "Unknow merge mode " << value << endl;
	        		                exit(EXIT_FAILURE);
					}
				} else {
	       				cerr << "Error in configuration file " << file << endl;
	       				cerr << "At line " << s << endl;
					cerr << "Invalid command" << endl;
				}
			}
		}

		f.close();
	}

	for(game_set::const_iterator i=gar.begin();i!=gar.end();++i)
		if (i->group_get().length()==0)
			i->group_set(GROUP_UNDEFINED);

	group_insert(grp,GROUP_UNDEFINED);
	type_insert(typ,TYPE_UNDEFINED);

	if (rs.include_group.size() == 0)
		rs.include_group = grp;
	if (rs.include_type.size() == 0)
		rs.include_type = typ;
}

void save_grp(const menu_state& rs, const game_set& gar, const group_container& grp, const type_container& typ, const string& file) {
	// skip if no write access
	if (access(file.c_str(), R_OK) == 0 && access(file.c_str(), W_OK) != 0)
		return;

	ofstream f(file.c_str(), ios::out );
        if (!f) {
        	cerr << "Error opening for writing the configuration file " << file << endl;
                exit(EXIT_FAILURE);
        }

	f << "# Events" << endl;
	f << "# event EVENT KEY0 KEY1" << endl;
	text_key_out_all(f);
	f << endl;

	f << "# Merge" << endl;
	f << "# merge no | parent | differential | any | disable" << endl;
	f << "merge ";
	switch (rs.merge) {
		case merge_no : f << "no"; break;
		case merge_parent : f << "parent"; break;
		case merge_differential : f << "differential"; break;
		case merge_any : f << "any"; break;
		case merge_disable : f << "disable"; break;
	}
	f << endl;
	f << endl;

	f << "# Idle" << endl;
	f << "# (idle_start | idle_move) SECOND_START SECOND_REPEAT" << endl;
	f << "idle_start " << rs.idle_start << " " << rs.idle_start_rep << endl;
	f << "idle_move " << rs.idle_move << " " << rs.idle_move_rep << endl;
	f << endl;

	f << "# Last game" << endl;
	f << "# last_game GAMENAME" << endl;
	if (rs.last_game.length()) {
		f << "last_game " << rs.last_game << endl;
		f << endl;
	}

	f << "# Sort mode" << endl;
	f << "# sort (group | parent | name | time | year | manufacturer | type | size)" << endl;
	f << "sort " << sort2string(rs.sort_mode) << endl;
	f << endl;

	f << "# Game selected" << endl;
	f << "# select_(neogeo | clone | bad | missing | vector) (include | exclude | exclude_not)" << endl;
	f << "select_neogeo " << tristate2string(rs.exclude_neogeo) << endl;
	f << "select_clone " << tristate2string(rs.exclude_clone) << endl;
	f << "select_bad " << tristate2string(rs.exclude_bad) << endl;
	f << "select_missing " << tristate2string(rs.exclude_missing) << endl;
	f << "select_vector " << tristate2string(rs.exclude_vector) << endl;
	f << endl;

	f << "# Group set" << endl;
	f << "# group STRING" << endl;
	for(group_container::const_iterator i=grp.begin();i!=grp.end();++i) {
		if (*i!=GROUP_UNDEFINED) {
			f << "group \"" << *i << "\"" << endl;
		}
	}
	f << endl;

	f << "# Group selected" << endl;
	f << "# select_group STRING" << endl;
	for(group_container::const_iterator i=rs.include_group.begin();i!=rs.include_group.end();++i) {
		f << "select_group \"" << *i << "\"" << endl;
	}
	f << endl;

	f << "# Type set" << endl;
	f << "# type = STRING" << endl;
	for(group_container::const_iterator i=typ.begin();i!=typ.end();++i) {
		if (*i!=TYPE_UNDEFINED) {
			f << "type \"" << *i << "\"" << endl;
		}
	}
	f << endl;

	f << "# Type selected" << endl;
	f << "# select_type = STRING" << endl;
	for(group_container::const_iterator i=rs.include_type.begin();i!=rs.include_type.end();++i) {
		f << "select_type \"" << *i << "\"" << endl;
	}
	f << endl;

	f << "# Game list" << endl;
	f << "# game GROUP_STRING TYPE_STRING SECONDS_PLAYED" << endl;
	for(game_set::const_iterator i=gar.begin();i!=gar.end();++i) {
		if ((i->group_get().length()!=0 && i->group_get()!=GROUP_UNDEFINED) ||
	 		(i->type_get().length()!=0 && i->type_get()!=TYPE_UNDEFINED) ||
			i->time_get()!=0) {
			f << "game " << i->name_get();
 			f << " \"";
			if (i->group_get()!=GROUP_UNDEFINED)
				f << i->group_get();
			f << "\"";
			f << " \"";
			if (i->type_get()!=TYPE_UNDEFINED)
				f << i->type_get();
			f << "\"";
			f << " " << i->time_get();
 			f << endl;
		}
	}
	f << endl;

	f.close();
}

void scan_arg(string& snap, string& font, int argc, char* argv[]) {
	int c;
	while ((c=getopt(argc, argv, "s:f:")) != -1) {
		switch (c) {
			case 's' :
				snap = snap + ";" + optarg;
				break;
			case 'f' :
				font = optarg;
				break;
			case '?':
				printf("Unknown option %c\n", c);
				exit(1);
		}
	}
}

// --------------------------------------------------------------------------
// Run

#define HELP_COLOR_BAR 0x70
#define HELP_COLOR_NORMAL 0x0B
#define HELP_COLOR_BOLD 0x0F

void run_help() {
	text_fill(0,0,text_dx_get(),' ',HELP_COLOR_BAR);
	text_put(2,0,"HELP",HELP_COLOR_BAR);

	text_fill(0,1,text_dx_get(),text_dy_get()-1,' ',HELP_COLOR_NORMAL);

	int y = 2;
	int xt = 0;
	int xd = 6;
	text_put(xt,y,"F1",HELP_COLOR_BOLD);
	text_put(xd,y,"Help screen",HELP_COLOR_NORMAL);
	++y;
	text_put(xt,y,"F2",HELP_COLOR_BOLD);
	text_put(xd,y,"Include specific games by group",HELP_COLOR_NORMAL);
	++y;
	text_put(xt,y,"F3",HELP_COLOR_BOLD);
	text_put(xd,y,"Include specific games by type",HELP_COLOR_NORMAL);
	++y;
	text_put(xt,y,"F4",HELP_COLOR_BOLD);
	text_put(xd,y,"Exclude specific games by attribute",HELP_COLOR_NORMAL);
	++y;
	text_put(xt,y,"F5",HELP_COLOR_BOLD);
	text_put(xd,y,"Select the game sort method",HELP_COLOR_NORMAL);
	++y;
	text_put(xt,y,"F9",HELP_COLOR_BOLD);
	text_put(xd,y,"Change the current game group",HELP_COLOR_NORMAL);
	++y;
	text_put(xt,y,"F10",HELP_COLOR_BOLD);
	text_put(xd,y,"Change the current game type",HELP_COLOR_NORMAL);
	++y;
	text_put(xt,y,"F12",HELP_COLOR_BOLD);
	text_put(xd,y,"Run a clone",HELP_COLOR_NORMAL);
	++y;
	text_put(xt,y,"ESC",HELP_COLOR_BOLD);
	text_put(xd,y,"Exit",HELP_COLOR_NORMAL);
	++y;
	++y;
	text_put(xt,y,"In selection:",HELP_COLOR_NORMAL);
	++y;
	text_put(xt,y,"ENTER",HELP_COLOR_BOLD);
	text_put(xd,y,"Accept",HELP_COLOR_NORMAL);
	++y;
	text_put(xt,y,"-",HELP_COLOR_BOLD);
	text_put(xd,y,"Deselect all",HELP_COLOR_NORMAL);
	++y;
	text_put(xt,y,"SPACE",HELP_COLOR_BOLD);
	text_put(xd,y,"Toggle (+ include, - exclude, * required)",HELP_COLOR_NORMAL);
	++y;
	text_put(xt,y,"ESC",HELP_COLOR_BOLD);
	text_put(xd,y,"Cancel",HELP_COLOR_NORMAL);
	++y;

	text_backdrop_clear();
	text_getkey();
}

int run_main(const string& font,const game_set& gar, const group_container& grp, const group_container& typ, unsigned orientation, double gamma, double brightness, menu_state& rs) {
	if (!text_sub_init(font.c_str(),orientation,gamma,brightness,rs.idle_start,rs.idle_start_rep,rs.idle_move,rs.idle_move_rep)) {
		cerr << "Error inizializing the video system" << endl;
		exit(EXIT_FAILURE);
	}

        bool done = false;
        int key = 0;

        while (!done) {
		key = run_menu(gar,rs, (orientation & ORIENTATION_FLIP_XY) != 0);
                switch (key) {
			case TEXT_KEY_HELP :
				run_help();
				break;
			case TEXT_KEY_GROUP :
				run_group(rs,grp);
				break;
			case TEXT_KEY_TYPE :
				run_type(rs,typ);
				break;
			case TEXT_KEY_EXCLUDE :
				run_exclude(rs);
				break;
			case TEXT_KEY_SORT :
				run_sort(rs);
				break;
			case TEXT_KEY_SETGROUP :
				if (rs.pgame)
					run_group_move(rs,grp);
				break;
			case TEXT_KEY_SETTYPE :
				if (rs.pgame)
					run_type_move(rs,typ);
				break;
			case TEXT_KEY_MODE :
				rs.backdrop_merge = !rs.backdrop_merge;
				break;
			case TEXT_KEY_RUN_CLONE :
				if (rs.pgame) {
					run_clone(rs,gar);
					if (rs.pclone)
						done = true;
				}
				break;
			case TEXT_KEY_IDLE_0 :
			case TEXT_KEY_ESC :
                        case TEXT_KEY_ENTER :
                        	done = true;
				break;
                }
        }

         text_sub_done();

         return key;
}

/*
INT 2F - SMARTDRV v4.00+ - FLUSH BUFFERS (COMMIT CACHE)
	AX = 4A10h
	BX = 0001h
Note:	this function is also supported by PC-Cache v8.0.
*/
static void flush_cache() {
	union REGS r;
	r.x.ax = 0x4A10;
	r.x.bx = 0x0001;
	int86(0x2F, &r, &r);
}

void run_game(const game& g) {
	char cmd[256];
	time_t start,stop;

	flush_cache();

	start = time(0);

	printf("%s, %s, %s\n", g.description_get().c_str(), g.manufacturer_get().c_str(), g.year_get().c_str());
	sprintf(cmd,"mame %s", g.name_get().c_str());
	printf("%s\n",cmd);

	__djgpp_exception_toggle();
        int r = system(cmd);
	__djgpp_exception_toggle();

        stop = time(0);
	if (stop - start > 8) {
		g.time_set( g.time_get() + stop - start );
	}

        if (r == -1) {
        	cerr << "DOS error " << -r << endl;
		sleep(8);
        } else if (r & 0xFF != 0) {
        	cerr << "MAME error " << (r & 0xFF) << endl;
		sleep(8);
        } else if ((r >> 8) & 0xFF != 0) {
        	cerr << "MAME signal " << ((r >> 8) & 0xFF) << endl;
		sleep(8);
	}

	flush_cache();
}

int run_all(menu_state& rs, const string& snap, const string& font, const game_set& gar, const group_container& grp, const type_container& typ, unsigned orientation, double gamma, double brightness) {
        bool done = false;
        int key = 0;

	rs.snap = snap;
        rs.pos_base = 0;
	rs.pos_rel = 0;
	rs.pgame = 0;
	rs.pclone = 0;
	rs.fast = "";
	rs.backdrop_merge = false;

        while (!done) {
		key = run_main(font,gar,grp,typ,orientation,gamma,brightness,rs);
                switch (key) {
			case TEXT_KEY_ESC :
                        	done = true;
                                break;
                        case TEXT_KEY_ENTER :
			case TEXT_KEY_IDLE_0 :
                        	if (rs.pgame) {
					// save before
					save_grp(rs,gar,grp,typ,FILE_CFG);
                        		run_game(*rs.pgame);
					// save after
					save_grp(rs,gar,grp,typ,FILE_CFG);
				}
				break;
			case TEXT_KEY_RUN_CLONE :
                        	if (rs.pclone) {
					// save before
					save_grp(rs,gar,grp,typ,FILE_CFG);
                        		run_game(*rs.pclone);
					// save after
					save_grp(rs,gar,grp,typ,FILE_CFG);
				}
                }
        }

        return key;
}

int main(int argc, char* argv[]) {
	game_set gar;
	group_container grp;
	type_container typ;
	string snap = "";
	string font = FILE_FONT;
	unsigned orientation = 0;
	double gamma = 1;
	double brightness = 1;
	menu_state rs;

	rs.idle_start = 0;
	rs.idle_start_rep = 0;
	rs.idle_move = 40;
	rs.idle_move_rep = 5;

	rs.sort_mode = sort_by_root_name;
	rs.exclude_neogeo = include;
	rs.exclude_clone = exclude;
	rs.exclude_bad = exclude;
        rs.exclude_missing = include;
	rs.exclude_vector = include;

	rs.merge = merge_differential;

	srand(time(0));

	if (!text_init()) {
		exit(1);
	}

	scan_info(gar,FILE_MAME_INFO,FILE_MAME);

        scan_cfg(snap,gar,orientation,gamma,brightness);

	scan_grp(rs,gar,grp,typ,FILE_CFG);

	scan_arg(snap,font,argc,argv);

	sound_signal();

        run_all(rs,snap,font,gar,grp,typ,orientation,gamma,brightness);

	save_grp(rs,gar,grp,typ,FILE_CFG);

	text_done();

        return EXIT_SUCCESS;
}

extern "C" void videoerror(const char *text,...) {
	/* no error report */
}
