/*
 * This file is part of the AdvanceMAME project
 *
 * Copyright (C) 1999-2000 Andrea Mazzoleni
 *
 * The contents of this file are free for non-commercial use.
 * Any commercial use is forbidden.
 */

#include "scrblit.h"

#include <sys/movedata.h>
#include <sys/segments.h>
#include <sys/farptr.h>
#include <assert.h>
#include <stdlib.h>

/***************************************************************************/
/* Clear */

/* Clear screen
 * in:
 *   x,y destination pos
 *   src data source
 */
static void video_chained8_clear(unsigned x, unsigned y, unsigned dx, unsigned dy, uint8 src) {
	unsigned j;
	unsigned pre_count;
	unsigned post_count;
	unsigned inner_count;
	unsigned sel = video_write_selector();
	uint32 mask;

	src &= 0xFF;
	mask = src | (src << 8) | (src << 16) | (src << 24);

	pre_count = (4 - x) & 3;
	if (pre_count >= dx) pre_count = dx;
	inner_count = (dx - pre_count) / 4;
	post_count = dx - pre_count - inner_count * 4;

	_farsetsel(sel);

	for(j=0;j<dy;++j) {
		unsigned dst = video_write_line(y) + x;
		unsigned i;
		for(i=pre_count;i;--i) {
			_farnspokeb(dst, mask);
			dst += 1;
		}
		for(i=inner_count;i;--i) {
			_farnspokel(dst, mask);
			dst += 4;
		}
		for(i=post_count;i;--i) {
			_farnspokeb(dst, mask);
			dst += 1;
		}
		++y;
	}
}

static void video_chained16_clear(unsigned x, unsigned y, unsigned dx, unsigned dy, unsigned src) {
	unsigned j;
	unsigned pre_count;
	unsigned post_count;
	unsigned inner_count;
	unsigned sel = video_write_selector();
	uint32 mask;

	src &= 0xFFFF;
	mask = src | (src << 16);

	pre_count = (2 - x) & 1;
	if (pre_count >= dx) pre_count = dx;
	inner_count = (dx - pre_count) / 2;
	post_count = dx - pre_count - inner_count * 2;

	_farsetsel(sel);

	for(j=0;j<dy;++j) {
		unsigned dst = video_write_line(y) + 2*x;
		unsigned i;
		if (pre_count) {
			_farnspokeb(dst, mask);
			_farnspokeb(dst + 1, mask >> 8);
			dst += 2;
		}
		for(i=inner_count;i;--i) {
			_farnspokel(dst, mask);
			dst += 4;
		}
		if (post_count) {
			_farnspokeb(dst, mask);
			_farnspokeb(dst + 1, mask >> 8);
			/* dst += 2; */
		}
		++y;
	}
}

static void video_unchained8_clear(unsigned x, unsigned y, unsigned dx, unsigned dy, uint8 src) {
	unsigned p;
	unsigned sel = video_write_selector();
	uint32 mask;

	src &= 0xFF;
	mask = src | (src << 8) | (src << 16) | (src << 24);

	_farsetsel(sel);

	for(p=0;p<4;++p) {
		unsigned j;

		unsigned p_x;
		unsigned p_dx;
		unsigned pre_count;
		unsigned post_count;
		unsigned inner_count;

		p_x = x / 4;
		p_dx = (x + dx - 1) / 4 - p_x + 1;
		if ((x + dx - 1) % 4 < p)
			--p_dx;
		if (x % 4 > p) {
			--p_dx;
			++p_x;
		}

		pre_count = (4 - p_x) & 3;
		if (pre_count >= p_dx) pre_count = p_dx;
		inner_count = (p_dx - pre_count) / 4;
		post_count = p_dx - pre_count - inner_count * 4;

		video_unchained_plane_set(p);
		for(j=0;j<dy;++j) {
			unsigned dst = video_write_line(y+j) + p_x;
			unsigned i;
			for(i=pre_count;i;--i) {
				_farnspokeb(dst, mask);
				dst += 1;
			}
			for(i=inner_count;i;--i) {
				_farnspokel(dst, mask);
				dst += 4;
			}
			for(i=post_count;i;--i) {
				_farnspokeb(dst, mask);
				dst += 1;
			}
		}
	}
}

/* Clear the screen */
void video_clear(unsigned dst_x, unsigned dst_y, unsigned dst_dx, unsigned dst_dy, unsigned src) {
	switch (video_bytes_per_pixel()) {
		case 1 :
			if (video_unchained_active())
				video_unchained8_clear(dst_x,dst_y,dst_dx,dst_dy,src);
			else
				video_chained8_clear(dst_x,dst_y,dst_dx,dst_dy,src);
			break;
		case 2:
			video_chained16_clear(dst_x,dst_y,dst_dx,dst_dy,src);
			break;
	}
}

/***************************************************************************/
/* internal */

/* Add a byte index to a generic pointer */
/* #define PADD(p,i) ((uint8*)p) += i */
#define PADD(p,i) *((uint8**)&p) += i /* pedantic */

/* Dereference a uint8* with a byte index */
#define P8DER(p,i) *(((uint8*)p) + i)
#define P8DER0(p) *(uint8*)p

/* Dereference a uint16* with a byte index */
#define P16DER(p,i) *(uint16*)(((uint8*)p) + i)
#define P16DER0(p) *(uint16*)p

/* Notes
1) The count argument is the number of source element to process
2) The count divisor at early stage of every functions is the number of
	source element processed in one subsequent loop iteration
3) The step argument refere always to bytes
4) All the internal_* functions use the fs register for the destination when
	is expressed as an unsigned, before the call the fs register
	must be set at the correct value
5) After any internal_* functions the internal_end function must be called
	before any use of the FPU (float or double operations)
*/

#if defined( __pentiumpro ) || defined( __k6 )

#if defined( __pentium )
#error Too many pentium, pentiumpro, k6 defines
#endif

static __inline__ void internal_copy8(unsigned dst, void* src, unsigned count)
{
	unsigned rest = count % 8;

	assert(((unsigned)src & 0x7)==0 && (dst & 0x7)==0);

	__asm__ __volatile__(
		"shr $3,%2\n"
		"jz 1f\n"
		"0:\n"
		"movq (%0),%%mm0\n"
		"movq %%mm0,%%fs:(%1)\n"
		"addl $8,%0\n"
		"addl $8,%1\n"
		"decl %2\n"
		"jnz 0b\n"
		"1:\n"
		: "+r" (src), "+r" (dst), "+r" (count)
		:
		: "cc"
	);

	switch (rest) {
		case 7 : _farnspokeb(dst + 6, ((uint8*)src)[6]);
		case 6 : _farnspokeb(dst + 5, ((uint8*)src)[5]);
		case 5 : _farnspokeb(dst + 4, ((uint8*)src)[4]);
		case 4 : _farnspokel(dst, ((uint32*)src)[0]);
		break;
		case 3 : _farnspokeb(dst + 2, ((uint8*)src)[2]);
		case 2 : _farnspokeb(dst + 1, ((uint8*)src)[1]);
		case 1 : _farnspokeb(dst, ((uint8*)src)[0]);
	}
}

static __inline__ void internal_double8(unsigned dst, void* src, unsigned count)
{
	assert(((unsigned)src & 0x7)==0 && (dst & 0x7)==0);
	__asm__ __volatile__(
		"shr $3,%2\n"
		"jz 1f\n"
		"0:\n"
		"movq (%0),%%mm0\n"
		"movq %%mm0,%%mm1\n"
		"punpcklbw %%mm0, %%mm0\n"
		"punpckhbw %%mm1, %%mm1\n"
		"movq %%mm0,%%fs:(%1)\n"
		"movq %%mm1,%%fs:8(%1)\n"
		"addl $8,%0\n"
		"addl $16,%1\n"
		"decl %2\n"
		"jnz 0b\n"
		"1:\n"
		: "+r" (src), "+r" (dst), "+r" (count)
		:
		: "cc"
	);
}

static __inline__ void internal_double16(unsigned dst, unsigned src, unsigned count)
{
	assert((src & 0x7)==0 && (dst & 0x7)==0);
	__asm__ __volatile__(
		"shr $2,%2\n"
		"jz 1f\n"
		"0:\n"
		"movq (%0),%%mm0\n"
		"movq %%mm0,%%mm1\n"
		"punpcklwd %%mm0, %%mm0\n"
		"punpckhwd %%mm1, %%mm1\n"
		"movq %%mm0,%%fs:(%1)\n"
		"movq %%mm1,%%fs:8(%1)\n"
		"addl $8,%0\n"
		"addl $16,%1\n"
		"decl %2\n"
		"jnz 0b\n"
		"1:\n"
		: "+r" (src), "+r" (dst), "+r" (count)
		:
 		: "cc"
	);
}

static __inline__ void internal_copy_end(void)
{
	__asm__ __volatile__ (
		"emms"
	);
}

#elif defined( __pentium )

#if defined( __pentiumpro ) || defined( __k6 )
#error Too many pentium, pentiumpro, k6 defines
#endif

static __inline__ void internal_copy8(unsigned dst, void* src, unsigned count)
{
	unsigned rest = count % 4;

	assert(((unsigned)src & 0x3)==0 && (dst & 0x3)==0);

	__asm__ __volatile__ (
		"shr $2,%2\n"
		"jz 1f\n"
		"0:\n"
		"movl (%0),%%eax\n"
		"movl %%eax,%%fs:(%1)\n"
		"addl $4,%0\n"
		"addl $4,%1\n"
		"decl %2\n"
		"jnz 0b\n"
		"1:\n"
		: "+r" (src), "+r" (dst), "+r" (count)
		:
		: "cc", "eax"
	);

	switch (rest) {
		case 3 : _farnspokeb(dst + 2, ((uint8*)src)[2]);
		case 2 : _farnspokeb(dst + 1, ((uint8*)src)[1]);
		case 1 : _farnspokeb(dst, ((uint8*)src)[0]);
	}
}

static __inline__ void internal_double8(unsigned dst, void* src, unsigned count)
{
	assert(((unsigned)src & 0x3)==0 && (dst & 0x3)==0);
	__asm__ __volatile__ (
		"shr $2,%2\n"
		"jz 1f\n"
		"0:\n"
		"movl (%0),%%eax\n"
		"movl %%eax, %%ebx\n"
		"bswap %%eax\n"
		"xchgw %%ax,%%bx\n"
		"roll $8, %%eax\n"
		"movl %%eax,%%fs:(%1)\n"
		"rorl $8, %%ebx\n"
		"movl %%ebx,%%fs:4(%1)\n"
		"addl $4,%0\n"
		"addl $8,%1\n"
		"decl %2\n"
		"jnz 0b\n"
		"1:\n"
		: "+r" (src), "+r" (dst), "+r" (count)
		:
		: "cc", "eax", "ebx"
	);
}

static __inline__ void internal_double16(unsigned dst, unsigned src, unsigned count)
{
	assert(((unsigned)src & 0x3)==0 && (dst & 0x3)==0);
	__asm__ __volatile__ (
		"shr $1,%2\n"
		"jz 1f\n"
		"0:\n"
		"movl (%0),%%eax\n"
		"movl %%eax, %%ebx\n"
		"roll $16, %%eax\n"
		"xchgw %%ax,%%bx\n"
		"movl %%eax,%%fs:(%1)\n"
		"movl %%ebx,%%fs:4(%1)\n"
		"addl $4,%0\n"
		"addl $8,%1\n"
		"decl %2\n"
		"jnz 0b\n"
		"1:\n"
		: "+r" (src), "+r" (dst), "+r" (count)
		:
		: "cc", "eax", "ebx"
	);
}

static __inline__ void internal_copy_end(void)
{
}

#else

#error No __pentiumpro,__pentium,__k6 define, use the compiler -march= option

static __inline__ void internal_copy8(unsigned dst, void* src, unsigned count) {
	uint32* src32 = (uint32*)src;
	unsigned rest = count % 4;

	count /= 4;
	while (count) {
		_farnspokel(dst, src32[0]);
		dst += 4;
		src32 += 1;
		--count;
	}

	switch (rest) {
		case 3 : _farnspokeb(dst + 2, ((uint8*)src32)[2]);
		case 2 : _farnspokeb(dst + 1, ((uint8*)src32)[1]);
		case 1 : _farnspokeb(dst, ((uint8*)src32)[0]);
	}
}

static __inline__ void internal_double8(unsigned dst, void* src, unsigned count) {
	uint8* src8 = (uint8*)src;

	count /= 2;
	while (count) {
		unsigned l0 = src8[0];
		unsigned l1 = src8[1];
		_farnspokel(dst, l0 | l0 << 8 | l1 << 16 | l1 << 24);
		dst += 4;
		src8 += 2;
		--count;
	}
}

static __inline__ void internal_double16(unsigned dst, unsigned src, unsigned count) {
	uint16* src16 = (uint16*)src;

	while (count) {
		unsigned l0 = src16[0];
		_farnspokel(dst, l0 | l0 << 16);
		dst += 4;
		src16 += 1;
		--count;
	}
}

static __inline__ void internal_copy_end(void)
{
}

#endif

static __inline__ void internal_copy16(unsigned dst, void* src, unsigned count) {
	internal_copy8(dst, src, 2*count);
}

static __inline__ void internal_copy8_step(unsigned dst, void* src, unsigned count, int step1) {
	int step2 = step1 + step1;
	int step3 = step2 + step1;
	int step4 = step3 + step1;
	uint8* src8 = (uint8*)src;

	count /= 4;
	while (count) {
		_farnspokel(dst, src8[0]
			| (unsigned)src8[step1] << 8
			| (unsigned)src8[step2] << 16
			| (unsigned)src8[step3] << 24
			);
		dst += 4;
		src8 += step4;
		--count;
	}
}

static __inline__ void internal_copy16_step(unsigned dst, void* src, unsigned count, int step1) {
	int step2 = step1 + step1;

	count /= 2;
	while (count) {
		_farnspokel(dst, P16DER0(src)
			| (unsigned)P16DER(src,step1) << 16);
		dst += 4;
		PADD(src,step2);
		--count;
	}
}

/***************************************************************************/
/* internal_unchained */

#if defined( __pentiumpro ) || defined( __pentium ) || defined( __k6 )

static __inline__ void internal_unchained8(unsigned dst, void* src, unsigned count) {
	unsigned rest = count % 4;

	__asm__ __volatile__ (
		"shr $2,%2\n"
		"jz 1f\n"
		"0:\n"
		"movb 12(%0),%%al\n"
		"movb 8(%0),%%ah\n"
		"bswap %%eax\n"
		"movb (%0),%%al\n"
		"movb 4(%0),%%ah\n"
		"movl %%eax,%%fs:(%1)\n"
		"addl $16,%0\n"
		"addl $4,%1\n"
		"decl %2\n"
		"jnz 0b\n"
		"1:\n"
		: "+r" (src), "+r" (dst), "+r" (count)
		:
		: "cc", "eax"
	);

	switch (rest) {
		case 3 : _farnspokeb(dst + 2, ((uint8*)src)[8] );
		case 2 : _farnspokeb(dst + 1, ((uint8*)src)[4] );
		case 1 : _farnspokeb(dst, ((uint8*)src)[0] );
	}
}

static __inline__ void internal_unchained8_step2(unsigned dst, void* src, unsigned count) {
	__asm__ __volatile__ (
		"shr $2,%2\n"
		"jz 1f\n"
		"0:\n"
		"movb 24(%0),%%al\n"
		"movb 16(%0),%%ah\n"
		"bswap %%eax\n"
		"movb (%0),%%al\n"
		"movb 8(%0),%%ah\n"
		"movl %%eax,%%fs:(%1)\n"
		"addl $32,%0\n"
		"addl $4,%1\n"
		"decl %2\n"
		"jnz 0b\n"
		"1:\n"
		: "+r" (src), "+r" (dst), "+r" (count)
		:
		: "cc", "eax"
	);
}

static __inline__ void internal_unchained8_double(unsigned dst, void* src, unsigned count) {
	__asm__ __volatile__ (
		"shr $2,%2\n"
		"jz 1f\n"
		"0:\n"
		"movb 6(%0),%%al\n"
		"movb 4(%0),%%ah\n"
		"bswap %%eax\n"
		"movb (%0),%%al\n"
		"movb 2(%0),%%ah\n"
		"movl %%eax,%%fs:(%1)\n"
		"addl $8,%0\n"
		"addl $4,%1\n"
		"decl %2\n"
		"jnz 0b\n"
		"1:\n"
		: "+r" (src), "+r" (dst), "+r" (count)
		:
		: "cc", "eax"
	);
}

#else

static __inline__ void internal_unchained8(unsigned dst, void* src, unsigned count) {
	uint8* src8 = (uint8*)src;
	unsigned rest = count % 4;

	count /= 4;
	while (count) {
		_farnspokel(dst,
			(unsigned)src8[0]
			| (unsigned)src8[4] << 8
			| (unsigned)src8[8] << 16
			| (unsigned)src8[12] << 24);
		dst += 4;
		src8 += 16;
		--count;
	}

	switch (rest) {
		case 3 : _farnspokeb(dst + 2, src8[8] );
		case 2 : _farnspokeb(dst + 1, src8[4] );
		case 1 : _farnspokeb(dst, src8[0] );
	}
}

static __inline__ void internal_unchained8_step2(unsigned dst, void* src, unsigned count) {
	uint8* src8 = (uint8*)src;

	count /= 4;
	while (count) {
		_farnspokel(dst,
			(unsigned)src8[0]
			| (unsigned)src8[8] << 8
			| (unsigned)src8[16] << 16
			| (unsigned)src8[24] << 24);
		dst += 4;
		src8 += 32;
		--count;
	}
}

static __inline__ void internal_unchained8_double(unsigned dst, void* src, unsigned count) {
	uint8* src8 = (uint8*)src;

	count /= 4;
	while (count) {
		_farnspokel(dst,
			(unsigned)src8[0]
			| (unsigned)src8[2] << 8
			| (unsigned)src8[4] << 16
			| (unsigned)src8[6] << 24);
		dst += 4;
		src8 += 8;
		--count;
	}
}

#endif

static __inline__ void internal_unchained8_step(unsigned dst, void* src, unsigned count, int step) {
	uint8* src8 = (uint8*)src;
	int step1 = 4 * step; /* 4 is for unchained mode */
	int step2 = step1 + step1;
	int step3 = step2 + step1;
	int step4 = step3 + step1;

	count /= 4;
	while (count) {
		_farnspokel(dst,
			(unsigned)src8[0]
			| (unsigned)src8[step1] << 8
			| (unsigned)src8[step2] << 16
			| (unsigned)src8[step3] << 24);
		dst += 4;
		src8 += step4;
		--count;
	}
}

static __inline__ void internal_unchained8_double_step2(unsigned dst, void* src, unsigned count) {
	internal_unchained8(dst,src,count);
}

static __inline__ void internal_unchained8_double_step(unsigned dst, void* src, unsigned count, int step) {
	uint8* src8 = (uint8*)src;
	int step1 = 2 * step; /* 2 is for unchained mode with double */
	int step2 = step1 + step1;
	int step3 = step2 + step1;
	int step4 = step3 + step1;

	count /= 4;
	while (count) {
		_farnspokel(dst,
			(unsigned)src8[0]
			| (unsigned)src8[step1] << 8
			| (unsigned)src8[step2] << 16
			| (unsigned)src8[step3] << 24);
		dst += 4;
		src8 += step4;
		--count;
	}
}

static __inline__ void internal_unchained8_palette16to8(unsigned dst, void* src, unsigned count, unsigned* palette) {
	uint16* src16 = (uint16*)src;
	unsigned rest = count % 4;

	count /= 4;
	while (count) {
		_farnspokel(dst,
			palette[src16[0]]
			| palette[src16[4]] << 8
			| palette[src16[8]] << 16
			| palette[src16[12]] << 24);
		dst += 4;
		src16 += 16;
		--count;
	}

	switch (rest) {
		case 3 : _farnspokeb(dst + 2, palette[src16[8]]);
		case 2 : _farnspokeb(dst + 1, palette[src16[4]]);
		case 1 : _farnspokeb(dst, palette[src16[0]]);
	}
}

static __inline__ void internal_unchained8_double_palette16to8(unsigned dst, void* src, unsigned count, unsigned* palette) {
	uint16* src16 = (uint16*)src;

	count /= 4;
	while (count) {
		uint32 mask;
		mask = palette[src16[0]]
			| palette[src16[2]] << 8
			| palette[src16[4]] << 16
			| palette[src16[6]] << 24;
		_farnspokel(dst, mask);
		dst += 4;
		src16 += 8;
		--count;
	}
}

/***************************************************************************/
/* internal_mean */

/* Notes:
1) The mean_mask_bit_h variable is the pixel value with the high bit of
	every rgb nibble at 1, duplicated to fill the 32 bit and negated.
2) The mean_mask_bit_l variable is the pixel value with the low bit of
	every rgb nibble at 1 and duplicated to fill the 32 bit
*/

static uint32 mean_mask_bit_h[1];
static uint32 mean_mask_bit_l[1];

static __inline__ void internal_mean_set(void) {

	video_rgb rgb_h = video_rgb_high_bit_get();
	video_rgb rgb_l = video_rgb_low_bit_get();
	video_rgb rgb_m = video_rgb_mask_bit_get();

	assert( video_index() == VIDEO_FLAGS_INDEX_RGB );

	switch (video_bytes_per_pixel()) {
		case 1 :
			mean_mask_bit_h[0] = rgb_h | rgb_h << 8 | rgb_h << 16 | rgb_h << 24;
			mean_mask_bit_l[0] = rgb_l | rgb_l << 8 | rgb_l << 16 | rgb_l << 24;
			rgb_m =	rgb_m | rgb_m << 8 | rgb_m << 16 | rgb_m << 24;
			break;
		case 2 :
			mean_mask_bit_h[0] = rgb_h | rgb_h << 16;
			mean_mask_bit_l[0] = rgb_l | rgb_l << 16;
			rgb_m = rgb_m | rgb_m << 16;
			break;
		case 4 :
			mean_mask_bit_h[0] = rgb_h;
			mean_mask_bit_l[0] = rgb_l;
			break;
	}

	mean_mask_bit_h[0] = ~mean_mask_bit_h[0] & rgb_m;
}

static __inline__ uint32 internal_mean_value(uint32 v0, uint32 v1) {
	/* compute the mean of many pixel and of many rgb nibble at one time */
	return ((v0 >> 1) & mean_mask_bit_h[0]) + ((v1 >> 1) & mean_mask_bit_h[0]) + (v0 & v1 & mean_mask_bit_l[0]);
}

#if 0
static __inline__ void internal_mean32_vert_self(void* dst, void* src, unsigned count)
{
	assert(((unsigned)src & 0x7)==0 && ((unsigned)dst & 0x7)==0);
	__asm__ __volatile__(
		"shrl $1,%2\n"
		"jz 1f\n"
		"andq mean_mask_bit_h,%%mm4\n"
		"andq mean_mask_bit_l,%%mm4\n"
		"0:\n"
		"movq (%0),%%mm0\n"
		"movq (%1),%%mm1\n"
		"movq %mm0,%%mm2\n"
		"movq %mm1,%%mm3\n"
		"shrq $1,%%mm0\n"
		"shrq $1,%%mm1\n"
		"andq %%mm4,%%mm0\n"
		"andq %%mm4,%%mm1\n"
		"addq %%mm1,%%mm0\n"
		"andq %%mm3,%%mm2\n"
		"andq %%mm5,%%mm2\n"
		"addq %%mm2,%%mm0\n"
		"decl %2\n"
		"jnz 0b\n"
		"1:\n"
		: "+r" (src), "+r" (dst), "+r" (count)
		:
		: "cc"
	);
}
#endif

static __inline__ void internal_mean8_vert_self(void* dst, void* src, unsigned count) {
	uint32* src32 = (uint32*)src;
	uint32* dst32 = (uint32*)dst;

	count /= 4;
	while (count) {
		dst32[0] = internal_mean_value(dst32[0],src32[0]);
		++src32;
		++dst32;
		--count;
	}
}

static __inline__ void internal_mean16_vert_self(void* dst, void* src, unsigned count) {
	uint32* src32 = (uint32*)src;
	uint32* dst32 = (uint32*)dst;

	count /= 2;
	while (count) {
		dst32[0] = internal_mean_value(dst32[0],src32[0]);
		++src32;
		++dst32;
		--count;
	}
}

static __inline__ void internal_mean8_vert_self_step(void* dst, void* src, unsigned count, int step1) {
	uint8* src8 = (uint8*)src;
	uint8* dst8 = (uint8*)dst;

	while (count) {
		dst8[0] = internal_mean_value(dst8[0],src8[0]);
		src8 += step1;
		++dst8;
		--count;
	}
}

static __inline__ void internal_mean16_vert_self_step(void* dst, void* src, unsigned count, int step1) {
	uint16* src16 = (uint16*)src;
	uint16* dst16 = (uint16*)dst;

	while (count) {
		dst16[0] = internal_mean_value(dst16[0],src16[0]);
		PADD(src16,step1);
		++dst16;
		--count;
	}
}

static __inline__ void internal_mean8_vert(unsigned dst, void* src0, void* src1, unsigned count) {
	uint32* src032 = (uint32*)src0;
	uint32* src132 = (uint32*)src1;

	count /= 4;
	while (count) {
		_farnspokel(dst, internal_mean_value(src032[0],src132[0]));
		dst += 4;
		++src032;
		++src132;
		--count;
	}
}

static __inline__ void internal_mean16_vert(unsigned dst, void* src0, void* src1, unsigned count) {
	uint32* src032 = (uint32*)src0;
	uint32* src132 = (uint32*)src1;

	count /= 2;
	while (count) {
		_farnspokel(dst, internal_mean_value(src032[0],src132[0]));
		dst += 4;
		++src032;
		++src132;
		--count;
	}
}

static __inline__ void internal_mean8_step1(unsigned dst, void* src, unsigned count) {
	count /= 4;
	if (count) {
		uint32* src32 = (uint32*)src;
		--count;
		while (count) {
			_farnspokel(dst, internal_mean_value(src32[0],(src32[0] >> 8) | (src32[1] << 24)));
			dst += 4;
			++src32;
			--count;
		}
		_farnspokel(dst, internal_mean_value(src32[0],(src32[0] >> 8) | (src32[0] & 0xFF000000)));
	}
}

static __inline__ void internal_mean8_step(unsigned dst, void* src, unsigned count, int step) {
	uint8* src8 = (uint8*)src;
	if (count) {
		--count;
		while (count) {
			_farnspokeb(dst, internal_mean_value(src8[0],src8[1]) );
			dst += 1;
			src8 += step;
			--count;
		}
		_farnspokeb(dst, src8[0]);
	}
}

static __inline__ void internal_mean16_step2(unsigned dst, void* src, unsigned count) {
	count /= 2;
	if (count) {
		uint32* src32 = (uint32*)src;
		--count;
		while (count) {
			_farnspokel(dst, internal_mean_value(src32[0],(src32[0] >> 16) | (src32[1] << 16)));
			dst += 4;
			++src32;
			--count;
		}
		_farnspokel(dst, internal_mean_value(src32[0],(src32[0] >> 16) | (src32[0] & 0xFFFF0000)));
	}
}

static __inline__ void internal_mean16_step(unsigned dst, void* src, unsigned count, int step) {
	uint16* src16 = (uint16*)src;
	if (count) {
		--count;
		while (count) {
			_farnspokew(dst, internal_mean_value(src16[0],src16[1]));
			dst += 2;
			PADD(src16,step);
			--count;
		}
		_farnspokew(dst, src16[0]);
	}
}

/***************************************************************************/
/* internal_max_rgb */

static unsigned max_rgb_mask_0;
static unsigned max_rgb_mask_1;
static unsigned max_rgb_mask_2;
static int max_rgb_shift_1;
static int max_rgb_shift_2;

static __inline__ void internal_max_rgb_set(void) {
	unsigned mask_0;
	unsigned mask_1;
	unsigned mask_2;
	unsigned shift_0;
	unsigned shift_1;
	unsigned shift_2;

	assert( video_index() == VIDEO_FLAGS_INDEX_RGB );

	if (video_state.rgb_red_shift >= video_state.rgb_green_shift &&
		video_state.rgb_red_shift >= video_state.rgb_blue_shift) {
		mask_0 = video_state.rgb_red_mask;
		mask_1 = video_state.rgb_green_mask;
		mask_2 = video_state.rgb_blue_mask;
		shift_0 = video_state.rgb_red_shift;
		shift_1 = video_state.rgb_green_shift;
		shift_2 = video_state.rgb_blue_shift;
	} else if (video_state.rgb_green_shift >= video_state.rgb_red_shift &&
		video_state.rgb_green_shift >= video_state.rgb_blue_shift) {
		mask_1 = video_state.rgb_red_mask;
		mask_0 = video_state.rgb_green_mask;
		mask_2 = video_state.rgb_blue_mask;
		shift_1 = video_state.rgb_red_shift;
		shift_0 = video_state.rgb_green_shift;
		shift_2 = video_state.rgb_blue_shift;
	} else {
		mask_2 = video_state.rgb_red_mask;
		mask_1 = video_state.rgb_green_mask;
		mask_0 = video_state.rgb_blue_mask;
		shift_2 = video_state.rgb_red_shift;
		shift_1 = video_state.rgb_green_shift;
		shift_0 = video_state.rgb_blue_shift;
	}

	max_rgb_mask_0 = mask_0;
	max_rgb_mask_1 = mask_1;
	max_rgb_mask_2 = mask_2;
	max_rgb_shift_1 = shift_0 - shift_1;
	max_rgb_shift_2 = shift_0 - shift_2;

	assert( max_rgb_shift_1 >= 0 &&	max_rgb_shift_2 >= 0);
}

static __inline__ unsigned internal_max_rgb_value(unsigned v) {
	return (v & max_rgb_mask_0)
		+ ((v << max_rgb_shift_1) & max_rgb_mask_1)
		+ ((v << max_rgb_shift_2) & max_rgb_mask_2);
}

static __inline__ void internal_max_rgb8_vert_self(void* dst, void* src, unsigned count) {
	uint8* src8 = (uint8*)src;
	uint8* dst8 = (uint8*)dst;

	while (count) {
		if (internal_max_rgb_value(dst8[0]) < internal_max_rgb_value(src8[0]))
			dst8[0] = src8[0];
		++src8;
		++dst8;
		--count;
	}
}

static __inline__ void internal_max_rgb8_vert_self_step(void* dst, void* src, unsigned count, int step1) {
	uint8* src8 = (uint8*)src;
	uint8* dst8 = (uint8*)dst;

	while (count) {
		if (internal_max_rgb_value(dst8[0]) < internal_max_rgb_value(src8[0]))
			dst8[0] = src8[0];
		src8 += step1;
		++dst8;
		--count;
	}
}

static __inline__ void internal_max_rgb16_vert_self(void* dst, void* src, unsigned count) {
	uint16* src16 = (uint16*)src;
	uint16* dst16 = (uint16*)dst;

	while (count) {
		if (internal_max_rgb_value(dst16[0]) < internal_max_rgb_value(src16[0]))
			dst16[0] = src16[0];
		++src16;
		++dst16;
		--count;
	}
}

static __inline__ void internal_max_rgb16_vert_self_step(void* dst, void* src, unsigned count, int step1) {
	uint16* src16 = (uint16*)src;
	uint16* dst16 = (uint16*)dst;

	while (count) {
		if (internal_max_rgb_value(dst16[0]) < internal_max_rgb_value(src16[0]))
			dst16[0] = src16[0];
		PADD(src16,step1);
		++dst16;
		--count;
	}
}

/***************************************************************************/
/* internal_max */

static __inline__ void internal_max8_vert_self(void* dst, void* src, unsigned count) {
	uint8* src8 = (uint8*)src;
	uint8* dst8 = (uint8*)dst;

	while (count) {
		if (dst8[0] < src8[0])
			dst8[0] = src8[0];
		++src8;
		++dst8;
		--count;
	}
}

static __inline__ void internal_max8_vert_self_step(void* dst, void* src, unsigned count, int step1) {
	uint8* src8 = (uint8*)src;
	uint8* dst8 = (uint8*)dst;

	while (count) {
		if (dst8[0] < src8[0])
			dst8[0] = src8[0];
		src8 += step1;
		++dst8;
		--count;
	}
}

static __inline__ void internal_max16_vert_self(void* dst, void* src, unsigned count) {
	uint16* src16 = (uint16*)src;
	uint16* dst16 = (uint16*)dst;

	while (count) {
		if (dst16[0] < src16[0])
			dst16[0] = src16[0];
		++src16;
		++dst16;
		--count;
	}
}

static __inline__ void internal_max16_vert_self_step(void* dst, void* src, unsigned count, int step1) {
	uint16* src16 = (uint16*)src;
	uint16* dst16 = (uint16*)dst;

	while (count) {
		if (dst16[0] < src16[0])
			dst16[0] = src16[0];
		PADD(src16,step1);
		++dst16;
		--count;
	}
}

/***************************************************************************/
/* fast_buffer */

/* A very fast dynamic buffers allocations */

/* Max number of allocable buffers */
#define FAST_BUFFER_MAX 24

struct fast_buffer_struct {
	void* buffer;
	void* buffer_aligned;
	unsigned size;
};

struct fast_buffer_struct fast_buffer_map[FAST_BUFFER_MAX];
unsigned fast_buffer_mac;

static void video_buffer_resize(struct fast_buffer_struct* buffer, unsigned size) {
	free(buffer->buffer);
	buffer->size = (size + 0xFF) & ~0xFF;
	buffer->buffer = malloc( buffer->size + ALIGN );
	assert(buffer->buffer);
	buffer->buffer_aligned = (void*)(((unsigned)buffer->buffer + ALIGN_MASK) & ~ALIGN_MASK);
}

static __inline__ void* video_buffer_alloc(unsigned size) {
	assert(	fast_buffer_mac < FAST_BUFFER_MAX );
	if (!fast_buffer_map[fast_buffer_mac].buffer
		|| fast_buffer_map[fast_buffer_mac].size < size)
		video_buffer_resize(&fast_buffer_map[fast_buffer_mac], size);
	return fast_buffer_map[fast_buffer_mac++].buffer_aligned;
}

/* Buffers must be allocated and freed in exact reverse order */
static __inline__ void video_buffer_free(void* buffer) {
	assert(	fast_buffer_mac > 0 );
	--fast_buffer_mac;
	assert( fast_buffer_map[fast_buffer_mac].buffer_aligned == buffer);
}

static void video_buffer_init(void) {
	unsigned i;	
	fast_buffer_mac = 0;
	for(i=0;i<FAST_BUFFER_MAX;++i) {
		fast_buffer_map[i].buffer = 0;
	}
}

static void video_buffer_done(void) {
	unsigned i;
	assert(	fast_buffer_mac == 0 );
	for(i=0;i<FAST_BUFFER_MAX;++i) {
		free(fast_buffer_map[i].buffer);
	}
}

/***************************************************************************/
/* stage/pipeline */

void video_pipeline_init(struct video_pipeline_struct* pipeline) {
	pipeline->stage_mac = 0;
}

void video_pipeline_done(struct video_pipeline_struct* pipeline) {
	int i;
	for(i=pipeline->stage_mac-1;i>=0;--i) {
		struct video_stage_horz_struct* stage = &pipeline->stage_map[i];
		if (stage->buffer)
			video_buffer_free(stage->buffer);
	}
}

static __inline__ struct video_stage_horz_struct* video_pipeline_begin_mutable(struct video_pipeline_struct* pipeline) {
	return pipeline->stage_map;
}

static __inline__ struct video_stage_horz_struct* video_pipeline_end_mutable(struct video_pipeline_struct* pipeline) {
	return pipeline->stage_map + pipeline->stage_mac;
}

static __inline__ struct video_stage_vert_struct* video_pipeline_vert_mutable(struct video_pipeline_struct* pipeline) {
	return &pipeline->stage_vert;
}

static __inline__ struct video_stage_horz_struct* video_pipeline_insert(struct video_pipeline_struct* pipeline) {
	struct video_stage_horz_struct* stage = video_pipeline_end_mutable(pipeline);
	++pipeline->stage_mac;
	return stage;
}

static __inline__ struct video_stage_horz_struct* video_pipeline_substitute(struct video_pipeline_struct* pipeline, struct video_stage_horz_struct* begin, struct video_stage_horz_struct* end) {
	struct video_stage_horz_struct* last = video_pipeline_end_mutable(pipeline);
	pipeline->stage_mac += 1 - (end - begin);
	while (end != last) {
		++begin;
		memcpy(begin,end,sizeof(struct video_stage_horz_struct));
		++end;
	}
	return begin;
}

static __inline__ void video_pipeline_realize(struct video_pipeline_struct* pipeline) {
	struct video_stage_horz_struct* stage =	video_pipeline_begin_mutable(pipeline);
	struct video_stage_horz_struct* stage_blit = video_pipeline_end_mutable(pipeline) - 1;
	while (stage != stage_blit) {
		if (stage->buffer_size) {
			stage->sel = _my_ds();
			stage->buffer =	video_buffer_alloc(stage->buffer_size);
		} else {
			stage->sel = 0;
			stage->buffer =	0;
		}
		++stage;
	}
	stage_blit->sel = video_write_selector();
	stage_blit->buffer = 0;
}

static __inline__ void* video_pipeline_run_partial(const struct video_stage_horz_struct* stage_begin, const struct video_stage_horz_struct* stage_end, void* src) {
	if (stage_begin == stage_end) {
		return src;
	} else {
		stage_begin->put(stage_begin, (unsigned)stage_begin->buffer, src);
		++stage_begin;

		while (stage_begin != stage_end) {
			stage_begin->put(stage_begin, (unsigned)stage_begin->buffer, stage_begin[-1].buffer);
			++stage_begin;
		}

		return stage_begin[-1].buffer;
	}
}

static __inline__ void video_pipeline_run(const struct video_stage_horz_struct* stage_begin, const struct video_stage_horz_struct* stage_end, unsigned dst, void* src) {
	--stage_end;
	if (stage_begin == stage_end) {
		stage_begin->put(stage_begin, dst, src);
	} else {
		stage_begin->put(stage_begin, (unsigned)stage_begin->buffer, src);
		++stage_begin;

		while (stage_begin != stage_end) {
			stage_begin->put(stage_begin, (unsigned)stage_begin->buffer, stage_begin[-1].buffer);
			++stage_begin;
		}

		stage_begin->put(stage_begin, dst, stage_begin[-1].buffer);
	}
}

static __inline__ void video_pipeline_run_plain(const struct video_stage_horz_struct* stage_begin, const struct video_stage_horz_struct* stage_end, unsigned dst, void* src) {
	--stage_end;
	if (stage_begin == stage_end) {
		stage_begin->put_plain(stage_begin, dst, src);
	} else {
		stage_begin->put_plain(stage_begin, (unsigned)stage_begin->buffer, src);
		++stage_begin;

		while (stage_begin != stage_end) {
			stage_begin->put(stage_begin, (unsigned)stage_begin->buffer, stage_begin[-1].buffer);
			++stage_begin;
		}

		stage_begin->put(stage_begin, dst, stage_begin[-1].buffer);
	}
}

static __inline__ void video_pipeline_vert_run(const struct video_pipeline_struct* pipeline, unsigned x, unsigned y, void* src) {
	video_pipeline_vert(pipeline)->put(video_pipeline_vert(pipeline),x,y,src);
	internal_copy_end();
}

const char* pipe_name(enum video_stage_enum pipe) {
	switch (pipe) {
		case pipe_x_stretch : return "horizontal stretch";
		case pipe_x_double : return "horizontal double";
		case pipe_x_triple : return "horizontal triple";
		case pipe_x_quadruple : return "horizontal quadruple";
		case pipe_x_filter : return "horizontal copy with low pass filter effect";
		case pipe_x_copy : return "horizontal copy";
		case pipe_unchained : return "unchained horizontal copy";
		case pipe_unchained_palette16to8 : return "unchained horizontal copy palette 16 to 8";
		case pipe_unchained_x_double : return "unchained horizontal double";
		case pipe_unchained_x_double_palette16to8  : return "unchained horizontal double palette 16 to 8";
		case pipe_palette8to16 : return "palette 8 to 16";
		case pipe_palette16to8 : return "palette 16 to 8";
		case pipe_palette16to16 : return "palette 16 to 16";
		case pipe_y_copy : return "vertical copy";
		case pipe_y_reduction_copy : return "vertical reduction";
		case pipe_y_expansion_copy : return "vertical expansion";
		case pipe_y_mean : return "vertical copy with mean effect";
		case pipe_y_reduction_mean : return "vertical reduction with mean effect";
		case pipe_y_expansion_mean : return "vertical expansion with mean effect";
		case pipe_y_filter : return "vertical copy with low pass filter effect";
		case pipe_y_reduction_filter : return "vertical reduction with low pass filter effect";
		case pipe_y_expansion_filter : return "vertical expansion with low pass filter effect";
		case pipe_y_reduction_max  : return "vertical reduction with max effect";
	}
	return 0;
}

static __inline__ int pipe_is_palette(enum video_stage_enum pipe) {
	switch (pipe) {
		case pipe_palette8to16 :
		case pipe_palette16to8 :
		case pipe_palette16to16 :
		case pipe_unchained_palette16to8 :
		case pipe_unchained_x_double_palette16to8 :
			return 1;
		default:
			return 0;
	}
}

/****************************************************************************/
/* filter8 */

static void video_line_filter8_step1(const struct video_stage_horz_struct* stage, unsigned dst, void* src) {
	_farsetsel(stage->sel);
	internal_mean8_step1(dst,src,stage->count);
}

static void video_line_filter8_step(const struct video_stage_horz_struct* stage, unsigned dst, void* src) {
	_farsetsel(stage->sel);
	internal_mean8_step(dst,src,stage->count,stage->sdp);
}

static void video_stage_filter8_set(struct video_stage_horz_struct* stage, unsigned sdx, int sdp) {
	stage->type = pipe_x_filter;
	stage->sdx = sdx;
	stage->sbpp = 1;
	stage->count = sdx;
	stage->sdp = sdp;
	stage->buffer_size = sdx;
	stage->plane_put = 0;
	stage->plane_put_plain = 0;

	stage->put_plain = video_line_filter8_step1;
	if (sdp == 1)
		stage->put = video_line_filter8_step1;
	else
		stage->put = video_line_filter8_step;

	internal_mean_set();
}

/****************************************************************************/
/* filter16 */

static void video_line_filter16_step2(const struct video_stage_horz_struct* stage, unsigned dst, void* src) {
	_farsetsel(stage->sel);
	internal_mean16_step2(dst,src,stage->count);
}

static void video_line_filter16_step(const struct video_stage_horz_struct* stage, unsigned dst, void* src) {
	_farsetsel(stage->sel);
	internal_mean16_step(dst,src,stage->count,stage->sdp);
}

static void video_stage_filter16_set(struct video_stage_horz_struct* stage, unsigned sdx, int sdp) {
	stage->type = pipe_x_filter;
	stage->sdx = sdx;
	stage->sbpp = 2;
	stage->count = sdx;
	stage->sdp = sdp;
	stage->buffer_size = 2*sdx;
	stage->plane_put = 0;
	stage->plane_put_plain = 0;

	stage->put_plain = video_line_filter16_step2;
	if (sdp == 2)
		stage->put = video_line_filter16_step2;
	else
		stage->put = video_line_filter16_step;

	internal_mean_set();
}

/****************************************************************************/
/* unchained8 */

static void video_line_unchained8_step1(const struct video_stage_horz_struct* stage, unsigned dst, void* src) {
	unsigned inner_count = stage->count / 4; /* 4 is the number of plane iterations */
	uint8* src8 = (uint8*)src;
	unsigned p;

	_farsetsel(stage->sel);

	for(p=0;p<4;++p) {
		video_unchained_plane_set(p);
		internal_unchained8(dst, src8 + p, inner_count);
	}
}

static void video_line_unchained8_step2(const struct video_stage_horz_struct* stage, unsigned dst, void* src) {
	unsigned inner_count = stage->count / 4; /* 4 is the number of plane iterations */
	uint8* src8 = (uint8*)src;
	unsigned p;

	_farsetsel(stage->sel);

	for(p=0;p<4;++p) {
		video_unchained_plane_set(p);
		internal_unchained8_step2(dst, src8 + 2*p, inner_count);
	}
}

static void video_line_unchained8(const struct video_stage_horz_struct* stage, unsigned dst, void* src) {
	unsigned inner_count = stage->count / 4; /* 4 is the number of plane iterations */
	uint8* src8 = (uint8*)src;
	unsigned p;

	_farsetsel(stage->sel);

	for(p=0;p<4;++p) {
		video_unchained_plane_set(p);
		internal_unchained8_step(dst, src8 + stage->sdp*p, inner_count, stage->sdp);
	}
}

static void video_line_unchained8_step1_plane(const struct video_stage_horz_struct* stage, unsigned dst, void* src) {
	_farsetsel(stage->sel);
	internal_unchained8(dst, src, stage->count / 4);
}

static void video_line_unchained8_step2_plane(const struct video_stage_horz_struct* stage, unsigned dst, void* src) {
	_farsetsel(stage->sel);
	internal_unchained8_step2(dst, src, stage->count / 4);
}

static void video_line_unchained8_plane(const struct video_stage_horz_struct* stage, unsigned dst, void* src) {
	_farsetsel(stage->sel);
	internal_unchained8_step(dst, src, stage->count / 4, stage->sdp);
}

static void video_stage_unchained8_set(struct video_stage_horz_struct* stage, unsigned sdx, int sdp) {
	stage->type = pipe_unchained;
	stage->sdx = sdx;
	stage->sbpp = 1;
	stage->count = sdx;
	stage->sdp = sdp;
	stage->buffer_size = 0;
	stage->plane_num = 4;

	stage->put_plain = video_line_unchained8_step1;
	stage->plane_put_plain = video_line_unchained8_step1_plane;
	if (sdp == 1) {
		stage->put = video_line_unchained8_step1;
		stage->plane_put = video_line_unchained8_step1_plane;
	} else if (sdp == 2) {
		stage->put = video_line_unchained8_step2;
		stage->plane_put = video_line_unchained8_step2_plane;
	} else {
		stage->put = video_line_unchained8;
		stage->plane_put = video_line_unchained8_plane;
	}
}

/****************************************************************************/
/* unchained8_palette16to8 */

static void video_line_unchained8_palette16to8_step1(const struct video_stage_horz_struct* stage, unsigned dst, void* src) {
	unsigned inner_count = stage->count / 4; /* 4 is the number of plane iteration */
	uint16* src16 = (uint16*)src;
	unsigned p;

	_farsetsel(stage->sel);

	for(p=0;p<4;++p) {
		video_unchained_plane_set(p);
		internal_unchained8_palette16to8(dst, src16 + p, inner_count, stage->palette);
	}
}

static void video_line_unchained8_palette16to8_step1_plane(const struct video_stage_horz_struct* stage, unsigned dst, void* src) {
	_farsetsel(stage->sel);
	internal_unchained8_palette16to8(dst, src, stage->count / 4, stage->palette);
}

static void video_stage_unchained8_palette16to8_set(struct video_stage_horz_struct* stage, unsigned sdx, unsigned* palette) {
	stage->type = pipe_unchained_palette16to8;
	stage->sdx = sdx;
	stage->sbpp = 2;
	stage->count = sdx;
	stage->sdp = 2;
	stage->palette = palette;
	stage->buffer_size = 0;
	stage->plane_num = 4;

	stage->put_plain = video_line_unchained8_palette16to8_step1;
	stage->plane_put_plain = video_line_unchained8_palette16to8_step1_plane;
	stage->put = video_line_unchained8_palette16to8_step1;
	stage->plane_put = video_line_unchained8_palette16to8_step1_plane;
}

/****************************************************************************/
/* unchained8_double */

static void video_line_unchained8_double_step1(const struct video_stage_horz_struct* stage, unsigned dst, void* src) {
	unsigned inner_count = stage->count / 2; /* 2 is the number of plane iterations */
	uint8* src8 = (uint8*)src;

	_farsetsel(stage->sel);

	video_unchained_plane_mask_set(0x3);
	internal_unchained8_double(dst, src8, inner_count);

	video_unchained_plane_mask_set(0xC);
	internal_unchained8_double(dst, src8 + 1, inner_count);
}

static void video_line_unchained8_double_step2(const struct video_stage_horz_struct* stage, unsigned dst, void* src) {
	unsigned inner_count = stage->count / 2; /* 2 is the number of plane iterations */
	uint8* src8 = (uint8*)src;

	_farsetsel(stage->sel);

	video_unchained_plane_mask_set(0x3);
	internal_unchained8_double_step2(dst, src8, inner_count);

	video_unchained_plane_mask_set(0xC);
	internal_unchained8_double_step2(dst, src8 + 2, inner_count);
}

static void video_line_unchained8_double(const struct video_stage_horz_struct* stage, unsigned dst, void* src) {
	unsigned inner_count = stage->count / 2; /* 2 is the number of plane iterations */
	uint8* src8 = (uint8*)src;

	_farsetsel(stage->sel);

	video_unchained_plane_mask_set(0x3);
	internal_unchained8_double_step(dst, src8, inner_count, stage->sdp);

	video_unchained_plane_mask_set(0xC);
	internal_unchained8_double_step(dst, src8 + stage->sdp, inner_count, stage->sdp);
}

static void video_stage_unchained8_double_set(struct video_stage_horz_struct* stage, unsigned sdx, int sdp) {
	stage->type = pipe_unchained_x_double;
	stage->sdx = sdx;
	stage->sbpp = 1;
	stage->count = sdx;
	stage->sdp = sdp;
	stage->buffer_size = 0;
	stage->plane_put = 0;
	stage->plane_put_plain = 0;

	stage->put_plain = video_line_unchained8_double_step1;
	if (stage->sdp == 1)
		stage->put = video_line_unchained8_double_step1;
	else if (stage->sdp == 2)
		stage->put = video_line_unchained8_double_step2;
	else
		stage->put = video_line_unchained8_double;
}

/****************************************************************************/
/* unchained8_double_palette16to8 */

static void video_line_unchained8_double_palette16to8_step1(const struct video_stage_horz_struct* stage, unsigned dst, void* src) {
	unsigned inner_count = stage->count / 2; /* 2 is the number of plane iterations */
	uint16* src16 = (uint16*)src;

	_farsetsel(stage->sel);

	video_unchained_plane_mask_set(0x3);
	internal_unchained8_double_palette16to8(dst, src16, inner_count,stage->palette);

	video_unchained_plane_mask_set(0xC);
	internal_unchained8_double_palette16to8(dst, src16 + 1, inner_count,stage->palette);
}

static void video_stage_unchained8_double_palette16to8_set(struct video_stage_horz_struct* stage, unsigned sdx, unsigned* palette) {
	stage->type = pipe_unchained_x_double_palette16to8;
	stage->sdx = sdx;
	stage->sbpp = 2;
	stage->count = sdx;
	stage->sdp = 2;
	stage->palette = palette;
	stage->buffer_size = 0;
	stage->plane_put = 0;
	stage->plane_put_plain = 0;

	stage->put_plain = video_line_unchained8_double_palette16to8_step1;
	stage->put = video_line_unchained8_double_palette16to8_step1;
}

/****************************************************************************/
/* stretchx8 */

static uint32 mask8_set_all[256];
static uint32 mask8_set_one[1024];

#define mask8_set_one1 (mask8_set_one+256)
#define mask8_set_one2 (mask8_set_one+512)
#define mask8_set_one3 (mask8_set_one+768)
#define mask8_set_one4 (mask8_set_one+1024)

#define VIDEO8_LINE_WRITE1 \
	mask |= mask_in[color]; \
	mask_in += 256; \
	if (mask_in == mask8_set_one4) { \
		_farnspokel(dst, mask); \
		mask_in = mask8_set_one; \
		mask = 0; \
		dst += 4; \
	}

#define VIDEO8_LINE_WRITE2 \
	if (mask_in == mask8_set_one3) { \
		mask |= mask_in[color]; \
		_farnspokel(dst, mask); \
		mask_in = mask8_set_one1; \
		mask = color; \
		dst += 4; \
	} else { \
		mask |= mask_in[color]; \
		mask_in += 256; \
		mask |= mask_in[color]; \
		mask_in += 256; \
		if (mask_in == mask8_set_one4) { \
			_farnspokel(dst, mask); \
			mask_in = mask8_set_one; \
			mask = 0; \
			dst += 4; \
		} \
	}

#define VIDEO8_LINE_WRITE3 \
	if (mask_in == mask8_set_one3) { \
		mask |= mask_in[color]; \
		_farnspokel(dst, mask); \
		mask_in = mask8_set_one2; \
		mask = color | color << 8; \
		dst += 4; \
	} else if (mask_in == mask8_set_one2) { \
		mask |= mask_in[color]; \
		mask_in += 256; \
		mask |= mask_in[color]; \
		_farnspokel(dst, mask); \
		mask_in = mask8_set_one1; \
		mask = color; \
		dst += 4; \
	} else if (mask_in == mask8_set_one1) { \
		mask |= mask8_set_all[color] & 0xFFFFFF00; \
		_farnspokel(dst, mask); \
		mask_in = mask8_set_one; \
		mask = 0; \
		dst += 4; \
	} else { \
		mask = mask8_set_all[color] & 0x00FFFFFF; \
		mask_in = mask8_set_one3; \
	}

static __inline__ void video_line_stretchx8_xx_step(const struct video_stage_horz_struct* stage, unsigned dst, uint8* src, unsigned sdp) {
	int error = -stage->up;
	unsigned inner_count = stage->count;
	uint32 mask = 0;
	uint32* mask_in = mask8_set_one;

	_farsetsel(stage->sel);

	/* assert( stage->whole >= 3 ); */

	while (inner_count) {
		unsigned color;
		unsigned run = stage->whole;
		if ((error += stage->up) > 0) {
			++run;
			error -= stage->down;
		}
		color = *src;

		if (mask_in == mask8_set_one3) {
			mask |= mask_in[color];
			_farnspokel(dst, mask);
			dst += 4;
			--run;
		} else if (mask_in == mask8_set_one2) {
			mask |= mask_in[color];
			mask_in += 256;
			mask |= mask_in[color];
			_farnspokel(dst, mask);
			dst += 4;
			run -= 2;
		} else if (mask_in == mask8_set_one1) {
			mask |= mask8_set_all[color] & 0xFFFFFF00;
			_farnspokel(dst, mask);
			dst += 4;
			run -= 3;
		}
		while (run>=4) {
			mask = mask8_set_all[color];
			_farnspokel(dst, mask);
			dst += 4;
			run -= 4;
		}
		if (run==1) {
			mask = color;
			mask_in = mask8_set_one1;
		} else if (run==2) {
			mask = color | color << 8;
			mask_in = mask8_set_one2;
		} else if (run==3) {
			mask = mask8_set_all[color] & 0x00FFFFFF;
			mask_in = mask8_set_one3;
		} else {
			mask_in = mask8_set_one;
		}
		PADD(src,sdp);
		--inner_count;
	}
}

static void video_line_stretchx8_xx(const struct video_stage_horz_struct* stage, unsigned dst, uint8* src) {
	video_line_stretchx8_xx_step(stage, dst, src, stage->sdp);
}

static void video_line_stretchx8_xx_step1(const struct video_stage_horz_struct* stage, unsigned dst, uint8* src) {
	video_line_stretchx8_xx_step(stage, dst, src, 1);
}

static __inline__ void video_line_stretchx8_01_step(const struct video_stage_horz_struct* stage, unsigned dst, uint8* src, unsigned sdp) {
	int error = -stage->up;
	unsigned inner_count = stage->count;
	uint32 mask = 0;
	uint32* mask_in = mask8_set_one;

	_farsetsel(stage->sel);

	while (inner_count) {
		unsigned color;
		unsigned run = stage->whole;
		if ((error += stage->up) > 0) {
			++run;
			error -= stage->down;
		}
		color = *src;
		while (run) {
			VIDEO8_LINE_WRITE1
			--run;
		}
		PADD(src,sdp);
		--inner_count;
	}
}

static void video_line_stretchx8_01(const struct video_stage_horz_struct* stage, unsigned dst, uint8* src) {
	video_line_stretchx8_01_step(stage, dst, src, stage->sdp);
}

static void video_line_stretchx8_01_step1(const struct video_stage_horz_struct* stage, unsigned dst, uint8* src) {
	video_line_stretchx8_01_step(stage, dst, src, 1);
}

static void video_line_stretchx8_11_step1(const struct video_stage_horz_struct* stage, unsigned dst, uint8* src) {
	_farsetsel(stage->sel);
	internal_copy8(dst,src,stage->count);
}

static void video_line_stretchx8_11(const struct video_stage_horz_struct* stage, unsigned dst, uint8* src) {
	_farsetsel(stage->sel);
	internal_copy8_step(dst,src,stage->count,stage->sdp);
}

static __inline__ void video_line_stretchx8_12_step(const struct video_stage_horz_struct* stage, unsigned dst, uint8* src, unsigned sdp) {
	int error = -stage->up;
	unsigned inner_count = stage->count;
	uint32 mask = 0;
	uint32* mask_in = mask8_set_one;

	_farsetsel(stage->sel);

	while (inner_count) {
		unsigned color = *src;
		if ((error += stage->up) > 0) {
			error -= stage->down;
			VIDEO8_LINE_WRITE2
		} else {
			VIDEO8_LINE_WRITE1
		}
		PADD(src,sdp);
		--inner_count;
	}
}

static void video_line_stretchx8_12(const struct video_stage_horz_struct* stage, unsigned dst, uint8* src) {
	video_line_stretchx8_12_step(stage, dst, src, stage->sdp);
}

static void video_line_stretchx8_12_step1(const struct video_stage_horz_struct* stage, unsigned dst, uint8* src) {
	video_line_stretchx8_12_step(stage, dst, src, 1);
}

static void video_line_stretchx8_22_step1(const struct video_stage_horz_struct* stage, unsigned dst, uint8* src) {
	_farsetsel(stage->sel);
	internal_double8(dst,src,stage->count);
}

static void video_line_stretchx8_22(const struct video_stage_horz_struct* stage, unsigned dst, uint8* src) {
	unsigned inner_count = stage->count / 2;

	_farsetsel(stage->sel);

 	while (inner_count) {
		unsigned color0,color1;
		uint32 mask;
		color0 = *src;
		PADD(src,stage->sdp);
		color1 = *src;
		PADD(src,stage->sdp);
		mask = color0 | color0 << 8 | color1 << 16 | color1 << 24;
		_farnspokel(dst, mask);
		dst += 4;
		--inner_count;
	}
}

static __inline__ void video_line_stretchx8_23_step(const struct video_stage_horz_struct* stage, unsigned dst, uint8* src, unsigned sdp) {
	int error = -stage->up;
	unsigned inner_count = stage->count;
	uint32 mask = 0;
	uint32* mask_in = mask8_set_one;

	_farsetsel(stage->sel);

	while (inner_count) {
		unsigned color = *src;
		if ((error += stage->up) > 0) {
			error -= stage->down;
			VIDEO8_LINE_WRITE3
		} else {
			VIDEO8_LINE_WRITE2
		}
		PADD(src,sdp);
		--inner_count;
	}
}

static void video_line_stretchx8_23(const struct video_stage_horz_struct* stage, unsigned dst, uint8* src) {
	video_line_stretchx8_23_step(stage, dst, src, stage->sdp);
}

static void video_line_stretchx8_23_step1(const struct video_stage_horz_struct* stage, unsigned dst, uint8* src) {
	video_line_stretchx8_23_step(stage, dst, src, 1);
}

static void video_line_stretchx8_33_step1(const struct video_stage_horz_struct* stage, unsigned dst, uint8* src) {
	unsigned inner_count = stage->count / 4;

	_farsetsel(stage->sel);

 	while (inner_count) {
		uint32 mask;
		unsigned color0 = src[0];
		unsigned color1 = src[1];
		unsigned color2 = src[2];
		unsigned color3 = src[3];
		mask = color0 | color0 << 8 | color0 << 16 | color1 << 24;
		_farnspokel(dst, mask);
		dst += 4;
		mask = color1 | color1 << 8 | color2 << 16 | color2 << 24;
		_farnspokel(dst, mask);
		dst += 4;
		mask = color2 | color3 << 8 | color3 << 16 | color3 << 24;
		_farnspokel(dst, mask);
		dst += 4;
		src += 4;
		--inner_count;
	}
}

static void video_line_stretchx8_33(const struct video_stage_horz_struct* stage, unsigned dst, uint8* src) {
	unsigned inner_count = stage->count / 4;

	_farsetsel(stage->sel);

 	while (inner_count) {
		uint32 mask;
		unsigned color0,color1,color2,color3;
		color0 = *src;
		PADD(src,stage->sdp);
		color1 = *src;
		PADD(src,stage->sdp);
		color2 = *src;
		PADD(src,stage->sdp);
		color3 = *src;
		PADD(src,stage->sdp);
		mask = color0 | color0 << 8 | color0 << 16 | color1 << 24;
		_farnspokel(dst, mask);
		dst += 4;
		mask = color1 | color1 << 8 | color2 << 16 | color2 << 24;
		_farnspokel(dst, mask);
		dst += 4;
		mask = color2 | color3 << 8 | color3 << 16 | color3 << 24;
		_farnspokel(dst, mask);
		dst += 4;
		--inner_count;
	}
}

static void video_line_stretchx8_44_step1(const struct video_stage_horz_struct* stage, unsigned dst, uint8* src) {
	unsigned inner_count = stage->count;

	_farsetsel(stage->sel);

	while (inner_count) {
		uint32 mask = mask8_set_all[*src];
		_farnspokel(dst, mask);
		dst += 4;
		++src;
		--inner_count;
	}
}

static void video_line_stretchx8_44(const struct video_stage_horz_struct* stage, unsigned dst, uint8* src) {
	unsigned inner_count = stage->count;

	_farsetsel(stage->sel);

	while (inner_count) {
		uint32 mask;
		mask = mask8_set_all[*src];
		_farnspokel(dst, mask);
		dst += 4;
		PADD(src,stage->sdp);
		--inner_count;
	}
}

/* Inizialize the line rescaling system */
static void video_stage_stretchx8_set(struct video_stage_horz_struct* stage, unsigned ddx, unsigned sdx, int sdp) {
	stage->type = pipe_x_stretch;
	stage->sdx = sdx;
	stage->sbpp = 1;
	stage->whole = ddx / sdx;
	stage->up = (ddx % sdx) * 2;
	stage->down = sdx * 2;
	stage->count = sdx;
	stage->sdp = sdp;
	stage->buffer_size = ddx;
	stage->plane_put = 0;
	stage->plane_put_plain = 0;

	if (stage->whole == 0) {
		stage->put_plain = (video_stage_hook*)video_line_stretchx8_01_step1;
		stage->put = (video_stage_hook*)video_line_stretchx8_01;
	} else if (stage->whole == 1 && stage->up == 0 && stage->sdp == 1) {
		stage->type = pipe_x_copy;
		stage->put_plain = (video_stage_hook*)video_line_stretchx8_11_step1;
		stage->put = (video_stage_hook*)video_line_stretchx8_11_step1;
	} else if (stage->whole == 1 && stage->up == 0) {
		stage->type = pipe_x_copy;
		stage->put_plain = (video_stage_hook*)video_line_stretchx8_11_step1;
		stage->put = (video_stage_hook*)video_line_stretchx8_11;
	} else if (stage->whole == 1) {
		stage->put_plain = (video_stage_hook*)video_line_stretchx8_12_step1;
		stage->put = (video_stage_hook*)video_line_stretchx8_12;
	} else if (stage->whole == 2 && stage->up == 0 && stage->sdp == 1) {
		stage->type = pipe_x_double;
		stage->put_plain = (video_stage_hook*)video_line_stretchx8_22_step1;
		stage->put = (video_stage_hook*)video_line_stretchx8_22_step1;
	} else if (stage->whole == 2 && stage->up == 0) {
		stage->type = pipe_x_double;
		stage->put_plain = (video_stage_hook*)video_line_stretchx8_22_step1;
		stage->put = (video_stage_hook*)video_line_stretchx8_22;
	} else if (stage->whole == 2) {
		stage->put_plain = (video_stage_hook*)video_line_stretchx8_23_step1;
		stage->put = (video_stage_hook*)video_line_stretchx8_23;
	} else if (stage->whole == 3 && stage->up == 0 && stage->sdp == 1) {
		stage->type = pipe_x_triple;
		stage->put_plain = (video_stage_hook*)video_line_stretchx8_33_step1;
		stage->put = (video_stage_hook*)video_line_stretchx8_33_step1;
	} else if (stage->whole == 3 && stage->up == 0) {
		stage->type = pipe_x_triple;
		stage->put_plain = (video_stage_hook*)video_line_stretchx8_33_step1;
		stage->put = (video_stage_hook*)video_line_stretchx8_33;
	} else if (stage->whole == 4 && stage->up == 0 && stage->sdp == 1) {
		stage->type = pipe_x_quadruple;
		stage->put_plain = (video_stage_hook*)video_line_stretchx8_44_step1;
		stage->put = (video_stage_hook*)video_line_stretchx8_44_step1;
	} else if (stage->whole == 4 && stage->up == 0) {
		stage->type = pipe_x_quadruple;
		stage->put_plain = (video_stage_hook*)video_line_stretchx8_44_step1;
		stage->put = (video_stage_hook*)video_line_stretchx8_44;
	} else {
		stage->put_plain = (video_stage_hook*)video_line_stretchx8_xx_step1;
		stage->put = (video_stage_hook*)video_line_stretchx8_xx;
	}
}

/****************************************************************************/
/* stretchx16 */

#define VIDEO16_LINE_WRITE1 \
	if (mask_in) { \
		mask |= color << 16; \
		_farnspokel(dst, mask); \
		--mask_in; \
		mask = 0; \
		dst += 4; \
	} else { \
		++mask_in; \
		mask = color; \
	}

#define VIDEO16_LINE_WRITE2 \
	if (mask_in) { \
		mask |= color << 16; \
		_farnspokel(dst, mask); \
		mask = color; \
		dst += 4; \
	} else { \
		mask = color | color << 16; \
		_farnspokel(dst, mask); \
		mask = 0; \
		dst += 4; \
	}

#define VIDEO16_LINE_WRITE3 \
	if (mask_in) { \
		mask |= color << 16; \
		_farnspokel(dst, mask); \
		dst += 4; \
		mask = color | color << 16; \
		_farnspokel(dst, mask); \
		--mask_in; \
		dst += 4; \
	} else { \
		mask = color | color << 16; \
		_farnspokel(dst, mask); \
		++mask_in; \
		mask = color; \
		dst += 4; \
	}

static __inline__ void video_line_stretchx16_xx_step(const struct video_stage_horz_struct* stage, unsigned dst, uint16* src, unsigned sdp) {
	int error = -stage->up;
	unsigned inner_count = stage->count;
	uint32 mask = 0;
	unsigned mask_in = 0;

	_farsetsel(stage->sel);

	while (inner_count) {
		unsigned color;
		unsigned run = stage->whole;
		if ((error += stage->up) > 0) {
			++run;
			error -= stage->down;
		}
		color = *src;
		if (mask_in) {
			mask |= color << 16;
			_farnspokel(dst, mask);
			--mask_in;
			dst += 4;
			--run;
		}
		while (run >= 2) {
			mask = color | color << 16;
			_farnspokel(dst, mask);
			dst += 4;
			run -= 2;
		}
		if (run) {
			mask = color;
			++mask_in;
		}
		PADD(src,sdp);
		--inner_count;
	}
}

static void video_line_stretchx16_xx(const struct video_stage_horz_struct* stage, unsigned dst, uint16* src) {
	video_line_stretchx16_xx_step(stage, dst, src, stage->sdp);
}

static void video_line_stretchx16_xx_step2(const struct video_stage_horz_struct* stage, unsigned dst, uint16* src) {
	video_line_stretchx16_xx_step(stage, dst, src, 2);
}

static __inline__ void video_line_stretchx16_01_step(const struct video_stage_horz_struct* stage, unsigned dst, uint16* src, unsigned sdp) {
	int error = -stage->up;
	unsigned inner_count = stage->count;
	uint32 mask = 0;
	unsigned mask_in = 0;

	_farsetsel(stage->sel);

	while (inner_count) {
		unsigned color;
		unsigned run = stage->whole;
		if ((error += stage->up) > 0) {
			++run;
			error -= stage->down;
		}
		color = *src;
		while (run) {
			VIDEO16_LINE_WRITE1
			--run;
		}
		PADD(src,sdp);
		--inner_count;
	}
}

static void video_line_stretchx16_01(const struct video_stage_horz_struct* stage, unsigned dst, uint16* src) {
	video_line_stretchx16_01_step(stage, dst, src, stage->sdp);
}

static void video_line_stretchx16_01_step2(const struct video_stage_horz_struct* stage, unsigned dst, uint16* src) {
	video_line_stretchx16_01_step(stage, dst, src, 2);
}

static void video_line_stretchx16_11_step2(const struct video_stage_horz_struct* stage, unsigned dst, uint16* src) {
	_farsetsel(stage->sel);
	internal_copy16(dst,src,stage->count);
}

static void video_line_stretchx16_11(const struct video_stage_horz_struct* stage, unsigned dst, uint16* src) {
	_farsetsel(stage->sel);
	internal_copy16_step(dst,src,stage->count,stage->sdp);
}

static void video_line_stretchx16_22_step2(const struct video_stage_horz_struct* stage, unsigned dst, uint16* src) {
	_farsetsel(stage->sel);
	internal_double16(dst,(unsigned)src,stage->count);
}

static void video_line_stretchx16_22(const struct video_stage_horz_struct* stage, unsigned dst, uint16* src) {
	unsigned inner_count = stage->count;

	_farsetsel(stage->sel);

	while (inner_count) {
		uint32 mask = *src;
		mask |= mask << 16;
		_farnspokel(dst, mask);
		dst += 4;
		PADD(src,stage->sdp);
		--inner_count;
	}
}

static void video_line_stretchx16_33_step2(const struct video_stage_horz_struct* stage, unsigned dst, uint16* src) {
	unsigned inner_count = stage->count / 2;

	_farsetsel(stage->sel);

	while (inner_count) {
		uint32 mask0 = src[0];
		uint32 mask2 = src[1];
		uint32 mask1 = mask0 | mask2 << 16;
		mask0 |= mask0 << 16;
		mask2 |= mask2 << 16;
		_farnspokel(dst, mask0);
		dst += 4;
		_farnspokel(dst, mask1);
		dst += 4;
		_farnspokel(dst, mask2);
		dst += 4;
		src += 2;
		--inner_count;
	}
}

static void video_line_stretchx16_33(const struct video_stage_horz_struct* stage, unsigned dst, uint16* src) {
	unsigned inner_count = stage->count / 2;

	_farsetsel(stage->sel);

	while (inner_count) {
		uint32 mask0,mask1,mask2;
		mask0 = *src;
		PADD(src,stage->sdp);
		mask2 = *src;
		PADD(src,stage->sdp);
		mask1 = mask0 | mask2 << 16;
		mask0 |= mask0 << 16;
		mask2 |= mask2 << 16;
		_farnspokel(dst, mask0);
		dst += 4;
		_farnspokel(dst, mask1);
		dst += 4;
		_farnspokel(dst, mask2);
		dst += 4;
		--inner_count;
	}
}

static void video_line_stretchx16_44_step2(const struct video_stage_horz_struct* stage, unsigned dst, uint16* src) {
	unsigned inner_count = stage->count;

	_farsetsel(stage->sel);

	while (inner_count) {
		uint32 mask = *src;
		mask |= mask << 16;
		_farnspokel(dst, mask);
		dst += 4;
		_farnspokel(dst, mask);
		dst += 4;
		++src;
		--inner_count;
	}
}

static void video_line_stretchx16_44(const struct video_stage_horz_struct* stage, unsigned dst, uint16* src) {
	unsigned inner_count = stage->count;

	_farsetsel(stage->sel);

	while (inner_count) {
		uint32 mask = *src;
		mask |= mask << 16;
		_farnspokel(dst, mask);
		dst += 4;
		_farnspokel(dst, mask);
		dst += 4;
		PADD(src,stage->sdp);
		--inner_count;
	}
}

static void video_stage_stretchx16_set(struct video_stage_horz_struct* stage, unsigned ddx, unsigned sdx, int sdp) {
	stage->type = pipe_x_stretch;
	stage->sdx = sdx;
	stage->sbpp = 2;
	stage->whole = ddx / sdx;
	stage->up = (ddx % sdx) * 2;
	stage->down = sdx * 2;
	stage->count = sdx;
	stage->sdp = sdp;
	stage->buffer_size = 2*ddx;
	stage->plane_put = 0;
	stage->plane_put_plain = 0;

	if (stage->whole == 0) {
		stage->put_plain = (video_stage_hook*)video_line_stretchx16_01_step2;
		stage->put = (video_stage_hook*)video_line_stretchx16_01;
	} else if (stage->whole == 1 && stage->up == 0 && stage->sdp == 2) {
		stage->type = pipe_x_copy;
		stage->put_plain = (video_stage_hook*)video_line_stretchx16_11_step2;
		stage->put = (video_stage_hook*)video_line_stretchx16_11_step2;
	} else if (stage->whole == 1 && stage->up == 0) {
		stage->type = pipe_x_copy;
		stage->put_plain = (video_stage_hook*)video_line_stretchx16_11_step2;
		stage->put = (video_stage_hook*)video_line_stretchx16_11;
	} else if (stage->whole == 2 && stage->up == 0 && stage->sdp == 2) {
		stage->type = pipe_x_double;
		stage->put_plain = (video_stage_hook*)video_line_stretchx16_22_step2;
		stage->put = (video_stage_hook*)video_line_stretchx16_22_step2;
	} else if (stage->whole == 2 && stage->up == 0) {
		stage->type = pipe_x_double;
		stage->put_plain = (video_stage_hook*)video_line_stretchx16_22_step2;
		stage->put = (video_stage_hook*)video_line_stretchx16_22;
	} else if (stage->whole == 3 && stage->up == 0 && stage->sdp == 2) {
		stage->type = pipe_x_triple;
		stage->put_plain = (video_stage_hook*)video_line_stretchx16_33_step2;
		stage->put = (video_stage_hook*)video_line_stretchx16_33_step2;
	} else if (stage->whole == 3 && stage->up == 0) {
		stage->type = pipe_x_triple;
		stage->put_plain = (video_stage_hook*)video_line_stretchx16_33_step2;
		stage->put = (video_stage_hook*)video_line_stretchx16_33;
	} else if (stage->whole == 4 && stage->up == 0 && stage->sdp == 2) {
		stage->type = pipe_x_quadruple;
		stage->put_plain = (video_stage_hook*)video_line_stretchx16_44_step2;
		stage->put = (video_stage_hook*)video_line_stretchx16_44_step2;
	} else if (stage->whole == 4 && stage->up == 0) {
		stage->type = pipe_x_quadruple;
		stage->put_plain = (video_stage_hook*)video_line_stretchx16_44_step2;
		stage->put = (video_stage_hook*)video_line_stretchx16_44;
	} else {
		stage->put_plain = (video_stage_hook*)video_line_stretchx16_xx_step2;
		stage->put = (video_stage_hook*)video_line_stretchx16_xx;
	}
}

/****************************************************************************/
/* chained8 */

static void video_stage_chained8_set(struct video_stage_horz_struct* stage, unsigned sdx, int sdp) {
	stage->type = pipe_x_copy;
	stage->sdx = sdx;
	stage->sbpp = 1;
	stage->count = sdx;
	stage->sdp = sdp;
	stage->buffer_size = sdx;
	stage->plane_put = 0;
	stage->plane_put_plain = 0;

	stage->put_plain = (video_stage_hook*)video_line_stretchx8_11_step1;
	if (sdp == 1)
		stage->put = (video_stage_hook*)video_line_stretchx8_11_step1;
	else
		stage->put = (video_stage_hook*)video_line_stretchx8_11;
}

/****************************************************************************/
/* chained16 */

static void video_stage_chained16_set(struct video_stage_horz_struct* stage, unsigned sdx, int sdp) {
	stage->type = pipe_x_copy;
	stage->sdx = sdx;
	stage->sbpp = 2;
	stage->count = sdx;
	stage->sdp = sdp;
	stage->buffer_size = 2*sdx;
	stage->plane_put = 0;
	stage->plane_put_plain = 0;

	stage->put_plain = (video_stage_hook*)video_line_stretchx16_11_step2;
	if (sdp == 2)
		stage->put = (video_stage_hook*)video_line_stretchx16_11_step2;
	else
		stage->put = (video_stage_hook*)video_line_stretchx16_11;
}

/****************************************************************************/
/* 8palette16 */

static void video_line_8palette16_step2(const struct video_stage_horz_struct* stage, unsigned dst, void* src) {
	unsigned inner_count = stage->count / 4;
	unsigned* palette = stage->palette;
	uint16* src16 = (uint16*)src;

	_farsetsel(stage->sel);

	while (inner_count) {
		uint32 mask;
		mask = palette[src16[0]]
			| palette[src16[1]] << 8
			| palette[src16[2]] << 16
			| palette[src16[3]] << 24;
		_farnspokel(dst, mask);
		dst += 4;
		src16 += 4;
		--inner_count;
	}
}

static void video_line_8palette16(const struct video_stage_horz_struct* stage, unsigned dst, void* src) {
	unsigned inner_count = stage->count / 4;
	int line_index1 = stage->sdp;
	int line_index2 = line_index1 + line_index1;
	int line_index3 = line_index2 + line_index1;
	int line_index4 = line_index3 + line_index1;
	unsigned* palette = stage->palette;

	_farsetsel(stage->sel);

	while (inner_count) {
		uint32 mask;
		mask = palette[P16DER0(src)]
			| palette[P16DER(src,line_index1)] << 8
			| palette[P16DER(src,line_index2)] << 16
			| palette[P16DER(src,line_index3)] << 24;
		_farnspokel(dst, mask);
		dst += 4;
		PADD(src,line_index4);
		--inner_count;
	}
}

static void video_stage_8palette16_set(struct video_stage_horz_struct* stage, unsigned sdx, int sdp, unsigned* palette) {
	stage->type = pipe_palette16to8;
	stage->sdx = sdx;
	stage->sbpp = 2;
	stage->count = sdx;
	stage->sdp = sdp;
	stage->palette = palette;
	stage->buffer_size = sdx;
	stage->plane_put = 0;
	stage->plane_put_plain = 0;

	stage->put_plain = video_line_8palette16_step2;
	if (sdp == 2)
		stage->put = video_line_8palette16_step2;
	else
		stage->put = video_line_8palette16;
}

/****************************************************************************/
/* 16palette8 */

static void video_line_16palette8_step1(const struct video_stage_horz_struct* stage, unsigned dst, void* src) {
	unsigned inner_count = stage->count / 2;
	unsigned* palette = stage->palette;
	uint8* src8 = (uint8*)src;

	_farsetsel(stage->sel);

	while (inner_count) {
		uint32 mask;
		mask = palette[src8[0]]
			| palette[src8[1]] << 16;
		_farnspokel(dst, mask);
		dst += 4;
		src8 += 2;
		--inner_count;
	}
}

static void video_line_16palette8(const struct video_stage_horz_struct* stage, unsigned dst, void* src) {
	unsigned inner_count = stage->count / 2;
	int line_index1 = stage->sdp;
	int line_index2 = line_index1 + line_index1;
	unsigned* palette = stage->palette;
	uint8* src8 = (uint8*)src;

	_farsetsel(stage->sel);

	while (inner_count) {
		uint32 mask;
		mask = palette[src8[0]]
			| palette[src8[line_index1]] << 16;
		_farnspokel(dst, mask);
		dst += 4;
		src8 += line_index2;
		--inner_count;
	}
}

static void video_stage_16palette8_set(struct video_stage_horz_struct* stage, unsigned sdx, int sdp, int unsigned* palette) {
	stage->type = pipe_palette8to16;
	stage->sdx = sdx;
	stage->sbpp = 1;
	stage->count = sdx;
	stage->sdp = sdp;
	stage->palette = palette;
	stage->buffer_size = 2*sdx;
	stage->plane_put = 0;
	stage->plane_put_plain = 0;

	stage->put_plain = video_line_16palette8_step1;
	if (sdp == 1)
		stage->put = video_line_16palette8_step1;
	else
		stage->put = video_line_16palette8;
}

/****************************************************************************/
/* 16palette16 */

static void video_line_16palette16_step2(const struct video_stage_horz_struct* stage, unsigned dst, void* src) {
	unsigned inner_count = stage->count / 2;
	unsigned* palette = stage->palette;
	uint16* src16 = (uint16*)src;

	_farsetsel(stage->sel);

	while (inner_count) {
		uint32 mask;
		mask = palette[src16[0]]
			| palette[src16[1]] << 16;
		_farnspokel(dst, mask);
		dst += 4;
		src16 += 2;
		--inner_count;
	}
}

static void video_line_16palette16(const struct video_stage_horz_struct* stage, unsigned dst, void* src) {
	unsigned inner_count = stage->count / 2;
	int line_index1 = stage->sdp;
	int line_index2 = line_index1 + line_index1;
	unsigned* palette = stage->palette;

	_farsetsel(stage->sel);

	while (inner_count) {
		uint32 mask;
		mask = palette[P16DER0(src)]
			| palette[P16DER(src,line_index1)] << 16;
		_farnspokel(dst, mask);
		dst += 4;
		PADD(src,line_index2);
		--inner_count;
	}
}

static void video_stage_16palette16_set(struct video_stage_horz_struct* stage, unsigned sdx, int sdp, unsigned* palette) {
	stage->type = pipe_palette16to16;
	stage->sdx = sdx;
	stage->sbpp = 2;
	stage->count = sdx;
	stage->sdp = sdp;
	stage->palette = palette;
	stage->buffer_size = 2*sdx;
	stage->plane_put = 0;
	stage->plane_put_plain = 0;

	stage->put_plain = video_line_16palette16_step2;
	if (sdp == 2)
		stage->put = video_line_16palette16_step2;
	else
		stage->put = video_line_16palette16;
}

/***************************************************************************/
/* stretchy reduction */

/* This example explains the difference of the effects in reduction

  ORI   COPY  MEAN  FILTER
  A     A     A+B+C A         (first line)
  B     D     D+E   C+D
  C     F     F+G+H E+F
  D     I     I+J   H+I
  ...

*/

static __inline__ void stage_copy(const struct video_stage_horz_struct* stage, void* dst, void* src) {
	_farsetsel(_my_ds());
	if (stage->sbpp == 1) {
		if (stage->sdp == 1)
			internal_copy8((unsigned)dst,src,stage->sdx);
		else
			internal_copy8_step((unsigned)dst,src,stage->sdx,stage->sdp);
	} else {
		if (stage->sdp == 2)
			internal_copy16((unsigned)dst,src,stage->sdx);
		else
			internal_copy16_step((unsigned)dst,src,stage->sdx,stage->sdp);
	}
}

static __inline__ void stage_max_vert_self(const struct video_stage_horz_struct* stage, void* dst, void* src) {
	if (stage->sbpp == 1) {
		if (stage->sdp == 1)
			internal_max8_vert_self(dst,src,stage->sdx);
		else
			internal_max8_vert_self_step(dst,src,stage->sdx,stage->sdp);
	} else {
		if (stage->sdp == 2)
			internal_max16_vert_self(dst,src,stage->sdx);
		else
			internal_max16_vert_self_step(dst,src,stage->sdx,stage->sdp);
	}
}

static __inline__ void stage_max_rgb_vert_self(const struct video_stage_horz_struct* stage, void* dst, void* src) {
	if (stage->sbpp == 1) {
		if (stage->sdp == 1)
			internal_max_rgb8_vert_self(dst,src,stage->sdx);
		else
			internal_max_rgb8_vert_self_step(dst,src,stage->sdx,stage->sdp);
	} else {
		if (stage->sdp == 2)
			internal_max_rgb16_vert_self(dst,src,stage->sdx);
		else
			internal_max_rgb16_vert_self_step(dst,src,stage->sdx,stage->sdp);
	}
}

static __inline__ void stage_mean_vert_self(const struct video_stage_horz_struct* stage, void* dst, void* src) {
	if (stage->sbpp == 1) {
		if (stage->sdp == 1)
			internal_mean8_vert_self(dst,src,stage->sdx);
		else
			internal_mean8_vert_self_step(dst,src,stage->sdx,stage->sdp);
	} else {
		if (stage->sdp == 2)
			internal_mean16_vert_self(dst,src,stage->sdx);
		else
			internal_mean16_vert_self_step(dst,src,stage->sdx,stage->sdp);
	}
}

static void video_stage_stretchy_x1(const struct video_stage_vert_struct* stage_vert, unsigned x, unsigned y, void* src) {
	unsigned x_off = video_offset(x);
	unsigned inner_count = stage_vert->ddy;

	int whole = stage_vert->whole;
	int up = stage_vert->up;
	int down = stage_vert->down;
	int error = -up;

	while (inner_count) {
		unsigned dst;
		unsigned run = whole;
		if ((error += up) > 0) {
			++run;
			error -= down;
		}

		dst = video_write_line(y) + x_off;
		video_pipeline_run(stage_vert->stage_begin, stage_vert->stage_end, dst, src);
		++y;

		PADD(src,stage_vert->sdw * run);
		--inner_count;
	}
}

static void video_stage_stretchy_max_x1(const struct video_stage_vert_struct* stage_vert, unsigned x, unsigned y, void* src) {
	unsigned x_off = video_offset(x);
	unsigned inner_count = stage_vert->ddy;

	const struct video_stage_horz_struct* stage_begin = stage_vert->stage_begin;
	const struct video_stage_horz_struct* stage_end = stage_vert->stage_end;
	const struct video_stage_horz_struct* stage_pivot = stage_vert->stage_pivot;

	void* buffer = video_buffer_alloc(stage_vert->stage_begin->sdx * stage_vert->stage_begin->sbpp);

	int whole = stage_vert->whole;
	int up = stage_vert->up;
	int down = stage_vert->down;
	int error = -up;

	while (inner_count) {
		unsigned dst;
		unsigned run = whole;
		if ((error += up) > 0) {
			++run;
			error -= down;
		}

		dst = video_write_line(y) + x_off;
		if (run == 1) {
			video_pipeline_run(stage_begin, stage_end, dst, src);
			PADD(src,stage_vert->sdw);
		} else {
			void* src_buffer = video_pipeline_run_partial(stage_begin, stage_pivot, src);
			stage_copy(stage_pivot,buffer,src_buffer);
			PADD(src,stage_vert->sdw);
			--run;

			while (run) {
				src_buffer = video_pipeline_run_partial(stage_begin, stage_pivot, src);
				if (video_index() == VIDEO_FLAGS_INDEX_RGB)
					stage_max_rgb_vert_self(stage_pivot,buffer,src_buffer);
				else
					stage_max_vert_self(stage_pivot,buffer,src_buffer);
				PADD(src,stage_vert->sdw);
				--run;
			}

			video_pipeline_run_plain(stage_vert->stage_pivot, stage_end, dst, buffer);
		}
		++y;
		--inner_count;
	}

	video_buffer_free(buffer);
}

/* Compute the mean of every lines reduced to a single line */
static void video_stage_stretchy_mean_x1(const struct video_stage_vert_struct* stage_vert, unsigned x, unsigned y, void* src) {
	unsigned x_off = video_offset(x);
	unsigned inner_count = stage_vert->ddy;

	const struct video_stage_horz_struct* stage_begin = stage_vert->stage_begin;
	const struct video_stage_horz_struct* stage_end = stage_vert->stage_end;
	const struct video_stage_horz_struct* stage_pivot = stage_vert->stage_pivot;

	void* buffer = video_buffer_alloc(stage_pivot->sdx * stage_pivot->sbpp);

	int whole = stage_vert->whole;
	int up = stage_vert->up;
	int down = stage_vert->down;
	int error = -up;

	while (inner_count) {
		unsigned dst;
		unsigned run = whole;
		if ((error += up) > 0) {
			++run;
			error -= down;
		}

		dst = video_write_line(y) + x_off;
		if (run == 1) {
			video_pipeline_run(stage_begin, stage_end, dst, src);
			PADD(src,stage_vert->sdw);
		} else {
			void* src_buffer = video_pipeline_run_partial(stage_begin, stage_pivot, src);
			stage_copy(stage_pivot,buffer,src_buffer);
			PADD(src,stage_vert->sdw);
			--run;

			while (run) {
				src_buffer = video_pipeline_run_partial(stage_begin, stage_pivot, src);
				stage_mean_vert_self(stage_pivot,buffer,src_buffer);
				PADD(src,stage_vert->sdw);
				--run;
			}

			video_pipeline_run_plain(stage_pivot, stage_end, dst, buffer);
		}
		++y;
		--inner_count;
	}

	video_buffer_free(buffer);
}

/* Compute the mean of the previous line and the first of every iteration */
static void video_stage_stretchy_filter_x1(const struct video_stage_vert_struct* stage_vert, unsigned x, unsigned y, void* src) {
	unsigned x_off = video_offset(x);
	unsigned inner_count = stage_vert->ddy;

	const struct video_stage_horz_struct* stage_begin = stage_vert->stage_begin;
	const struct video_stage_horz_struct* stage_end = stage_vert->stage_end;
	const struct video_stage_horz_struct* stage_pivot = stage_vert->stage_pivot;

	int buffer_full = 0;
	void* buffer = video_buffer_alloc(stage_pivot->sdx * stage_pivot->sbpp);

	int whole = stage_vert->whole;
	int up = stage_vert->up;
	int down = stage_vert->down;
	int error = -up;

	while (inner_count) {
		unsigned dst;
		void* src_buffer;
		unsigned run = whole;
		if ((error += up) > 0) {
			++run;
			error -= down;
		}

		dst = video_write_line(y) + x_off;
		src_buffer = video_pipeline_run_partial(stage_begin, stage_pivot, src);
		if (buffer_full) {
			stage_mean_vert_self(stage_pivot,buffer,src_buffer);
			video_pipeline_run_plain(stage_pivot, stage_end, dst, buffer);
		} else {
			video_pipeline_run(stage_pivot, stage_end, dst, src_buffer);
			buffer_full = 1;
		}

		if (run > 1) {
			PADD(src, (run - 1) * stage_vert->sdw);
			src_buffer = video_pipeline_run_partial(stage_begin, stage_pivot, src);
		}
		stage_copy(stage_pivot,buffer,src_buffer);
		PADD(src,stage_vert->sdw);

		++y;
		--inner_count;
	}

	video_buffer_free(buffer);
}

/***************************************************************************/
/* stretchy expansion */

/* This example explain the difference of effects in expansion

  ORI   COPY  MEAN  FILTER
  A     A     A     A         (first line)
  B     B     A+B   A+B
        B     B     B (== B+B)
  C     C     B+C   B+C
        C     C     C (== C+C)
  D     D     D     C+D
  E     E     D+E   D+E
        E     E     E (== E+E)
*/

static void video_stage_stretchy_1x(const struct video_stage_vert_struct* stage_vert, unsigned x, unsigned y, void* src) {
	unsigned x_off = video_offset(x);
	unsigned inner_count = stage_vert->sdy;

	int whole = stage_vert->whole;
	int up = stage_vert->up;
	int down = stage_vert->down;
	int error = -up;

	while (inner_count) {
		void* buffer;

		unsigned run = whole;
		if ((error += up) > 0) {
			++run;
			error -= down;
		}

		buffer = video_pipeline_run_partial(stage_vert->stage_begin, stage_vert->stage_pivot, src);
		while (run) {
			unsigned dst = video_write_line(y) + x_off;
			video_pipeline_run(stage_vert->stage_pivot, stage_vert->stage_end, dst, buffer);
			++y;
			--run;
		}

		PADD(src,stage_vert->sdw);
		--inner_count;
	}
}

/* The mean effect is applied at the first line duplicated, if no line */
/* duplication is done, no effect is applied */
static void video_stage_stretchy_mean_1x(const struct video_stage_vert_struct* stage_vert, unsigned x, unsigned y, void* src) {
	unsigned x_off = video_offset(x);
	unsigned inner_count = stage_vert->sdy;

	const struct video_stage_horz_struct* stage_begin = stage_vert->stage_begin;
	const struct video_stage_horz_struct* stage_end = stage_vert->stage_end;
	const struct video_stage_horz_struct* stage_pivot = stage_vert->stage_pivot;

	int buffer_full = 0;
	void* buffer = video_buffer_alloc(stage_pivot->sdx * stage_pivot->sbpp);

	int whole = stage_vert->whole;
	int up = stage_vert->up;
	int down = stage_vert->down;
	int error = -up;

	while (inner_count) {
		void* src_buffer;
		unsigned dst;
		unsigned run = whole;
		if ((error += up) > 0) {
			++run;
			error -= down;
		}
		src_buffer = video_pipeline_run_partial(stage_begin, stage_pivot, src);
		dst = video_write_line(y) + x_off;

		/* don't apply the effect if no duplication is required */
		if (buffer_full && run >= 2) {
			stage_mean_vert_self(stage_pivot,buffer,src_buffer);
			video_pipeline_run_plain(stage_pivot, stage_end, dst, buffer);
		} else {
			video_pipeline_run(stage_pivot, stage_end, dst, src_buffer);
			buffer_full = 1;
		}

		stage_copy(stage_pivot,buffer,src_buffer);

		++y;
		--run;

		while (run) {
			dst = video_write_line(y) + x_off;
			video_pipeline_run(stage_pivot, stage_end, dst, src_buffer);
			++y;
			--run;
		}

		PADD(src,stage_vert->sdw);
		--inner_count;
	}

	video_buffer_free(buffer);
}

/* The effect is applied at every line */
static void video_stage_stretchy_filter_1x(const struct video_stage_vert_struct* stage_vert, unsigned x, unsigned y, void* src) {
	unsigned x_off = video_offset(x);
	unsigned inner_count = stage_vert->sdy;

	const struct video_stage_horz_struct* stage_begin = stage_vert->stage_begin;
	const struct video_stage_horz_struct* stage_end = stage_vert->stage_end;
	const struct video_stage_horz_struct* stage_pivot = stage_vert->stage_pivot;

	int buffer_full = 0;
	void* buffer = video_buffer_alloc(stage_pivot->sdx * stage_pivot->sbpp);

	int whole = stage_vert->whole;
	int up = stage_vert->up;
	int down = stage_vert->down;
	int error = -up;

	while (inner_count) {
		void* src_buffer;
		unsigned dst;
		unsigned run = whole;
		if ((error += up) > 0) {
			++run;
			error -= down;
		}
		src_buffer = video_pipeline_run_partial(stage_begin, stage_pivot, src);
		dst = video_write_line(y) + x_off;

		/* apply the effect only at the first line because on others */
		/* lines the two source are equals and the mean operation */
		/* don't have any effects */
		if (buffer_full) {
			stage_mean_vert_self(stage_pivot,buffer,src_buffer);
			video_pipeline_run_plain(stage_pivot, stage_end, dst, buffer);
		} else {
			video_pipeline_run(stage_pivot, stage_end, dst, src_buffer);
			buffer_full = 1;
		}
		stage_copy(stage_pivot,buffer,src_buffer);
		++y;
		--run;

		while (run) {
			dst = video_write_line(y) + x_off;
			video_pipeline_run(stage_pivot, stage_end, dst, src_buffer);
			++y;
			--run;
		}

		PADD(src,stage_vert->sdw);
		--inner_count;
	}

	video_buffer_free(buffer);
}

/***************************************************************************/
/* stretchy copy */

static void video_stage_stretchy_11(const struct video_stage_vert_struct* stage_vert, unsigned x, unsigned y, void* src) {
	unsigned inner_count = stage_vert->sdy;
	unsigned x_off = video_offset(x);

	while (inner_count) {
		unsigned dst;

		dst = video_write_line(y) + x_off;
		video_pipeline_run(stage_vert->stage_begin, stage_vert->stage_end, dst, src);
		++y;

		PADD(src,stage_vert->sdw);
		--inner_count;
	}
}

/***************************************************************************/
/* stretchy plane */

/* Vertical wrapper for 4 plane mode unchained mode */
static void video_stage_planey4(const struct video_stage_vert_struct* stage_vert, unsigned x, unsigned y, void* src) {
	uint8* src8 = (uint8*)src;
	unsigned p;

	for(p=0;p<4;++p) {
		video_unchained_plane_set(p);
		stage_vert->plane_put(stage_vert,x,y,src8 + p * stage_vert->stage_begin->sdp);
	}
}

/* Vertical wrapper for 2 plane mode unchained */
static void video_stage_planey2(const struct video_stage_vert_struct* stage_vert, unsigned x, unsigned y, void* src) {
	uint8* src8 = (uint8*)src;

	video_unchained_plane_mask_set(0x3);
	stage_vert->plane_put(stage_vert,x,y,src8);
	video_unchained_plane_mask_set(0xC);
	stage_vert->plane_put(stage_vert,x,y,src8 + stage_vert->stage_begin->sdp);
}

/* Modify the vertical stage for supporting the separated plane put */
static __inline__ void video_stage_planey_set(struct video_stage_vert_struct* stage_vert) {
	/* supported only if the pipeline has only one stage and this stage */
	/* support the plane put */
	if (stage_vert->stage_begin + 1 == stage_vert->stage_end
		&& stage_vert->stage_begin->plane_put) {
		stage_vert->plane_put =	stage_vert->put;
		if (stage_vert->stage_begin->plane_num == 4) {
			stage_vert->put = video_stage_planey4;
		} else {
			stage_vert->put = video_stage_planey2;
		}
		((struct video_stage_horz_struct*)(stage_vert->stage_begin))->put = stage_vert->stage_begin->plane_put;
		((struct video_stage_horz_struct*)(stage_vert->stage_begin))->put_plain = stage_vert->stage_begin->plane_put_plain;
	}
}

/***************************************************************************/
/* stretchy */

/* Inizialize the vertical stage */
static void video_stage_stretchy_set(struct video_stage_vert_struct* stage_vert, const struct video_pipeline_struct* pipeline, unsigned ddy, unsigned sdy, int sdw, unsigned combine) {
	stage_vert->sdy = sdy;
	stage_vert->sdw = sdw;
	stage_vert->ddy = ddy;

	stage_vert->stage_begin = video_pipeline_begin(pipeline);
	stage_vert->stage_end = video_pipeline_end(pipeline);

	if (sdy < ddy) {
		stage_vert->whole = ddy / sdy;
		stage_vert->up = (ddy % sdy) * 2;
		stage_vert->down = sdy * 2;

		switch (combine) {
			case VIDEO_COMBINE_MEAN :
				stage_vert->stage_pivot = stage_vert->stage_end - 1;
				stage_vert->put = video_stage_stretchy_mean_1x;
				stage_vert->type = pipe_y_expansion_mean;
				break;
			case VIDEO_COMBINE_FILTER :
				stage_vert->stage_pivot = stage_vert->stage_end - 1;
				stage_vert->put = video_stage_stretchy_filter_1x;
				stage_vert->type = pipe_y_expansion_filter;
				break;
			default:
				stage_vert->stage_pivot = stage_vert->stage_end - 1;
				stage_vert->put = video_stage_stretchy_1x;
				stage_vert->type = pipe_y_expansion_copy;
				break;
		}
	} else if (sdy == ddy) {
		stage_vert->whole = 1;
		stage_vert->up = 0;
		stage_vert->down = sdy * 2;

		switch (combine) {
			case VIDEO_COMBINE_MEAN :
				stage_vert->stage_pivot = stage_vert->stage_end;
				while (stage_vert->stage_pivot != stage_vert->stage_begin
					&& !pipe_is_palette(stage_vert->stage_pivot[-1].type)
				)
					--stage_vert->stage_pivot;
				stage_vert->put = video_stage_stretchy_mean_1x;
				stage_vert->type = pipe_y_mean;
				break;
			case VIDEO_COMBINE_FILTER :
				stage_vert->stage_pivot = stage_vert->stage_end;
				while (stage_vert->stage_pivot != stage_vert->stage_begin
					&& !pipe_is_palette(stage_vert->stage_pivot[-1].type)
				)
					--stage_vert->stage_pivot;
				stage_vert->put = video_stage_stretchy_filter_1x;
				stage_vert->type = pipe_y_filter;
				break;
			default:
				stage_vert->stage_pivot = stage_vert->stage_begin;
				stage_vert->put = video_stage_stretchy_11;
				stage_vert->type = pipe_y_copy;
				break;
		}
	} else {
		stage_vert->whole = sdy / ddy;
		stage_vert->up = (sdy % ddy) * 2;
		stage_vert->down = ddy * 2;

		switch (combine) {
			case VIDEO_COMBINE_MAX :
				stage_vert->stage_pivot = stage_vert->stage_begin;
				stage_vert->put = video_stage_stretchy_max_x1;
				stage_vert->type = pipe_y_reduction_max;
				break;
			case VIDEO_COMBINE_MEAN :
				stage_vert->stage_pivot = stage_vert->stage_end;
				while (stage_vert->stage_pivot != stage_vert->stage_begin
					&& !pipe_is_palette(stage_vert->stage_pivot[-1].type)
				)
					--stage_vert->stage_pivot;
				stage_vert->put = video_stage_stretchy_mean_x1;
				stage_vert->type = pipe_y_reduction_mean;
				break;
			case VIDEO_COMBINE_FILTER :
				stage_vert->stage_pivot = stage_vert->stage_end;
				while (stage_vert->stage_pivot != stage_vert->stage_begin
					&& !pipe_is_palette(stage_vert->stage_pivot[-1].type)
				)
					--stage_vert->stage_pivot;
				stage_vert->put = video_stage_stretchy_filter_x1;
				stage_vert->type = pipe_y_reduction_filter;
				break;
			default:
				stage_vert->stage_pivot = stage_vert->stage_begin;
				stage_vert->put = video_stage_stretchy_x1;
				stage_vert->type = pipe_y_reduction_copy;
				break;
		}
	}

	/* try to activate the plane put if avaliable */
	if (combine == VIDEO_COMBINE_NONE)
		video_stage_planey_set(stage_vert);

	if (video_index() == VIDEO_FLAGS_INDEX_RGB ) {
		/* inizialize the mean tables */
		if (combine == VIDEO_COMBINE_MEAN || combine == VIDEO_COMBINE_FILTER)
			internal_mean_set();

		/* inizialize the max tables */
		if (combine == VIDEO_COMBINE_MAX)
			internal_max_rgb_set();
	}
}

/* Inizialize the vertical stage for the no transformation special case */
static __inline__ void video_stage_stretchy_11_set(struct video_stage_vert_struct* stage_vert, const struct video_pipeline_struct* pipeline, unsigned sdy, int sdw) {
	stage_vert->sdy = sdy;
	stage_vert->sdw = sdw;
	stage_vert->ddy = sdy;

	stage_vert->whole = 1;
	stage_vert->up = 0;
	stage_vert->down = sdy * 2;

	stage_vert->stage_begin = video_pipeline_begin(pipeline);
	stage_vert->stage_end = video_pipeline_end(pipeline);
	stage_vert->stage_pivot = stage_vert->stage_begin;

	stage_vert->put = video_stage_stretchy_11;
	stage_vert->type = pipe_y_copy;

	/* try to activate the plane put if avaliable */
	video_stage_planey_set(stage_vert);
}

/***************************************************************************/
/* inizialization */

/* Inizialize the stretch system */
void video_stretch_init() {
	unsigned i;
	for(i=0;i<256;++i) {
		mask8_set_one[i+0*256] = i;
		mask8_set_one[i+1*256] = i << 8;
		mask8_set_one[i+2*256] = i << 16;
		mask8_set_one[i+3*256] = i << 24;
		mask8_set_all[i] = i | i << 8 | i << 16 | i << 24;
	}
	video_buffer_init();
}

/* Deinizialize */
void video_stretch_done(void) {
	video_buffer_done();
}

/***************************************************************************/
/* blit */

void video_blit_pipeline_init(struct video_pipeline_struct* pipeline, unsigned src_dx, unsigned src_dy, int src_dw, int src_dp) {
	video_pipeline_init(pipeline);

	switch (video_bytes_per_pixel()) {
		case 1 :
			if (video_unchained_active())
				video_stage_unchained8_set( video_pipeline_insert(pipeline), src_dx, src_dp );
			else
				video_stage_chained8_set( video_pipeline_insert(pipeline), src_dx, src_dp );
			break;
		case 2 :
			video_stage_chained16_set( video_pipeline_insert(pipeline), src_dx, src_dp );
			break;
	}

	video_stage_stretchy_11_set(video_pipeline_vert_mutable(pipeline),pipeline,src_dy,src_dw);

	video_pipeline_realize(pipeline);
}

void video_blit_palette8_pipeline_init(struct video_pipeline_struct* pipeline, unsigned src_dx, unsigned src_dy, int src_dw, int src_dp, unsigned* palette) {
	video_pipeline_init(pipeline);

	switch (video_bytes_per_pixel()) {
		case 2 :
			video_stage_16palette8_set( video_pipeline_insert(pipeline), src_dx, src_dp, palette );
			break;
		default:
			assert(0);
	}

	video_stage_stretchy_11_set(video_pipeline_vert_mutable(pipeline),pipeline,src_dy,src_dw);

	video_pipeline_realize(pipeline);
}

void video_blit_palette16_pipeline_init(struct video_pipeline_struct* pipeline, unsigned src_dx, unsigned src_dy, int src_dw, int src_dp, unsigned* palette) {
	video_pipeline_init(pipeline);

	switch (video_bytes_per_pixel()) {
		case 1 :
			video_stage_8palette16_set( video_pipeline_insert(pipeline), src_dx, src_dp, palette );
			if (video_unchained_active())
				video_stage_unchained8_set( video_pipeline_insert(pipeline), src_dx, 1 );
			break;
		case 2 :
			video_stage_16palette16_set( video_pipeline_insert(pipeline), src_dx, src_dp, palette );
			break;
	}

	video_stage_stretchy_11_set(video_pipeline_vert_mutable(pipeline),pipeline,src_dy,src_dw);

	video_pipeline_realize(pipeline);
}

/***************************************************************************/
/* stretch */

/* require that the last stage isn't a palette stage */
#define LAST_NOT_PALETTE (combine == VIDEO_COMBINE_MEAN || combine == VIDEO_COMBINE_FILTER)

static __inline__ void video_pipeline_make(struct video_pipeline_struct* pipeline, unsigned dst_dx, unsigned src_dx, int src_dp, unsigned combine) {
	unsigned bytes_per_pixel = video_bytes_per_pixel();
	struct video_stage_horz_struct* end;

	if (combine == VIDEO_COMBINE_FILTER && src_dx > dst_dx) {
		switch (bytes_per_pixel) {
			case 1 : video_stage_filter8_set( video_pipeline_insert(pipeline), src_dx, src_dp ); break;
			case 2 : video_stage_filter16_set( video_pipeline_insert(pipeline), src_dx, src_dp ); break;
		}
		src_dp = bytes_per_pixel;
	}

	if (dst_dx != src_dx) {
		switch (bytes_per_pixel) {
			case 1 : video_stage_stretchx8_set( video_pipeline_insert(pipeline), dst_dx, src_dx, src_dp ); break;
			case 2 : video_stage_stretchx16_set( video_pipeline_insert(pipeline), dst_dx, src_dx, src_dp ); break;
		}
		src_dp = bytes_per_pixel;
	}

	if (combine == VIDEO_COMBINE_FILTER && src_dx <= dst_dx) {
		switch (bytes_per_pixel) {
			case 1 : video_stage_filter8_set( video_pipeline_insert(pipeline), dst_dx, src_dp ); break;
			case 2 : video_stage_filter16_set( video_pipeline_insert(pipeline), dst_dx, src_dp ); break;
		}
		src_dp = bytes_per_pixel;
	}

	if (bytes_per_pixel == 1 && video_unchained_active()) {
		video_stage_unchained8_set( video_pipeline_insert(pipeline), dst_dx, src_dp );
	} else {
		if ((video_pipeline_size(pipeline) == 0)
			|| (LAST_NOT_PALETTE && pipe_is_palette(video_pipeline_end(pipeline)[-1].type))) {
			switch (bytes_per_pixel) {
				case 1 : video_stage_chained8_set( video_pipeline_insert(pipeline), dst_dx, src_dp ); break;
				case 2 : video_stage_chained16_set( video_pipeline_insert(pipeline), dst_dx, src_dp ); break;
			}
			src_dp = bytes_per_pixel;
		}
	}

	/* optimize */
	end = video_pipeline_end_mutable(pipeline);

	if (!LAST_NOT_PALETTE) {
		if (video_pipeline_size(pipeline) >= 2
			&& end[-1].type == pipe_unchained
			&& end[-2].type == pipe_palette16to8 && end[-2].sdp == 2) {
			video_stage_unchained8_palette16to8_set( video_pipeline_substitute(pipeline,end-2,end), end[-2].sdx, end[-2].palette );
			return;
		}

		if (video_pipeline_size(pipeline) >= 3
			&& end[-1].type == pipe_unchained
			&& end[-2].type == pipe_x_double
			&& end[-3].type == pipe_palette16to8 && end[-3].sdp == 2) {
			video_stage_unchained8_double_palette16to8_set( video_pipeline_substitute(pipeline,end-3,end), end[-3].sdx, end[-3].palette );
			return;
		}
	}

	if (video_pipeline_size(pipeline) >= 2
		&& end[-1].type == pipe_unchained
		&& end[-2].type == pipe_x_double) {
		video_stage_unchained8_double_set( video_pipeline_substitute(pipeline,end-2,end), end[-2].sdx, end[-2].sdp);
		return;
	}
}

void video_stretch_pipeline_init(struct video_pipeline_struct* pipeline, unsigned dst_dx, unsigned dst_dy, unsigned src_dx, unsigned src_dy, int src_dw, int src_dp, unsigned combine) {
	video_pipeline_init(pipeline);

	video_pipeline_make(pipeline,dst_dx,src_dx,src_dp,combine);

	video_stage_stretchy_set(video_pipeline_vert_mutable(pipeline), pipeline, dst_dy, src_dy, src_dw, combine);

	video_pipeline_realize(pipeline);
}

void video_stretch_palette8_pipeline_init(struct video_pipeline_struct* pipeline, unsigned dst_dx, unsigned dst_dy, unsigned src_dx, unsigned src_dy, int src_dw, int src_dp, unsigned* palette, unsigned combine) {
	unsigned bytes_per_pixel = video_bytes_per_pixel();

	video_pipeline_init(pipeline);

	switch (bytes_per_pixel) {
		case 2 :
			video_stage_16palette8_set( video_pipeline_insert(pipeline), src_dx, src_dp, palette );
			break;
	}

	video_pipeline_make(pipeline,dst_dx,src_dx,bytes_per_pixel,combine);

	video_stage_stretchy_set(video_pipeline_vert_mutable(pipeline), pipeline, dst_dy, src_dy, src_dw, combine);

	video_pipeline_realize(pipeline);
}

void video_stretch_palette16_pipeline_init(struct video_pipeline_struct* pipeline, unsigned dst_dx, unsigned dst_dy, unsigned src_dx, unsigned src_dy, int src_dw, int src_dp, unsigned* palette, unsigned combine) {
	unsigned bytes_per_pixel = video_bytes_per_pixel();

	video_pipeline_init(pipeline);

	switch (bytes_per_pixel) {
		case 1 :
			video_stage_8palette16_set( video_pipeline_insert(pipeline), src_dx, src_dp, palette );
			break;
		case 2 :
			video_stage_16palette16_set( video_pipeline_insert(pipeline), src_dx, src_dp, palette );
			break;
	}

	video_pipeline_make(pipeline,dst_dx,src_dx,bytes_per_pixel,combine);

	video_stage_stretchy_set(video_pipeline_vert_mutable(pipeline), pipeline, dst_dy, src_dy, src_dw, combine);

	video_pipeline_realize(pipeline);

}

void video_blit_pipeline(const struct video_pipeline_struct* pipeline, unsigned dst_x, unsigned dst_y, void* src) {
	video_pipeline_vert_run(pipeline,dst_x,dst_y,src);
}
