/*
 * This file is part of the AdvanceMAME project
 *
 * Copyright (C) 1999-2000 Andrea Mazzoleni
 *
 * The contents of this file are free for non-commercial use.
 * Any commercial use is forbidden.
 */

#ifndef __SCRBLIT_H
#define	__SCRBLIT_H

#include "screen.h"

/***************************************************************************/
/* stage/pipeline */

struct video_stage_horz_struct;

typedef	void video_stage_hook(const struct video_stage_horz_struct* stage, unsigned dst, void* src);

enum video_stage_enum {
	pipe_x_stretch,
	pipe_x_double,
	pipe_x_triple,
	pipe_x_quadruple,
	pipe_x_filter,
	pipe_x_copy,
	pipe_unchained,
	pipe_unchained_palette16to8,
	pipe_unchained_x_double,
	pipe_unchained_x_double_palette16to8,
	pipe_palette8to16,
	pipe_palette16to8,
	pipe_palette16to16,
	pipe_y_copy,
	pipe_y_reduction_copy,
	pipe_y_expansion_copy,
	pipe_y_mean,
	pipe_y_reduction_mean,
	pipe_y_expansion_mean,
	pipe_y_filter,
	pipe_y_expansion_filter,
	pipe_y_reduction_filter,
	pipe_y_reduction_max
};

const char* pipe_name(enum video_stage_enum pipe_type);

/* Pipeline horizontal trasformation stage */
struct video_stage_horz_struct {
	/* hook */
	video_stage_hook* put;

	/* hook assuming sdp == sbpp */
	video_stage_hook* put_plain;

	/* type */
	enum video_stage_enum type;

	/* dest */
	unsigned sel; /* dest selector */
	void* buffer; /* eventually dest buffer */
	unsigned buffer_size; /* size of the dest buffer, ==0 if not required */

	/* source */
	int sdp; /* step in the src for the next pixel (in bytes) */
	unsigned sdx; /* size of the source (in bytes) */
	unsigned sbpp; /* bytes per pixel of the source */

	/* brehesnam x */
	int whole; /* segment length */
	int up; /* up step */
	int down; /* down step */
	int count; /* number of segment */

	/* palette conversion */
	unsigned* palette;

	/* unchained plane */
	unsigned plane_num; /* number of plane */
	video_stage_hook* plane_put; /* hook for single plane put, !=0 if plane put supported */
	video_stage_hook* plane_put_plain; /* hook assuming sdp == sbpp */
};

/* Special effect */
#define VIDEO_COMBINE_NONE 0 /* no effect */
#define VIDEO_COMBINE_MAX 1 /* use the max value in y reductions */
#define VIDEO_COMBINE_MEAN 2 /* use the mean value in y transformations for added or removed lines */
#define VIDEO_COMBINE_FILTER 3 /* apply a FIR lowpass filter with 2 point and fc 0.5 in x and y directions */

struct video_stage_vert_struct;

typedef void video_stage_vert_hook(const struct video_stage_vert_struct* stage_vert, unsigned x, unsigned y, void* src);

/* Pipeline vertical trasformation stage */
struct video_stage_vert_struct {
	/* hook */
	video_stage_vert_hook* put;

	/* type */
	enum video_stage_enum type;

	/* source */
	unsigned sdy;
	unsigned sdw;

	/* dest */
	unsigned ddy;

	/* brehesnam y */
	int whole;
	int up;
	int down;

	/* pipeline */
	const struct video_stage_horz_struct* stage_begin;
	const struct video_stage_horz_struct* stage_end;

	/* stage used for splitting the pipeline, on the split is applied the vertical stage */
	const struct video_stage_horz_struct* stage_pivot;

	/* unchained plane */
	video_stage_vert_hook* plane_put; /* hook for plane put, !=0 if plane put supported */
};

#define VIDEO_STAGE_MAX 8

struct video_pipeline_struct {
	struct video_stage_vert_struct stage_vert;
	struct video_stage_horz_struct stage_map[VIDEO_STAGE_MAX];
	unsigned stage_mac;
};

void video_pipeline_init(struct video_pipeline_struct* pipeline);
void video_pipeline_done(struct video_pipeline_struct* pipeline);

static __inline__ unsigned video_pipeline_size(const struct video_pipeline_struct* pipeline) {
	return pipeline->stage_mac;
}

static __inline__ const struct video_stage_horz_struct* video_pipeline_begin(const struct video_pipeline_struct* pipeline) {
	return pipeline->stage_map;
}

static __inline__ const struct video_stage_horz_struct* video_pipeline_end(const struct video_pipeline_struct* pipeline) {
	return pipeline->stage_map + pipeline->stage_mac;
}

static __inline__ const struct video_stage_vert_struct* video_pipeline_vert(const struct video_pipeline_struct* pipeline) {
	return &pipeline->stage_vert;
}

static __inline__ const struct video_stage_horz_struct* video_pipeline_pivot(const struct video_pipeline_struct* pipeline) {
	return video_pipeline_vert(pipeline)->stage_pivot;
}

/***************************************************************************/
/* blit */

void video_stretch_init(void);
void video_stretch_done(void);

/***************************************************************************/
/* pipeline blit */

void video_blit_pipeline_init(struct video_pipeline_struct* pipeline, unsigned src_dx, unsigned src_dy, int src_dw, int src_dp);
void video_blit_palette8_pipeline_init(struct video_pipeline_struct* pipeline, unsigned src_dx, unsigned src_dy, int src_dw, int src_dp, unsigned* palette);
void video_blit_palette16_pipeline_init(struct video_pipeline_struct* pipeline, unsigned src_dx, unsigned src_dy, int src_dw, int src_dp, unsigned* palette);

void video_stretch_pipeline_init(struct video_pipeline_struct* pipeline, unsigned dst_dx, unsigned dst_dy, unsigned src_dx, unsigned src_dy, int src_dw, int src_dp, unsigned combine);
void video_stretch_palette8_pipeline_init(struct video_pipeline_struct* pipeline, unsigned dst_dx, unsigned dst_dy, unsigned src_dx, unsigned src_dy, int src_dw, int src_dp, unsigned* palette, unsigned combine);
void video_stretch_palette16_pipeline_init(struct video_pipeline_struct* pipeline, unsigned dst_dx, unsigned dst_dy, unsigned src_dx, unsigned src_dy, int src_dw, int src_dp, unsigned* palette, unsigned combine);

void video_blit_pipeline(const struct video_pipeline_struct* pipeline, unsigned dst_x, unsigned dst_y, void* src);

/***************************************************************************/
/* blit */

void video_clear(unsigned dst_x, unsigned dst_y, unsigned dst_dx, unsigned dst_dy, unsigned src);

static __inline__ void video_blit(unsigned dst_x, unsigned dst_y, void* src, unsigned src_dx, unsigned src_dy, int src_dw, int src_dp) {
	struct video_pipeline_struct pipeline;

	video_blit_pipeline_init(&pipeline, src_dx, src_dy, src_dw, src_dp);

	video_blit_pipeline(&pipeline, dst_x, dst_y, src);

	video_pipeline_done(&pipeline);
}

static __inline__ void video_blit_palette8(unsigned dst_x, unsigned dst_y, void* src, unsigned src_dx, unsigned src_dy, int src_dw, int src_dp, unsigned* palette) {
	struct video_pipeline_struct pipeline;

	video_blit_palette8_pipeline_init(&pipeline, src_dx, src_dy, src_dw, src_dp, palette);

	video_blit_pipeline(&pipeline, dst_x, dst_y, src);

	video_pipeline_done(&pipeline);
}

static __inline__ void video_blit_palette16(unsigned dst_x, unsigned dst_y, void* src, unsigned src_dx, unsigned src_dy, int src_dw, int src_dp, unsigned* palette) {
	struct video_pipeline_struct pipeline;

	video_blit_palette16_pipeline_init(&pipeline, src_dx, src_dy, src_dw, src_dp, palette);

	video_blit_pipeline(&pipeline, dst_x, dst_y, src);

	video_pipeline_done(&pipeline);
}

static __inline__ void video_stretch(unsigned dst_x, unsigned dst_y, unsigned dst_dx, unsigned dst_dy, void* src, unsigned src_dx, unsigned src_dy, int src_dw, int src_dp, unsigned combine) {
	struct video_pipeline_struct pipeline;

	video_stretch_pipeline_init(&pipeline, dst_dx, dst_dy, src_dx, src_dy, src_dw, src_dp, combine);

	video_blit_pipeline(&pipeline, dst_x, dst_y, src);

	video_pipeline_done(&pipeline);
}

static __inline__ void video_stretch_palette8(unsigned dst_x, unsigned dst_y, unsigned dst_dx, unsigned dst_dy, uint8* src, unsigned src_dx, unsigned src_dy, int src_dw, int src_dp, unsigned* palette, unsigned combine) {
	struct video_pipeline_struct pipeline;

	video_stretch_palette8_pipeline_init(&pipeline, dst_dx, dst_dy, src_dx, src_dy, src_dw, src_dp, palette, combine);

	video_blit_pipeline(&pipeline, dst_x, dst_y, src);

	video_pipeline_done(&pipeline);
}

static __inline__ void video_stretch_palette16(unsigned dst_x, unsigned dst_y, unsigned dst_dx, unsigned dst_dy, uint16* src, unsigned src_dx, unsigned src_dy, int src_dw, int src_dp, unsigned* palette, unsigned combine) {
	struct video_pipeline_struct pipeline;

	video_stretch_palette16_pipeline_init(&pipeline, dst_dx, dst_dy, src_dx, src_dy, src_dw, src_dp, palette, combine);

	video_blit_pipeline(&pipeline, dst_x, dst_y, src);

	video_pipeline_done(&pipeline);
}

#endif
