/*
 * This file is part of the AdvanceMAME project
 *
 * Copyright (C) 1999-2000 Andrea Mazzoleni
 *
 * The contents of this file are free for non-commercial use.
 * Any commercial use is forbidden.
 */

#include "screen.h"
#include "scrvga.h"
#include "scrvbe.h"
#include "scrcard.h"

#include <stdarg.h>
#include <sys/farptr.h>
#include <dpmi.h>
#include <stdio.h>
#include <dos.h>
#include <assert.h>
#include <time.h>
#include <stdlib.h>

/* State */
video_internal video_state;
unsigned FAST (*video_write_selector)(void);
unsigned FAST (*video_write_line)(unsigned y);

/* Inizialize the video system
	Return the list of video mode avaliable
*/
void video_init(void) {
	assert( !video_active() );

	vbe_init();
	vga_init();
	card_init();

	memset(&video_state,0,sizeof(video_state));

	/* enable */
	video_state.active = 1;

	/* grab the actual video mode */
	if (video_mode_grab(&video_state.old_mode_info) == 0) {
		strcpy(video_state.old_mode_info.name,"startup_mode");
		video_state.old_mode_grabbed = 1;
	}
}

/* Deinizialize the video system */
void video_done() {
	assert( video_active() );

	if (video_mode_active())
		video_mode_done();

	if (card_active())
		card_done();
	if (vbe_active())
		vbe_done();
	if (vga_active())
		vga_done();

	/* disable */
	video_state.active = 0;
}

/***************************************************************************/
/* Options */

static int video_option_vbeline_tweak = VIDEO_OPTION_VBELINE_TWEAK_AGGRESSIVE;

void video_option_vbeline_tweak_set(int mode) {
	assert( mode ==	VIDEO_OPTION_VBELINE_TWEAK_MODERATE || mode == VIDEO_OPTION_VBELINE_TWEAK_AGGRESSIVE );
	video_option_vbeline_tweak = mode;
}

int video_option_vbeline_tweak_get(void) {
	return video_option_vbeline_tweak;
}

static int video_option_vbeline_driver = VIDEO_OPTION_VBELINE_DRIVER_AUTO;

void video_option_vbeline_driver_set(int mode) {
	video_option_vbeline_driver = mode;
}

int video_option_vbeline_driver_get(void) {
	return video_option_vbeline_driver;
}

static int video_option_default_mode = VIDEO_OPTION_DEFAULT_MODE_SMART;

void video_option_default_mode_set(int mode) {
	assert( mode ==	VIDEO_OPTION_DEFAULT_MODE_ALL || mode == VIDEO_OPTION_DEFAULT_MODE_SMART );
	video_option_default_mode = mode;
}

int video_option_default_mode_get(void) {
	return video_option_default_mode;
}

static int video_option_crtc_blank = VIDEO_OPTION_CRTC_BLANK_AGGRESSIVE;

void video_option_crtc_blank_set(int mode) {
	assert( mode ==	VIDEO_OPTION_CRTC_BLANK_AGGRESSIVE || mode == VIDEO_OPTION_CRTC_BLANK_MODERATE );
	video_option_crtc_blank = mode;
}

int video_option_crtc_blank_get(void) {
	return video_option_crtc_blank;
}

static double video_option_vbeline_clock_multiplier = 1.0;

void video_option_vbeline_clock_multiplier_set(double factor) {
	video_option_vbeline_clock_multiplier = factor;
}

double video_option_vbeline_clock_multiplier_get(void) {
	return video_option_vbeline_clock_multiplier;
}

/***************************************************************************/
/* CRCT */

void video_crtc_realize_h(const video_crtc* crtc, unsigned *de, unsigned* bs, unsigned* rs, unsigned* re, unsigned* be, unsigned* tt, unsigned x_div) {
	/* output horizontal */
	*tt = tweak_div_low(crtc->ht,x_div);
	*de = tweak_div_high(crtc->hde,x_div);
	*rs = tweak_div_high(crtc->hrs,x_div);
	*re = tweak_div_low(crtc->hre,x_div);
	if (video_option_crtc_blank_get() == VIDEO_OPTION_CRTC_BLANK_AGGRESSIVE) {
		*bs = *de + 1;
		if (*bs > *rs)
			*bs = *rs;
		*be = *re + 1;
	} else {
		*bs = *rs;
		*be = *re + 1;
	}
}

void video_crtc_realize_v(const video_crtc* crtc, unsigned *de, unsigned* bs, unsigned* rs, unsigned* re, unsigned* be, unsigned* tt, unsigned y_mul) {
	*tt = crtc->vt * y_mul;
	*de = crtc->vde * y_mul;
	*rs = crtc->vrs * y_mul;
	*re = crtc->vre * y_mul;
	if (video_option_crtc_blank_get() == VIDEO_OPTION_CRTC_BLANK_AGGRESSIVE) {
		*bs = *de + 1;
		if (*bs > *rs)
			*bs = *rs;
		*be = *re + 1;
	} else {
		*bs = *rs;
		*be = *re + 1;
	}
}

/* Compute crtc value from vga tweak information
	in:
		info tweak information
		mode video mode
		vert_freq input vertical frequency
	out:
		crtc computed value
	note:
		suppose NOT interlaced mode (VGA regs don't support this feature)
*/
video_error video_crtc_import(video_crtc* crtc, struct tweak_crtc* info, video_mode* mode, double vert_freq) {
	unsigned y_mul = 0;
	unsigned x_div = 0;

	switch (video_mode_driver(mode)) {
		case VIDEO_FLAGS_DRIVER_VGA :
		case VIDEO_FLAGS_DRIVER_VGALINE :
		case VIDEO_FLAGS_DRIVER_VGATWEAK :
			x_div = info->pixels_per_clock;
		break;
		case VIDEO_FLAGS_DRIVER_VBE :
		case VIDEO_FLAGS_DRIVER_VBELINE :
			if (info->hde == mode->size_x)
				x_div = 1;
			else if (info->hde*2 == mode->size_x)
				x_div = 2;
			else if (info->hde*4 == mode->size_x)
				x_div = 4;
			else if (info->hde*8 == mode->size_x)
				x_div = 8;
			else if (info->hde*16 == mode->size_x)
				x_div = 16;
		break;
	}

	if (!x_div) {
		video_error_description_set("Video mode don't support VGA CRTC regs grabbing");
		return -1;
	}

	crtc->hde = info->hde * x_div;
 	crtc->hrs = info->hrs * x_div;
	crtc->hre = info->hre * x_div;
	crtc->ht = info->ht * x_div;

	switch (video_mode_driver(mode)) {
		case VIDEO_FLAGS_DRIVER_VGA :
		case VIDEO_FLAGS_DRIVER_VGALINE :
		case VIDEO_FLAGS_DRIVER_VGATWEAK :
			if (info->is_graphics_mode) {
				y_mul = info->scanlines_per_line;
			} else {
				y_mul = 1;
			}
		break;
		case VIDEO_FLAGS_DRIVER_VBE :
		case VIDEO_FLAGS_DRIVER_VBELINE :
			if (info->vde == mode->size_y)
				y_mul = 1;
			else if (info->vde == 2*mode->size_y)
				y_mul = 2;
			else if (info->vde == 4*mode->size_y)
				y_mul = 4;
		break;
	}

	if (!y_mul) {
		video_error_description_set("Video mode don't support VGA CRTC regs grabbing");
		return -1;
	}

	crtc->doublescan = y_mul >= 2;

	crtc->vde = info->vde / y_mul;
	crtc->vrs = info->vrs / y_mul;
	crtc->vre = info->vre / y_mul;
	crtc->vt = info->vt / y_mul;

	/* pixelclock is computed, not imported */
	crtc->pixelclock = crtc->vt * y_mul * crtc->ht * vert_freq;

	crtc->hsync = info->hsync;
	crtc->vsync = info->vsync;

	crtc->interlaced = 0;
	return 0;
}

/***************************************************************************/
/* vga */

/* Import information for one video mode
	vm->driver.vga video mode
	Return 0 if successful
*/
int vga_mode_import(video_mode* mode) {
	if (mode->driver.vga.mode == 0x13) {
		mode->flags = VIDEO_FLAGS_DRIVER_VGA |
			VIDEO_FLAGS_LINEAR |
			VIDEO_FLAGS_INDEX_PACKED |
			VIDEO_FLAGS_TYPE_GRAPHICS |
			(mode->flags & VIDEO_FLAGS_USER_MASK);
		mode->size_x = 320;
		mode->size_y = 200;
		mode->virtual_x = 320;
		mode->virtual_y = 200;
		mode->vert_freq = 0;
		mode->horz_freq = 0;
		mode->bits = 8;
		mode->aspect_x = 4 * mode->size_y;
		mode->aspect_y = 3 * mode->size_x;
	} else if (mode->driver.vga.mode == 0x6A) {
		mode->flags = VIDEO_FLAGS_DRIVER_VGA |
			VIDEO_FLAGS_INDEX_PACKED |
			VIDEO_FLAGS_TYPE_GRAPHICS |
			(mode->flags & VIDEO_FLAGS_USER_MASK);
		mode->size_x = 400;
		mode->size_y = 300;
		mode->virtual_x = 400;
		mode->virtual_y = 400;
		mode->vert_freq = 0;
		mode->horz_freq = 0;
		mode->bits = 8;
		mode->aspect_x = 4 * mode->size_y;
		mode->aspect_y = 3 * mode->size_x;
	} else if (mode->driver.vga.mode == 0x3) {
		mode->flags = VIDEO_FLAGS_DRIVER_VGA |
			VIDEO_FLAGS_LINEAR |
			VIDEO_FLAGS_INDEX_PACKED |
			VIDEO_FLAGS_TYPE_TEXT |
			(mode->flags & VIDEO_FLAGS_USER_MASK);
		mode->size_x = 80*9;
		mode->size_y = 25*16;
		mode->virtual_x = 80*9;
		mode->virtual_y = 25*16;
		mode->vert_freq = 0;
		mode->horz_freq = 0;
		mode->bits = 0;
		mode->aspect_x = 4 * mode->size_y;
		mode->aspect_y = 3 * mode->size_x;
	} else if (mode->driver.vga.mode == 0x1) {
		mode->flags = VIDEO_FLAGS_DRIVER_VGA |
			VIDEO_FLAGS_LINEAR |
			VIDEO_FLAGS_INDEX_PACKED |
			VIDEO_FLAGS_TYPE_TEXT |
			(mode->flags & VIDEO_FLAGS_USER_MASK);
		mode->size_x = 40*9;
		mode->size_y = 25*16;
		mode->virtual_x = 80*9;
		mode->virtual_y = 25*16;
		mode->vert_freq = 0;
		mode->horz_freq = 0;
		mode->bits = 0;
		mode->aspect_x = 4 * mode->size_y;
		mode->aspect_y = 3 * mode->size_x;
	} else {
		video_error_description_set("Unsupported VGA mode %d",mode->driver.vga.mode);
		return -1;
	}

	return 0;
}

/***************************************************************************/
/* vgatweak */

static vgatweak_video_mode tweak_13_default = {
	0x13,
	25, /* CHECK */
	{
		{ 0x3c2, 0x0, 0x63 },
		{ 0x3d4, 0x0, 0x5f },
		{ 0x3d4, 0x1, 0x4f },
		{ 0x3d4, 0x2, 0x50 },
		{ 0x3d4, 0x3, 0x82 },
		{ 0x3d4, 0x4, 0x54 },
		{ 0x3d4, 0x5, 0x80 },
		{ 0x3d4, 0x6, 0xbf },
		{ 0x3d4, 0x7, 0x1f },
		{ 0x3d4, 0x8, 0x0 },
		{ 0x3d4, 0x9, 0x41 },
		{ 0x3d4, 0x10, 0x9c },
		{ 0x3d4, 0x11, 0x9e }, // (0x8e on S3)
		{ 0x3d4, 0x12, 0x8f },
		{ 0x3d4, 0x13, 0x28 },
		{ 0x3d4, 0x14, 0x40 },
		{ 0x3d4, 0x15, 0x96 },
		{ 0x3d4, 0x16, 0x79 }, // (0xb9 on S3)
		{ 0x3d4, 0x17, 0xa3 },
		{ 0x3c4, 0x1, 0x1 },
		{ 0x3c4, 0x4, 0xe },
		{ 0x3ce, 0x5, 0x40 },
		{ 0x3ce, 0x6, 0x5 },
		{ 0x3c0, 0x10, 0x41 },
		{ 0x3c0, 0x13, 0x0 }
	}
};

/*
 * plz  freq dsen blst rtst rten blen totl ch type pclock scan
 * h+         400  400  428  492  520  528  8 grph         100
 * v+         300  299  300  302  384  314  2 blank8bit
 */

static vgatweak_video_mode tweak_6A_default = {
	0x6A,
	25, /* CHECK */
	{
		{ 0x3c2, 0x0, 0x2f },
		{ 0x3d4, 0x0, 0x7f },
		{ 0x3d4, 0x1, 0x63 },
		{ 0x3d4, 0x2, 0x64 },
		{ 0x3d4, 0x3, 0x82 },
		{ 0x3d4, 0x4, 0x6b },
		{ 0x3d4, 0x5, 0x1b },
		{ 0x3d4, 0x6, 0x72 },
		{ 0x3d4, 0x7, 0xf0 },
		{ 0x3d4, 0x8, 0x0 },
		{ 0x3d4, 0x9, 0x61 },
		{ 0x3d4, 0x10, 0x58 },
		{ 0x3d4, 0x11, 0x8c },
		{ 0x3d4, 0x12, 0x57 },
		{ 0x3d4, 0x13, 0x32 },
		{ 0x3d4, 0x14, 0x0 },
		{ 0x3d4, 0x15, 0x57 },
		{ 0x3d4, 0x16, 0x0 },
		{ 0x3d4, 0x17, 0xe3 },
		{ 0x3c4, 0x1, 0x1 },
		{ 0x3c4, 0x4, 0x6 },
		{ 0x3ce, 0x5, 0x40 },
		{ 0x3ce, 0x6, 0x5 },
		{ 0x3c0, 0x10, 0x41 },
		{ 0x3c0, 0x13, 0x0 }
	}
};

static vgatweak_video_mode tweak_3_default = {
	0x3,
	25, /* CHECK */
	{
		{ 0x3c2, 0x0, 0x67 },
		{ 0x3d4, 0x0, 0x5f },
		{ 0x3d4, 0x1, 0x4f },
		{ 0x3d4, 0x2, 0x50 },
		{ 0x3d4, 0x3, 0x82 },
		{ 0x3d4, 0x4, 0x55 },
		{ 0x3d4, 0x5, 0x81 },
		{ 0x3d4, 0x6, 0xbf },
		{ 0x3d4, 0x7, 0x1f },
		{ 0x3d4, 0x8, 0x0 },
		{ 0x3d4, 0x9, 0x4f },
		{ 0x3d4, 0x10, 0x9c },
		{ 0x3d4, 0x11, 0x8e },
		{ 0x3d4, 0x12, 0x8f },
		{ 0x3d4, 0x13, 0x28 },
		{ 0x3d4, 0x14, 0x1f },
		{ 0x3d4, 0x15, 0x96 },
		{ 0x3d4, 0x16, 0xb9 },
		{ 0x3d4, 0x17, 0xa3 },
		{ 0x3c4, 0x1, 0x0 },
		{ 0x3c4, 0x4, 0x2 },
		{ 0x3ce, 0x5, 0x10 },
		{ 0x3ce, 0x6, 0xe },
		{ 0x3c0, 0x10, 0xc },
		{ 0x3c0, 0x13, 0x8 }
	}
};

static vgatweak_video_mode tweak_1_default = {
	0x1,
	25, /* CHECK */
	{
	{ 0x3c2, 0x0, 0x67 },
	{ 0x3d4, 0x0, 0x2d },
	{ 0x3d4, 0x1, 0x27 },
	{ 0x3d4, 0x2, 0x28 },
	{ 0x3d4, 0x3, 0x90 },
	{ 0x3d4, 0x4, 0x2b },
	{ 0x3d4, 0x5, 0xa0 },
	{ 0x3d4, 0x6, 0xbf },
	{ 0x3d4, 0x7, 0x1f },
	{ 0x3d4, 0x8, 0x0 },
	{ 0x3d4, 0x9, 0x4f },
	{ 0x3d4, 0x10, 0x9c },
	{ 0x3d4, 0x11, 0x8e },
	{ 0x3d4, 0x12, 0x8f },
	{ 0x3d4, 0x13, 0x14 },
	{ 0x3d4, 0x14, 0x1f },
	{ 0x3d4, 0x15, 0x96 },
	{ 0x3d4, 0x16, 0xb9 },
	{ 0x3d4, 0x17, 0xa3 },
	{ 0x3c4, 0x1, 0x8 },
	{ 0x3c4, 0x4, 0x2 },
	{ 0x3ce, 0x5, 0x10 },
	{ 0x3ce, 0x6, 0xe },
	{ 0x3c0, 0x10, 0xc },
	{ 0x3c0, 0x13, 0x8 }
	}
};

static int vgatweak_mode_set(vgatweak_video_mode* mode) {
	if (vga_mode_set(mode->mode,mode->regs,mode->regs_num)!=0) {
		return -1;
	}

	return 0;
}

static void vgatweak_mode_done(void) {
	vga_mode_done();
}

int vgatweak_mode_grab(vgatweak_video_mode* mode) {
	if (!vga_active())
		return -1;

	if (vga_mode_grab(&mode->mode,mode->regs,&mode->regs_num,VIDEO_MODE_REGS_MAX)!=0)
		return -1;

	return 0;
}

/* Import information for one video mode
	mode->tweak video mode
	Return 0 if successful
*/
static int vgatweak_mode_import_ext(video_mode* mode,vgatweak_video_mode* driver) {
	struct tweak_crtc info;
	tweak_crtc_get(&info,tweak_register_get, driver);

	if (info.is_graphics_mode && info.bits_per_pixel!=8) {
		video_error_description_set("Not a 8 bit mode");
		return -1;
	}

	mode->flags = VIDEO_FLAGS_DRIVER_VGATWEAK |
		VIDEO_FLAGS_ASYNC_SETPAGE |
		VIDEO_FLAGS_INDEX_PACKED |
		(mode->flags & VIDEO_FLAGS_USER_MASK);

	if (info.is_linear)
		mode->flags |= VIDEO_FLAGS_LINEAR;

	if (!info.is_graphics_mode) {
		mode->flags |= VIDEO_FLAGS_TYPE_TEXT;
		mode->bits = 0;
	} else {
		mode->flags |= VIDEO_FLAGS_TYPE_GRAPHICS;
		mode->bits = 8;
	}
	mode->size_x = info.size_x;
	mode->size_y = info.size_y;
        mode->virtual_x = info.virtual_x;
	mode->virtual_y = info.virtual_y;
        mode->vert_freq = info.vert_clock;
	mode->horz_freq = info.horz_clock;

#if 0
	/* euristic */
	mode->aspect_x = info.horz_display_end * info.vert_total;
	mode->aspect_y = info.vert_display_end * info.horz_total;
#else
	mode->aspect_x = 4 * mode->size_y;
	mode->aspect_y = 3 * mode->size_x;
#endif

	return 0;
}

int vgatweak_mode_import(video_mode* mode) {
	return vgatweak_mode_import_ext(mode,&mode->driver.vgatweak);
}

/***************************************************************************/
/* vgaline */

static int vgaline_mode_13_realize(vgatweak_video_mode* regs, const video_crtc* crtc) {
	unsigned x_div = 4; /* 4 pixel for clock in 8 bit mode */
	unsigned y_mul = crtc->doublescan!=0 ? 2 : 1;
	unsigned de,bs,rs,re,be,tt;

	/* Get dot clock */
	unsigned dotclock = crtc->pixelclock * 8 / x_div;
	if (vga_dotclock_get(&dotclock) != 0)
		return -1;

	/* output horizontal */
	video_crtc_realize_h(crtc,&de,&bs,&rs,&re,&be,&tt,x_div);
	tweak_htt_set(tt, tweak_register_set, tweak_register_get, regs);
	tweak_hde_set(de, tweak_register_set, tweak_register_get, regs);
	tweak_hrs_set(rs, tweak_register_set, tweak_register_get, regs);
	tweak_hre_set(re, tweak_register_set, tweak_register_get, regs);
	tweak_hbs_set(bs, tweak_register_set, tweak_register_get, regs);
	tweak_hbe_set(be, tweak_register_set, tweak_register_get, regs);

	/* output vertical */
	video_crtc_realize_v(crtc,&de,&bs,&rs,&re,&be,&tt,y_mul);
	tweak_vtt_set(tt, tweak_register_set, tweak_register_get, regs);
	tweak_vde_set(de, tweak_register_set, tweak_register_get, regs);
	tweak_vrs_set(rs, tweak_register_set, tweak_register_get, regs);
	tweak_vre_set(re, tweak_register_set, tweak_register_get, regs);
	tweak_vbs_set(bs, tweak_register_set, tweak_register_get, regs);
	tweak_vbe_set(be, tweak_register_set, tweak_register_get, regs);

	/* set char size */
	tweak_char_size_x_set(8, tweak_register_set, tweak_register_get, regs);

#if 0
	/* set char size to 1 */
	tweak_scanlines_per_line_set(1, tweak_register_set, tweak_register_get, regs);

	/* set double scan as requested */
	tweak_doublescan_set(crtc->doublescan != 0, tweak_register_set, tweak_register_get, regs);
#else
	/* set char size emulating doublescan */
	tweak_char_size_y_set(crtc->doublescan != 0 ? 2 : 1, tweak_register_set, tweak_register_get, regs);

	/* reset double scan */
	tweak_doublescan_set(0, tweak_register_set, tweak_register_get, regs);
#endif

	/* set dot clock */
	switch (dotclock) {
		case TWEAK_DOTCLOCK_HIGH:
			tweak_dotclock_middle_set(0, tweak_register_set, tweak_register_get, regs);
			tweak_masterclock_input_set(1, tweak_register_set, tweak_register_get, regs);
		break;
		case TWEAK_DOTCLOCK_HIGH/2 :
			tweak_dotclock_middle_set(1, tweak_register_set, tweak_register_get, regs);
			tweak_masterclock_input_set(1, tweak_register_set, tweak_register_get, regs);
		break;
		case TWEAK_DOTCLOCK_LOW :
			tweak_dotclock_middle_set(0, tweak_register_set, tweak_register_get, regs);
			tweak_masterclock_input_set(0, tweak_register_set, tweak_register_get, regs);
		break;
		case TWEAK_DOTCLOCK_LOW/2 :
			tweak_dotclock_middle_set(1, tweak_register_set, tweak_register_get, regs);
			tweak_masterclock_input_set(0, tweak_register_set, tweak_register_get, regs);
		break;
		default:
			return -1;
	}

	/* set polarity */
	tweak_hsync_set(crtc->hsync, tweak_register_set, tweak_register_get, regs);
	tweak_vsync_set(crtc->vsync, tweak_register_set, tweak_register_get, regs);

	/* set scanline length */
	tweak_offset_set((crtc->hde + 2 * x_div - 1) / (2 * x_div), tweak_register_set, tweak_register_get, regs);

	/* set chained/unchained mode */
	if (crtc->hde *	crtc->vde > 0x10000) {
		tweak_chained_set(0,tweak_register_set, tweak_register_get, regs);
	} else {
		tweak_chained_set(1,tweak_register_set, tweak_register_get, regs);
	}

	return 0;
}

static int vgaline_mode_6A_realize(vgatweak_video_mode* regs, const video_crtc* crtc) {
	unsigned x_div = 4; /* 4 pixel for clock in 8 bit mode */
	unsigned y_mul = crtc->doublescan!=0 ? 2 : 1;
	unsigned de,bs,rs,re,be,tt;

	/* output horizontal */
	video_crtc_realize_h(crtc,&de,&bs,&rs,&re,&be,&tt,x_div);
	tweak_htt_set(tt, tweak_register_set, tweak_register_get, regs);
	tweak_hde_set(de, tweak_register_set, tweak_register_get, regs);
	tweak_hrs_set(rs, tweak_register_set, tweak_register_get, regs);
	tweak_hre_set(re, tweak_register_set, tweak_register_get, regs);
	tweak_hbs_set(bs, tweak_register_set, tweak_register_get, regs);
	tweak_hbe_set(be, tweak_register_set, tweak_register_get, regs);

	/* output vertical */
	video_crtc_realize_v(crtc,&de,&bs,&rs,&re,&be,&tt,y_mul);
	tweak_vtt_set(tt, tweak_register_set, tweak_register_get, regs);
	tweak_vde_set(de, tweak_register_set, tweak_register_get, regs);
	tweak_vrs_set(rs, tweak_register_set, tweak_register_get, regs);
	tweak_vre_set(re, tweak_register_set, tweak_register_get, regs);
	tweak_vbs_set(bs, tweak_register_set, tweak_register_get, regs);
	tweak_vbe_set(be, tweak_register_set, tweak_register_get, regs);

	/* set char size */
	tweak_char_size_x_set(8, tweak_register_set, tweak_register_get, regs);

#if 0
	/* set char size to 1 */
	tweak_scanlines_per_line_set(1, tweak_register_set, tweak_register_get, regs);

	/* set double scan as requested */
	tweak_doublescan_set(crtc->doublescan != 0, tweak_register_set, tweak_register_get, regs);
#else
	/* set char size emulating doublescan */
	tweak_char_size_y_set(crtc->doublescan != 0 ? 2 : 1, tweak_register_set, tweak_register_get, regs);

	/* reset double scan */
	tweak_doublescan_set(0, tweak_register_set, tweak_register_get, regs);
#endif

	/* set polarity */
	tweak_hsync_set(crtc->hsync, tweak_register_set, tweak_register_get, regs);
	tweak_vsync_set(crtc->vsync, tweak_register_set, tweak_register_get, regs);

	/* set scanline length */
	tweak_offset_set((crtc->hde + 2 * x_div - 1) / (2 * x_div), tweak_register_set, tweak_register_get, regs);

	return 0;
}

static int vgaline_mode_text_realize(vgatweak_video_mode* regs, const video_crtc* crtc, unsigned font_x, unsigned font_y) {
	unsigned x_div = font_x;
	unsigned y_mul = crtc->doublescan!=0 ? 2 : 1;
	unsigned de,bs,rs,re,be,tt;

	/* Get dot clock (for text mode is equal at dot clock) */
	unsigned dotclock = crtc->pixelclock;
	if (vga_dotclock_get(&dotclock) != 0)
		return -1;

	/* output horizontal */
	video_crtc_realize_h(crtc,&de,&bs,&rs,&re,&be,&tt,x_div);
	tweak_htt_set(tt, tweak_register_set, tweak_register_get, regs);
	tweak_hde_set(de, tweak_register_set, tweak_register_get, regs);
	tweak_hrs_set(rs, tweak_register_set, tweak_register_get, regs);
	tweak_hre_set(re, tweak_register_set, tweak_register_get, regs);
	tweak_hbs_set(bs, tweak_register_set, tweak_register_get, regs);
	tweak_hbe_set(be, tweak_register_set, tweak_register_get, regs);

	/* output vertical */
	video_crtc_realize_v(crtc,&de,&bs,&rs,&re,&be,&tt,y_mul);
	tweak_vtt_set(tt, tweak_register_set, tweak_register_get, regs);
	tweak_vde_set(de, tweak_register_set, tweak_register_get, regs);
	tweak_vrs_set(rs, tweak_register_set, tweak_register_get, regs);
	tweak_vre_set(re, tweak_register_set, tweak_register_get, regs);
	tweak_vbs_set(bs, tweak_register_set, tweak_register_get, regs);
	tweak_vbe_set(be, tweak_register_set, tweak_register_get, regs);

	/* set char size */
	tweak_char_size_x_set(font_x, tweak_register_set, tweak_register_get, regs);
	tweak_char_size_y_set(font_y, tweak_register_set, tweak_register_get, regs);

	/* set double scan as requested */
	tweak_doublescan_set(crtc->doublescan != 0, tweak_register_set, tweak_register_get, regs);

	/* set dot clock */
	switch (dotclock) {
		case TWEAK_DOTCLOCK_HIGH :
			tweak_dotclock_middle_set(0, tweak_register_set, tweak_register_get, regs);
			tweak_masterclock_input_set(1, tweak_register_set, tweak_register_get, regs);
		break;
		case TWEAK_DOTCLOCK_HIGH/2 :
			tweak_dotclock_middle_set(1, tweak_register_set, tweak_register_get, regs);
			tweak_masterclock_input_set(1, tweak_register_set, tweak_register_get, regs);
		break;
		case TWEAK_DOTCLOCK_LOW :
			tweak_dotclock_middle_set(0, tweak_register_set, tweak_register_get, regs);
			tweak_masterclock_input_set(0, tweak_register_set, tweak_register_get, regs);
		break;
		case TWEAK_DOTCLOCK_LOW/2 :
			tweak_dotclock_middle_set(1, tweak_register_set, tweak_register_get, regs);
			tweak_masterclock_input_set(0, tweak_register_set, tweak_register_get, regs);
		break;
		default:
			return -1;
	}

	/* set polarity */
	tweak_hsync_set(crtc->hsync, tweak_register_set, tweak_register_get, regs);
	tweak_vsync_set(crtc->vsync, tweak_register_set, tweak_register_get, regs);

	/* set scanline length */
	tweak_offset_set((crtc->hde + 2 * x_div - 1) / (2 * x_div), tweak_register_set, tweak_register_get, regs);

	return 0;
}

int vgaline_mode_realize(vgatweak_video_mode* regs, const vgaline_video_mode* mode) {
	switch (mode->mode) {
		case 0x13:
			*regs = tweak_13_default;
			if (vgaline_mode_13_realize(regs,&mode->crtc)!=0) {
				video_error_description_set("Error in vgaline mode realize 0x13");
				return -1;
			}
			break;
		case 0x6A:
			*regs = tweak_6A_default;
			if (vgaline_mode_6A_realize(regs,&mode->crtc)!=0) {
				video_error_description_set("Error in vgaline mode realize 0x6A");
				return -1;
			}
			break;
		case 0x1 :
			*regs = tweak_1_default;
			if (vgaline_mode_text_realize(regs,&mode->crtc,9,8)!=0) {
				video_error_description_set("Error in vgaline mode realize 0x1");
				return -1;
			}
			break;
		case 0x3 :
			*regs = tweak_3_default;
			if (vgaline_mode_text_realize(regs,&mode->crtc,9,8)!=0) {
				video_error_description_set("Error in vgaline mode realize 0x3");
				return -1;
			}
			break;
		default:
			video_error_description_set("Unsupported mode %02x in vgaline mode realize", (unsigned)mode->mode);
			return -1;
	}
	return 0;
}

static int vgaline_mode_set(vgaline_video_mode* mode) {
	vgatweak_video_mode regs;

	video_log_modeline_m("vgaline: mode_set", mode->mode, mode->crtc.pixelclock, mode->crtc.hde, mode->crtc.hrs, mode->crtc.hre, mode->crtc.ht, mode->crtc.vde, mode->crtc.vrs, mode->crtc.vre, mode->crtc.vt, mode->crtc.hsync, mode->crtc.vsync, mode->crtc.doublescan, mode->crtc.interlaced );

	if (vgaline_mode_realize(&regs,mode)!=0) {
		return -1;
	}

	if (vga_mode_set(mode->mode,regs.regs,regs.regs_num)!=0) {
		return -1;
	}

	return 0;
}

static void vgaline_mode_done(void) {
	vga_mode_done();
}

int vgaline_mode_grab(vgaline_video_mode* mode) {
	struct tweak_crtc info;
	video_mode dummy;
	dummy.flags = VIDEO_FLAGS_DRIVER_VGALINE;

	if (!vga_active())
		return -1;

	/* get the video mode */
	if (vga_mode_get(&dummy.driver.vgaline.mode) != 0)
		return -1;

	/* import information */
	tweak_crtc_get(&info,tweak_reg_read,0);

	/* generate crtc values */
	if (video_crtc_import(&dummy.driver.vgaline.crtc,&info,&dummy,info.vert_clock)!=0)
		return -1;

	*mode = dummy.driver.vgaline;

	return 0;
}

/* Import information for one video mode
	Return 0 if successful
*/
int vgaline_mode_import(video_mode* mode) {
	vgatweak_video_mode regs;

	/* adjust the pixel clock to an acceptable value */
	if (vga_pixelclock_get(&mode->driver.vgaline.crtc.pixelclock,mode->driver.vgaline.mode) != 0)
		return -1;

	if (vgaline_mode_realize(&regs,&mode->driver.vgaline)!=0) {
		return -1;
	}

	if (vgatweak_mode_import_ext(mode,&regs)!=0) {
		return -1;
	}

	/* set values that depend on crtc values */
	mode->size_x = mode->driver.vgaline.crtc.hde;
	mode->size_y = mode->driver.vgaline.crtc.vde;
	mode->horz_freq = (double)mode->driver.vgaline.crtc.pixelclock / mode->driver.vgaline.crtc.ht;
        mode->vert_freq =  mode->horz_freq / mode->driver.vgaline.crtc.vt;
	if (mode->driver.vgaline.crtc.doublescan)
		mode->vert_freq /= 2;
	if (mode->driver.vgaline.crtc.interlaced)
		mode->vert_freq *= 2;

	mode->flags = (mode->flags & ~VIDEO_FLAGS_DRIVER_MASK) | VIDEO_FLAGS_DRIVER_VGALINE;

	return 0;
}

/***************************************************************************/
/* vbe */

int vbe_mode_grab(vbe_video_mode* mode) {
	if (!vbe_active())
		return -1;

	if (vbe_mode_get(&mode->mode)!=0) {
		return -1;
	}

	if (mode->mode < 0x100)
		return -1;

	return 0;
}

/* Import information for one video mode
	vm->vbe video mode
	Return 0 if successful
*/
int vbe_mode_import(video_mode* mode) {
	vbe_ModeInfoBlock info;

	if (vbe_mode_info_get( &info, mode->driver.vbe.mode)!=0) {
		video_error_description_set("VBE report that mode %d is unsupported", mode->driver.vbe.mode);
		return -1;
	}

	if ((info.ModeAttributes & vbeMdAvailable) == 0) {
		video_error_description_set("VBE report that mode %d is not avaliable", mode->driver.vbe.mode);
		return -1;
	}

	if ((mode->driver.vbe.mode & vbeLinearBuffer) != 0) {
		if (vbe_state.info.VESAVersion < 0x200) {
			video_error_description_set("Linear frame buffer not available on VBE version prior 2.0");
			return -1;
		}
		if ((info.ModeAttributes & vbeMdLinear) == 0) {
			video_error_description_set("Linear frame buffer not available in this mode");
			return -1;
		}
	} else {
		if ((info.ModeAttributes & vbeMdNonBanked) != 0) {
			video_error_description_set("Banked frame buffer not available in this mode");
			return -1;
		}
	}
	/* 8,15,16 bits per pixel */
	if (info.BitsPerPixel!=8 && info.BitsPerPixel!=15 && info.BitsPerPixel!=16) {
		video_error_description_set("Unsupported bits per pixel size");
		return -1;
	}
	/* No text mode */
	if (info.MemoryModel==vbeMemTXT) {
		video_error_description_set("Text mode not supported");
		return -1;
	}
	/* Packed or RGB memory model */
	if (info.MemoryModel!=vbeMemPK && info.MemoryModel!=vbeMemRGB) {
		video_error_description_set("Unsupported memory model");
		return -1;
	}
	/* Non planar mode */
	if (info.NumberOfPlanes!=1) {
		video_error_description_set("Unsupported number of planes");
		return -1;
	}
	mode->flags = VIDEO_FLAGS_DRIVER_VBE |
		VIDEO_FLAGS_ASYNC_SETPAGE |
		VIDEO_FLAGS_TYPE_GRAPHICS |
		(mode->flags & VIDEO_FLAGS_USER_MASK);
	if ((info.ModeAttributes & vbeMdTripleBuffer) != 0)
		mode->flags |= VIDEO_FLAGS_SYNC_SETPAGE;
	if (info.MemoryModel==vbeMemPK)
		mode->flags |= VIDEO_FLAGS_INDEX_PACKED;
	else
		mode->flags |= VIDEO_FLAGS_INDEX_RGB;
	if (mode->driver.vbe.mode & vbeLinearBuffer)
		mode->flags |= VIDEO_FLAGS_LINEAR;

	mode->size_x = info.XResolution;
	mode->size_y = info.YResolution;
        mode->virtual_x = 0;
	mode->virtual_y = 0;
	mode->vert_freq = 0;
	mode->horz_freq = 0;
	mode->bits = info.BitsPerPixel;
	mode->aspect_x = 4 * mode->size_y;
	mode->aspect_y = 3 * mode->size_x;
	return 0;
}

/***************************************************************************/
/* vbeline */

static int vbeline_mode_set(vbeline_video_mode* mode) {
	video_log_modeline_m("vbeline: mode_set", mode->mode, mode->crtc.pixelclock, mode->crtc.hde, mode->crtc.hrs, mode->crtc.hre, mode->crtc.ht, mode->crtc.vde, mode->crtc.vrs, mode->crtc.vre, mode->crtc.vt, mode->crtc.hsync, mode->crtc.vsync, mode->crtc.doublescan, mode->crtc.interlaced );
	return card_mode_set(mode->mode,&mode->crtc);
}

static void vbeline_mode_done(void) {
	card_mode_done();
}

int vbeline_mode_import(video_mode* mode) {
	return card_mode_import(mode);
}

/***************************************************************************/

void video_mode_done(void) {
	assert( video_mode_active() );

	switch (video_state.mode_info.flags & VIDEO_FLAGS_DRIVER_MASK) {
		case VIDEO_FLAGS_DRIVER_VGATWEAK :
			vgatweak_mode_done();
			break;
		case VIDEO_FLAGS_DRIVER_VGALINE :
			vgaline_mode_done();
			break;
		case VIDEO_FLAGS_DRIVER_VGA :
			vga_mode_done();
			break;
		case VIDEO_FLAGS_DRIVER_VBE :
			vbe_mode_done();
			break;
		case VIDEO_FLAGS_DRIVER_VBELINE :
			vbeline_mode_done();
			break;
		default:
			assert(0);
	}

	video_write_line = 0;
	video_write_selector = 0;

	/* disable */
	video_state.mode_active = 0;
}

static void video_state_rgb_set(void);
static void video_state_rgb_clear(void);

/* Set one video mode
 * return:
 *   ==0 if successfull
 */
video_error video_mode_set(video_mode* mode) {
	if (video_mode_active())
		video_mode_done();

	video_log("video: mode_set %s\n",video_mode_name(mode));

	switch (mode->flags & VIDEO_FLAGS_DRIVER_MASK) {
		case VIDEO_FLAGS_DRIVER_VGATWEAK :
			video_log("video: vgatweak_mode_set %x\n",mode->driver.vgatweak);
			if (vgatweak_mode_set(&mode->driver.vgatweak)!=0) {
				return -1;
			}
			video_state.mode_active = 1;
			video_state.mode_info = *mode;
			video_write_line = vga_write_line;
			video_write_selector = vga_write_selector;
		break;
		case VIDEO_FLAGS_DRIVER_VGALINE :
			video_log("video: vgaline_mode_set %x\n",mode->driver.vgaline);
			if (vgaline_mode_set(&mode->driver.vgaline)!=0) {
				return -1;
			}

			video_state.mode_active = 1;
			video_state.mode_info = *mode;
			video_write_line = vga_write_line;
			video_write_selector = vga_write_selector;
		break;
		case VIDEO_FLAGS_DRIVER_VGA :
			video_log("video: vga_mode_set %x\n",mode->driver.vga.mode);
			if (vga_mode_set(mode->driver.vga.mode,0,0)!=0) {
				return -1;
			}

			video_state.mode_active = 1;
			video_state.mode_info = *mode;
			video_write_line = vga_write_line;
			video_write_selector = vga_write_selector;
		break;
		case VIDEO_FLAGS_DRIVER_VBE :
			video_log("video: vbe_mode_set %x\n",mode->driver.vbe.mode);
			if (vbe_mode_set(mode->driver.vbe.mode,0)!=0) {
				return -1;
			}
			video_state.mode_active = 1;
			video_state.mode_info = *mode;
			video_state.mode_info.virtual_x = vbe_state.virtual_x;
			video_state.mode_info.virtual_y = vbe_state.virtual_y;
			video_write_line = vbe_write_line;
			video_write_selector = vbe_write_selector;
		break;
		case VIDEO_FLAGS_DRIVER_VBELINE : {
			video_log("video: vbeline_mode_set %x\n",mode->driver.vbeline);
			if (vbeline_mode_set(&mode->driver.vbeline)!=0) {
				return -1;
			}
			video_state.mode_active = 1;
			video_state.mode_info = *mode;
			video_state.mode_info.size_x = vbe_state.mode_info.XResolution;
			video_state.mode_info.size_y = vbe_state.mode_info.YResolution;
			video_state.mode_info.virtual_x = vbe_state.virtual_x;
			video_state.mode_info.virtual_y = vbe_state.virtual_y;
			video_write_line = vbe_write_line;
			video_write_selector = vbe_write_selector;
		} break;
	}

	/* rgb */
	if (video_index() == VIDEO_FLAGS_INDEX_RGB) {
		switch (mode->flags & VIDEO_FLAGS_DRIVER_MASK) {
			case VIDEO_FLAGS_DRIVER_VBE :
			case VIDEO_FLAGS_DRIVER_VBELINE :
				video_state.rgb_red_shift = vbe_state.rgb_red_shift;
				video_state.rgb_green_shift = vbe_state.rgb_green_shift;
				video_state.rgb_blue_shift = vbe_state.rgb_blue_shift;
				video_state.rgb_red_mask = vbe_state.rgb_red_mask;
				video_state.rgb_green_mask = vbe_state.rgb_green_mask;
				video_state.rgb_blue_mask = vbe_state.rgb_blue_mask;
			break;
		}
		video_state_rgb_set();
	} else {
		video_state_rgb_clear();
	}

	{
		double freq = video_measure_vert_freq();
		video_log("video: measured vert freq %.2f\n",freq);
	}

	return 0;
}

/* Grab the current video mode */
int video_mode_grab(video_mode* mode) {
	if (vbe_mode_grab(&mode->driver.vbe)==0 && vbe_mode_import(mode)==0)
		return 0;
	if (vgatweak_mode_grab(&mode->driver.vgatweak)==0 && vgatweak_mode_import(mode)==0)
		return 0;
	if (vgaline_mode_grab(&mode->driver.vgaline)==0 && vgaline_mode_import(mode)==0)
		return 0;
	return -1;
}

/* Import the specified video mode */
video_error video_mode_import(video_mode* vm) {
	switch (video_mode_driver(vm)) {
		case VIDEO_FLAGS_DRIVER_VGA: return vga_mode_import(vm);
		case VIDEO_FLAGS_DRIVER_VGALINE: return vgaline_mode_import(vm);
		case VIDEO_FLAGS_DRIVER_VGATWEAK: return vgatweak_mode_import(vm);
		case VIDEO_FLAGS_DRIVER_VBE: return vbe_mode_import(vm);
		case VIDEO_FLAGS_DRIVER_VBELINE: return vbeline_mode_import(vm);
	}
	return -1;
}

/* Import the specified video mode and insert in the container */
video_error video_mode_import_and_insert(video_mode_container* vmc, video_mode* vm) {
	switch (video_mode_driver(vm)) {
		case VIDEO_FLAGS_DRIVER_VGA: return vga_mode_import_and_insert(vmc,vm);
		case VIDEO_FLAGS_DRIVER_VGALINE: return vgaline_mode_import_and_insert(vmc,vm);
		case VIDEO_FLAGS_DRIVER_VGATWEAK: return vgatweak_mode_import_and_insert(vmc,vm);
		case VIDEO_FLAGS_DRIVER_VBE: return vbe_mode_import_and_insert(vmc,vm);
		case VIDEO_FLAGS_DRIVER_VBELINE: return vbeline_mode_import_and_insert(vmc,vm);
	}
	return -1;
}

/* Set the startup mode */
void video_mode_reset(void) {
	int mode_reset = 0;

	if (video_mode_active())
		video_mode_done();

	if (video_state.old_mode_grabbed) {
		if (video_mode_set(&video_state.old_mode_info) == 0) {
			video_mode_done();
			mode_reset = 1;
		}
	}

	if (!mode_reset) {
		vga_mode_set(0x3,0,0);
		vga_mode_done();
	}
}

/* Wait for a vertical retrace */
void video_wait_vsync(void) {
	switch (video_driver()) {
		default:
			assert(0);
		case VIDEO_FLAGS_DRIVER_VGA :
		case VIDEO_FLAGS_DRIVER_VGALINE :
		case VIDEO_FLAGS_DRIVER_VGATWEAK :
			vga_wait_vsync();
			break;
		case VIDEO_FLAGS_DRIVER_VBE :
		case VIDEO_FLAGS_DRIVER_VBELINE :
			vbe_wait_vsync();
			break;
	}
}

/* Measure the frame rate */
float video_measure_vert_freq(void) {
	double vert_freq = 0;
	unsigned retry = 4;

	uclock(); /* reset clock */

	do {
		clock_t start, stop;
		video_wait_vsync();
		start = uclock();
		video_wait_vsync();
		stop = uclock();
		if (start != stop) {
			vert_freq = (double)UCLOCKS_PER_SEC / (stop - start);
		}
		--retry;
	} while (retry && (vert_freq < 10 || vert_freq > 300));

	if (!retry)
		return 0;

	return vert_freq;
}

/* Set the visual offset, return immediatially */
video_error video_display_set_async(unsigned offset, int waitvsync) {
	assert( video_mode_active() );

	switch (video_driver()) {
		default:
			assert(0);
		case VIDEO_FLAGS_DRIVER_VGA :
		case VIDEO_FLAGS_DRIVER_VGALINE :
		case VIDEO_FLAGS_DRIVER_VGATWEAK :
			if (vga_scroll(offset,waitvsync)!=0) {
				return -1;
			}
			break;
		case VIDEO_FLAGS_DRIVER_VBE :
		case VIDEO_FLAGS_DRIVER_VBELINE :
			if (vbe_scroll(offset,waitvsync)!=0) {
				return -1;
			}
			break;
	}
	return 0;
}

/* Set the visual offset at next vsync, return immeditially */
video_error video_display_set_sync(unsigned offset) {
	assert( video_mode_active() );

	switch (video_driver()) {
		default:
			video_error_description_set("Syncronous scroll unsupported with this video mode");
			return -1;
		case VIDEO_FLAGS_DRIVER_VBE :
		case VIDEO_FLAGS_DRIVER_VBELINE :
			if (vbe_request_scroll(offset)!=0) {
				return -1;
			}
			break;
	}
	return 0;
}

unsigned video_bytes_per_page(void) {
	unsigned bytes_per_page = video_size_y() * video_bytes_per_scanline();
	switch (video_driver()) {
		default:
			assert(0);
		case VIDEO_FLAGS_DRIVER_VBE :
		case VIDEO_FLAGS_DRIVER_VBELINE :
			return vbe_adjust_bytes_per_page(bytes_per_page);
		case VIDEO_FLAGS_DRIVER_VGA :
		case VIDEO_FLAGS_DRIVER_VGALINE :
		case VIDEO_FLAGS_DRIVER_VGATWEAK :
			return vga_adjust_bytes_per_page(bytes_per_page);
	}
}

static __inline__ void video_chained_put_pixel(unsigned x, unsigned y, unsigned color) {
	switch (video_bytes_per_pixel()) {
		case 1 :
			_farpokeb( video_write_selector(), video_write_line(y) + x, color);
		break;
		case 2 :
			_farpokew( video_write_selector(), video_write_line(y) + 2*x, color);
		break;
		case 4 :
			_farpokel( video_write_selector(), video_write_line(y) + 4*x, color);
		break;
	}
}

static __inline__ void video_unchained_put_pixel(unsigned x, unsigned y, unsigned color) {
	vga_unchained_plane_set(x % 4);
	_farpokeb( vga_write_selector(), vga_write_line(y) + x / 4, color);
}

void video_put_char(unsigned x, unsigned y, char c, unsigned color) {
	assert( video_is_text() );
	_farpokew( vga_write_selector(), vga_write_line(y) + 2*x, (color << 8) | (unsigned char)c);
}

static __inline__ void video_put_pixel_noclip(unsigned x, unsigned y, unsigned color) {
	assert( video_is_graphics() );
	switch (video_driver()) {
		case VIDEO_FLAGS_DRIVER_VGA :
		case VIDEO_FLAGS_DRIVER_VGALINE :
		case VIDEO_FLAGS_DRIVER_VGATWEAK :
			if (vga_state.unchained_active)
				video_unchained_put_pixel(x,y,color);
			else
				video_chained_put_pixel(x,y,color);
		break;
		case VIDEO_FLAGS_DRIVER_VBE :
		case VIDEO_FLAGS_DRIVER_VBELINE :
			video_chained_put_pixel(x,y,color);
			break;
		default:
			assert(0);
	}
}

void video_put_pixel(unsigned x, unsigned y, unsigned color) {
	if (x < video_virtual_x() && y < video_virtual_y())
		video_put_pixel_noclip(x,y,color);
}

/****************************************************************************/
/* color */

/* Compute the distance of two color */
unsigned video_color_dist(const video_color* A, const video_color* B) {
	int r,g,b;
	r = (int)A->red - B->red;
	g = (int)A->green - B->green;
	b = (int)A->blue - B->blue;
	return r*r + g*g + b*b;
}

/* Compute the size in bit of the mask */
static unsigned video_rgb_len(unsigned mask) {
	unsigned len = 0;
	while ((mask & 1) == 0)
		mask >>= 1;
	while (mask) {
		++len;
		mask >>= 1;
	}
	return len;
}

/* Update the video_state.rgb_* variables */
static void video_state_rgb_set(void) {
	video_state.rgb_red_len = video_rgb_len(video_state.rgb_red_mask);
	video_state.rgb_green_len = video_rgb_len(video_state.rgb_green_mask);
	video_state.rgb_blue_len = video_rgb_len(video_state.rgb_blue_mask);
	video_state.rgb_mask_bit = video_rgb_get(0xFF,0xFF,0xFF);
	video_state.rgb_high_bit = video_rgb_get(0x80,0x80,0x80);
	video_state.rgb_low_bit =  video_rgb_get(
		1 << (8-video_state.rgb_red_len),
		1 << (8-video_state.rgb_green_len),
		1 << (8-video_state.rgb_blue_len));
}

static void video_state_rgb_clear(void) {
	video_state.rgb_red_shift = 0;
	video_state.rgb_green_shift = 0;
	video_state.rgb_blue_shift = 0;
	video_state.rgb_red_mask = 0;
	video_state.rgb_green_mask = 0;
	video_state.rgb_blue_mask = 0;
	video_state.rgb_red_len = 0;
	video_state.rgb_green_len = 0;
	video_state.rgb_blue_len = 0;
	video_state.rgb_mask_bit = 0;
	video_state.rgb_high_bit = 0;
	video_state.rgb_low_bit =  0;
}

static unsigned video_rgb_approx(unsigned value, unsigned len) {
	unsigned fill = len;
	while (fill < 8) {
		value |= value >> fill;
		fill *= 2;
	}
	return value;
}

unsigned video_red_get_approx(unsigned rgb) {
	return video_rgb_approx(video_red_get(rgb), video_state.rgb_red_len);
}

unsigned video_green_get_approx(unsigned rgb) {
	return video_rgb_approx(video_green_get(rgb), video_state.rgb_green_len);
}

unsigned video_blue_get_approx(unsigned rgb) {
	return video_rgb_approx(video_blue_get(rgb), video_state.rgb_blue_len);
}

/* Compute shift and mask value */
static __inline__ void video_rgb_data(int* shift, unsigned* mask, unsigned posbit, unsigned masksize) {
	*mask = ((1 << masksize) - 1) << posbit;
	*shift = posbit + masksize - 8;
}

/* Convert one video mode from PACKED to RGB */
void video_palette_to_rgb(int waitvsync) {
	unsigned bit_r;
	unsigned bit_g;
	unsigned bit_b;

	video_color* palette;
	unsigned i;
	unsigned colors = 1 << video_bits_per_pixel();

	assert( (video_flags() & VIDEO_FLAGS_INDEX_MASK) == VIDEO_FLAGS_INDEX_PACKED );

	bit_b = video_bits_per_pixel() / 3;
	bit_r =	(video_bits_per_pixel() - bit_b) / 2;
	bit_g = video_bits_per_pixel() - bit_b - bit_r;

	video_rgb_data(&video_state.rgb_red_shift,&video_state.rgb_red_mask,0,bit_r);
	video_rgb_data(&video_state.rgb_green_shift,&video_state.rgb_green_mask,bit_r,bit_g);
	video_rgb_data(&video_state.rgb_blue_shift,&video_state.rgb_blue_mask,bit_r+bit_g,bit_b);

	video_state_rgb_set();

	palette = malloc(colors * sizeof(video_color));
	assert( palette );

	for(i=0;i<colors;++i) {
		palette[i].red = video_red_get_approx(i);
		palette[i].green = video_green_get_approx(i);
		palette[i].blue = video_blue_get_approx(i);
	}

	video_palette_set(palette,0,colors,waitvsync);

	free(palette);

	/* set RGB */
	video_state.mode_info.flags &= ~VIDEO_FLAGS_INDEX_MASK;
	video_state.mode_info.flags |= VIDEO_FLAGS_INDEX_RGB;
}

video_error video_palette_set(video_color* palette, unsigned start, unsigned count, int waitvsync) {
	assert( (video_flags() & VIDEO_FLAGS_INDEX_MASK) == VIDEO_FLAGS_INDEX_PACKED );
	assert( count <= 256 );

	switch (video_driver()) {
		default:
			assert(0);
		case VIDEO_FLAGS_DRIVER_VGA :
		case VIDEO_FLAGS_DRIVER_VGALINE :
		case VIDEO_FLAGS_DRIVER_VGATWEAK : {
			video_color vga_pal[256];
			unsigned i;
			for(i=0;i<count;++i) {
				vga_pal[i].red = palette[i].red >> 2;
				vga_pal[i].green = palette[i].green >> 2;
				vga_pal[i].blue = palette[i].blue >> 2;
				vga_pal[i].alpha = 0;
			}
			vga_palette6_set(vga_pal,start,count,waitvsync);
                }
		break;
		case VIDEO_FLAGS_DRIVER_VBE :
		case VIDEO_FLAGS_DRIVER_VBELINE : {
			video_color vbe_pal[256];
			unsigned shift = 8 - vbe_state.palette_width;
			if (shift) {
				unsigned i;
				for(i=0;i<count;++i) {
					vbe_pal[i].red = palette[i].red >> shift;
					vbe_pal[i].green = palette[i].green >> shift;
					vbe_pal[i].blue = palette[i].blue >> shift;
					vbe_pal[i].alpha = 0;
				}
			} else {
				unsigned i;
				for(i=0;i<count;++i) {
					vbe_pal[i].red = palette[i].red;
					vbe_pal[i].green = palette[i].green;
					vbe_pal[i].blue = palette[i].blue;
					vbe_pal[i].alpha = 0;
				}
			}

			if (vbe_palette_set(vbe_pal,start,count,waitvsync)!=0) {
				return -1;
			}
                }
		break;
	}

	return 0;
}

/****************************************************************************/
/* error */

void video_error_description_set(const char* error,...) {
	va_list arg;
	va_start(arg,error);
	vsprintf(video_state.error,error,arg);
	va_end(arg);
}

/****************************************************************************/
/* log */

void video_log_modeline_mb(const char *text, unsigned mode, unsigned pixel_clock, unsigned hde, unsigned hbs, unsigned hrs, unsigned hre, unsigned hbe, unsigned ht, unsigned vde, unsigned vbs, unsigned vrs, unsigned vre, unsigned vbe, unsigned vt, int hsync_pol, int vsync_pol, int doublescan, int interlace) {
	const char* flag1 = hsync_pol ? " -hsync" : " +hsync";
	const char* flag2 = vsync_pol ? " -vsync" : " +vsync";
	const char* flag3 = doublescan ? " doublescan" : "";
	const char* flag4 = interlace ? " interlaced" : "";
	video_log("%s(%x %.2f %d %d %d %d %d %d %d %d %d %d %d %d%s%s%s%s)\n",
       		text, mode, (double)pixel_clock / 1E6,
       		hde, hbs, hrs, hre, hbe, ht,
		vde, vbs, vrs, vre, vbe, vt,
		flag1, flag2, flag3, flag4
	);
}

void video_log_modeline_m(const char *text, unsigned mode, unsigned pixel_clock, unsigned hde, unsigned hrs, unsigned hre, unsigned ht, unsigned vde, unsigned vrs, unsigned vre, unsigned vt, int hsync_pol, int vsync_pol, int doublescan, int interlace) {
	const char* flag1 = hsync_pol ? " -hsync" : " +hsync";
	const char* flag2 = vsync_pol ? " -vsync" : " +vsync";
	const char* flag3 = doublescan ? " doublescan" : "";
	const char* flag4 = interlace ? " interlaced" : "";
	video_log("%s(%x %.2f %d %d %d %d %d %d %d %d%s%s%s%s)\n",
       		text, mode, (double)pixel_clock / 1E6,
       		hde, hrs, hre, ht,
		vde, vrs, vre, vt,
		flag1, flag2, flag3, flag4
	);
}


void video_log_modeline_b(const char *text, unsigned pixel_clock, unsigned hde, unsigned hbs, unsigned hrs, unsigned hre, unsigned hbe, unsigned ht, unsigned vde, unsigned vbs, unsigned vrs, unsigned vre, unsigned vbe, unsigned vt, int hsync_pol, int vsync_pol, int doublescan, int interlace) {
	const char* flag1 = hsync_pol ? " -hsync" : " +hsync";
	const char* flag2 = vsync_pol ? " -vsync" : " +vsync";
	const char* flag3 = doublescan ? " doublescan" : "";
	const char* flag4 = interlace ? " interlaced" : "";
	video_log("%s(%.2f %d %d %d %d %d %d %d %d %d %d %d %d%s%s%s%s)\n",
       		text, (double)pixel_clock / 1E6,
       		hde, hbs, hrs, hre, hbe, ht,
		vde, vbs, vrs, vre, vbe, vt,
		flag1, flag2, flag3, flag4
	);
}

