/*
 * This file is part of the AdvanceMAME project
 *
 * Copyright (C) 1999-2000 Andrea Mazzoleni
 *
 * The contents of this file are free for non-commercial use.
 * Any commercial use is forbidden.
 */

#include "screen.h"
#include "scrvga.h"
#include "scrvbe.h"
#include "scrcard.h"

/* from allegro.h + const added */
extern char *get_config_string(const char *section, const char *name, const char *def);
extern void set_config_string(const char *section, const char *name, const char *val);

#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/movedata.h>
#include <sys/farptr.h>

extern char _video_error[];

/***************************************************************************/
/* Video mode container */

/* Inizialize the video mode container */
void video_mode_container_init(video_mode_container* vmc) {
	vmc->base = 0;
}

/* Deinizialize the video mode container */
void video_mode_container_done(video_mode_container* vmc) {
	video_mode* p = vmc->base;
	while (p) {
		video_mode* todelete = p;
		p = p->next;
		free(todelete);
	}
}

/* Check if mode is already in container
	Don't compare name, but the mode parameter
	Return ==0 if not found, otherwise the mode in the container
*/
const video_mode* video_mode_container_has(video_mode_container* vmc, const video_mode* vm, int (*compare)(const video_mode* a,const video_mode* b)) {
	video_mode* p = vmc->base;
	while (p) {
		if (compare(p,vm)==0)
			return p;
		p = p->next;
	}
	return 0;
}

/* Remove a mode in the container
	Return the new container
*/
void video_mode_container_remove(video_mode_container* vmc, const video_mode* vm, int (*compare)(const video_mode* a,const video_mode* b)) {
	if (vmc->base) {
		if (compare(vmc->base,vm)==0) {
			video_mode* todelete = vmc->base;
			vmc->base = vmc->base->next;
			free(todelete);
		} else {
			video_mode* p = vmc->base;
			while (p->next) {
				if (compare(p->next,vm)==0) {
					video_mode* todelete = p->next;
					p->next = p->next->next;
					free(todelete);
				} else
					p = p->next;
			}
		}
	}
}

void video_mode_container_insert(video_mode_container* vmc, const video_mode* vm) {
	video_mode* p =	malloc(sizeof(video_mode));
	*p = *vm;
	p->next = vmc->base;
	vmc->base = p;
}

/* Insert in order */
void video_mode_container_insert_sort(video_mode_container* vmc,  const video_mode* vm, int (*compare)(const video_mode* a,const video_mode* b)) {
	if (vmc->base && compare(vmc->base,vm)<=0) {
		video_mode* p = vmc->base;
		video_mode* c;
		while (p->next && compare(p->next,vm)<=0) {
			p = p->next;
		}
		c = malloc(sizeof(video_mode));
		*c = *vm;
		c->next = p->next;
		p->next = c;
	} else {
		video_mode* p =	malloc(sizeof(video_mode));
		*p = *vm;
		p->next = vmc->base;
		vmc->base = p;
	}
}

/* Compare two video mode
	Return ==0 if equal
*/
int video_mode_compare(const video_mode* vm0, const video_mode* vm1) {
	int m0 = vm0->flags & VIDEO_FLAGS_DRIVER_MASK;
	int m1 = vm1->flags & VIDEO_FLAGS_DRIVER_MASK;
	if (m0!=m1)
		return 1;
	switch (m0) {
		case VIDEO_FLAGS_DRIVER_VGA :
			return memcmp(&vm0->driver.vga,&vm1->driver.vga,sizeof(vga_video_mode))!=0;
		case VIDEO_FLAGS_DRIVER_VGATWEAK : {
			/* TODO do an effective compare ? order may be important ! */
			unsigned i;
			if (vm0->driver.vgatweak.mode != vm1->driver.vgatweak.mode)
				return 1;
			if (vm0->driver.vgatweak.regs_num != vm1->driver.vgatweak.regs_num)
				return 1;
			for(i=0;i<vm0->driver.vgatweak.regs_num;++i) {
				if (vm0->driver.vgatweak.regs[i].port != vm1->driver.vgatweak.regs[i].port)
					return 1;
				if (vm0->driver.vgatweak.regs[i].index != vm1->driver.vgatweak.regs[i].index)
					return 1;
				if (vm0->driver.vgatweak.regs[i].value != vm1->driver.vgatweak.regs[i].value)
					return 1;
			}
			return 0;
		}
		case VIDEO_FLAGS_DRIVER_VGALINE :
			return memcmp(&vm0->driver.vgaline,&vm1->driver.vgaline,sizeof(vgaline_video_mode))!=0;
		case VIDEO_FLAGS_DRIVER_VBE :
			return memcmp(&vm0->driver.vbe,&vm1->driver.vbe,sizeof(vbe_video_mode))!=0;
		case VIDEO_FLAGS_DRIVER_VBELINE :
			return memcmp(&vm0->driver.vbeline,&vm1->driver.vbeline,sizeof(vbeline_video_mode))!=0;
 	}
	return 1;
}

/* ----------------------------------------------------------------------- */
/* Parse */

/* Line in parsing */
static char video_mode_parse_error[1024];

static video_error parse_int(int* r, const char** begin, const char* end) {
	char* e;
 	*r = strtol(*begin,&e,10);
	if (e==*begin || e > end) {
		sprintf(video_mode_parse_error,"Error in value");
		return -1;
	}
	*begin = e;
	return 0;
}

static void parse_separator(const char* separator, const char** begin, const char* end) {
	while (*begin < end && strchr(separator,**begin))
		++*begin;
}

static video_error parse_token(char* token, unsigned size, const char* separator, const char** begin, const char* end) {
	unsigned pos = 0;
	while (*begin < end && !strchr(separator,**begin)) {
		if (pos + 2 > size) {
			sprintf(video_mode_parse_error,"Token too long");
			return -1;
		}
		token[pos] = **begin;
		++pos;
		++*begin;
	}
	if (!pos) {
		sprintf(video_mode_parse_error,"Expected token");
		return -1;
	}
	token[pos] = 0;
	return 0;
}

static video_error parse_hex(unsigned* r, const char** begin, const char* end) {
	char* e;
 	*r = strtoul(*begin,&e,16);
	if (e==*begin || e > end) {
		sprintf(video_mode_parse_error,"Error in value");
		return -1;
	}
	*begin = e;
	return 0;
}

static video_error parse_double(double* r, const char** begin, const char* end) {
	char* e;
 	*r = strtod(*begin,&e);
	if (e==*begin || e > end) {
		sprintf(video_mode_parse_error,"Error in value");
		return -1;
	}
	*begin = e;
	return 0;
}

static video_error parse_reg(tweak_register* r, const char** begin, const char* end) {
	unsigned v;
	parse_separator(" \t",begin,end);
	if (parse_hex(&v,begin,end))
		return -1;
	r->port = v;
	parse_separator(" \t",begin,end);
	if (**begin!='[') {
		sprintf(video_mode_parse_error,"Miss [");
		return -1;
	}
	++*begin;
	parse_separator(" \t",begin,end);
	if (parse_hex(&v,begin,end))
		return -1;
	r->index = v;
	parse_separator(" \t",begin,end);
	if (**begin!=']') {
		sprintf(video_mode_parse_error,"Miss ]");
		return -1;
	}
	++*begin;
	parse_separator(" \t",begin,end);
	if (**begin!='=') {
		sprintf(video_mode_parse_error,"Miss =");
		return -1;
	}
	++*begin;
	parse_separator(" \t",begin,end);
	if (parse_hex(&v,begin,end))
		return -1;
	r->value = v;
	parse_separator(" \t",begin,end);
	return 0;
}

static video_error parse_vgatweak(vgatweak_video_mode* mode, const char* begin, const char* end) {
	unsigned v;
	parse_separator(" \t",&begin,end);
	if (parse_hex(&v,&begin,end))
		return -1;
	mode->mode = v;
	mode->regs_num = 0;
	while (begin != end) {
		if (mode->regs_num >= VIDEO_MODE_REGS_MAX) {
			sprintf(video_mode_parse_error,"Too many regs");
			return -1;
		}
		if (parse_reg( mode->regs + mode->regs_num, &begin, end))
			return -1;
		++mode->regs_num;
	}
	return 0;
}

static video_error parse_modecrtc(unsigned*mode, video_crtc* crtc, const char* begin, const char* end) {
	int v;
	unsigned u;
	double d;
	parse_separator(" \t",&begin,end);
	if (parse_hex(&u,&begin,end))
		return -1;
	*mode = u;
	parse_separator(" \t",&begin,end);
	if (parse_double(&d,&begin,end))
		return -1;
	crtc->pixelclock = d * 1E6;
	parse_separator(" \t",&begin,end);
	if (parse_int(&v,&begin,end))
		return -1;
	crtc->hde = v;
	parse_separator(" \t",&begin,end);
	if (parse_int(&v,&begin,end))
		return -1;
	crtc->hrs = v;
	parse_separator(" \t",&begin,end);
	if (parse_int(&v,&begin,end))
		return -1;
	crtc->hre = v;
	parse_separator(" \t",&begin,end);
	if (parse_int(&v,&begin,end))
		return -1;
	crtc->ht = v;
	parse_separator(" \t",&begin,end);
	if (parse_int(&v,&begin,end))
		return -1;
	crtc->vde = v;
	parse_separator(" \t",&begin,end);
	if (parse_int(&v,&begin,end))
		return -1;
	crtc->vrs = v;
	parse_separator(" \t",&begin,end);
	if (parse_int(&v,&begin,end))
		return -1;
	crtc->vre = v;
	parse_separator(" \t",&begin,end);
	if (parse_int(&v,&begin,end))
		return -1;
	crtc->vt = v;

	/* defaults */
	crtc->doublescan = 0;
	crtc->interlaced = 0;
	crtc->hsync = 1;
	crtc->vsync = 0;

	parse_separator(" \t",&begin,end);
	while (begin != end) {
		char t[16];
		if (parse_token(t,sizeof(t)," \t",&begin,end))
			return -1;
		if (stricmp(t,"doublescan")==0) {
			crtc->doublescan = 1;
		} else if (stricmp(t,"+hsync")==0) {
			crtc->hsync = 0;
		} else if (stricmp(t,"-hsync")==0) {
			crtc->hsync = 1;
		} else if (stricmp(t,"+vsync")==0) {
			crtc->vsync = 0;
		} else if (stricmp(t,"-vsync")==0) {
			crtc->vsync = 1;
		} else if (stricmp(t,"interlaced")==0) {
			crtc->interlaced = 1;
		} else {
			sprintf(video_mode_parse_error,"Unknow token");
			return -1;
		}
		parse_separator(" \t",&begin,end);
	}
	return 0;
}

static int print_modecrtc(char* s, unsigned mode, const video_crtc* crtc) {
	const char* flag1 = crtc->doublescan ? " doublescan" : "";
	const char* flag2 = crtc->hsync ? " -hsync" : " +hsync";
	const char* flag3 = crtc->vsync ? " -vsync" : " +vsync";
	const char* flag4 = crtc->interlaced ? " interlaced" : "";
	return sprintf(s,"%x %.2f %d %d %d %d %d %d %d %d%s%s%s%s",
       		mode, (double)crtc->pixelclock / 1E6,
       		crtc->hde, crtc->hrs, crtc->hre, crtc->ht,
		crtc->vde, crtc->vrs, crtc->vre, crtc->vt,
		flag1, flag2, flag3, flag4
	);
}

/* Parse a line */
video_error video_mode_parse_line(video_mode* mode, const char* begin) {
	char name[VIDEO_MODE_NAME_MAX];
	char type[64];
	const char* end = begin + strlen(begin);
	const char* argend;

	parse_separator(" \t",&begin,end);
	if (parse_token(name,sizeof(name)," \t",&begin,end))
		return -1;
	parse_separator(" \t",&begin,end);
	if (parse_token(type,sizeof(type)," \t(",&begin,end))
		return -1;
	parse_separator(" \t",&begin,end);
	if (*begin!='(') {
		sprintf(video_mode_parse_error,"Miss (");
		return -1;
	}
	++begin;

	argend = strchr(begin,')');
	if (!argend || argend >= end) {
		sprintf(video_mode_parse_error,"Miss )");
		return -1;
	}

	memset(mode,0,sizeof(video_mode));
	if (strlen(name) >= VIDEO_MODE_NAME_MAX) {
		sprintf(video_mode_parse_error,"Video mode name too long");
		return -1;
	}
	strcpy(mode->name,name);

	/* type */
	if (strcmp(type,"vbe")==0) {
		mode->flags = VIDEO_FLAGS_DRIVER_VBE;
 		if (parse_hex(&mode->driver.vbe.mode,&begin,argend))
 			return -1;
	} else if (strcmp(type,"vga")==0) {
		mode->flags = VIDEO_FLAGS_DRIVER_VGA;
		if (parse_hex(&mode->driver.vga.mode,&begin,argend))
			return -1;
	} else if (strcmp(type,"vgatweak")==0) {
		mode->flags = VIDEO_FLAGS_DRIVER_VGATWEAK;
		if (parse_vgatweak(&mode->driver.vgatweak,begin,argend))
			return -1;
	} else if (strcmp(type,"vgaline")==0) {
		mode->flags = VIDEO_FLAGS_DRIVER_VGALINE;
		if (parse_modecrtc(&mode->driver.vgaline.mode,&mode->driver.vgaline.crtc,begin,argend))
			return -1;
	} else if (strcmp(type,"vbeline")==0) {
		mode->flags = VIDEO_FLAGS_DRIVER_VBELINE;
		if (parse_modecrtc(&mode->driver.vbeline.mode,&mode->driver.vbeline.crtc,begin,argend))
			return -1;
	} else {
		sprintf(video_mode_parse_error,"Unknow mode type");
		return -1;
	}

	return 0;
}

/* ----------------------------------------------------------------------- */
/* VGA */

video_error vga_mode_import_and_insert(video_mode_container* vmc, video_mode* mode) {
	video_mode mode8 = *mode;
	if (video_mode_import(&mode8)==0) {
		video_mode_container_insert(vmc,&mode8);
	}
	return 0;
}

/* Add all vga mode */
void video_mode_container_insert_vga(video_mode_container* vmc) {
	video_mode v;

        memset(&v,0,sizeof(v));
	v.driver.vga.mode = 0x13;
	if (vga_mode_import(&v)==0) {
		sprintf(v.name,"bios_vga_%dx%dx%d",(unsigned)v.size_x,(unsigned)v.size_y,(unsigned)v.bits);
		video_mode_container_insert(vmc,&v);
	}

        memset(&v,0,sizeof(v));
	v.driver.vga.mode = 0x6A;
	if (vga_mode_import(&v)==0) {
		sprintf(v.name,"bios_vga_%dx%dx%d",(unsigned)v.size_x,(unsigned)v.size_y,(unsigned)v.bits);
		video_mode_container_insert(vmc,&v);
	}

	memset(&v,0,sizeof(v));
	v.driver.vga.mode = 0x3;
	if (vga_mode_import(&v)==0) {
		sprintf(v.name,"bios_vga_text");
		video_mode_container_insert(vmc,&v);
	}

#if 0
	memset(&v,0,sizeof(v));
	v.driver.vga.mode = 0x1;
	if (vga_mode_import(&v)==0) {
		sprintf(v.name,"bios_vga_text40",(unsigned)v.size_x,(unsigned)v.size_y);
		video_mode_container_insert(vmc,&v);
	}
#endif
}

static void video_mode_vga_print(char* s, const video_mode* vm) {
       sprintf(s,"%s vga(%x)",vm->name,vm->driver.vga.mode);
}

/* ----------------------------------------------------------------------- */
/* VBE */

video_error vbe_mode_import_and_insert(video_mode_container* vmc, video_mode* mode) {
	video_mode mode8 = *mode;
	if (video_mode_import(&mode8)==0) {
		video_mode_container_insert(vmc,&mode8);
	}
	return 0;
}

/* Add all vbe modes */
static void video_mode_container_insert_vbe_all(video_mode_container* vmc) {
	if (vbe_active() && vbe_state.info.VESAVersion >= 0x0102) {
		vbe_mode_iterator vmi;
		vbe_mode_iterator_begin(&vmi);
		while (!vbe_mode_iterator_end(&vmi)) {
			video_mode v;
                        memset(&v,0,sizeof(v));
			v.driver.vbe.mode = vbe_mode_iterator_get(&vmi);

			if (vbe_mode_import(&v)==0) {
				sprintf(v.name,"bios_vbebanked_%dx%dx%d",(unsigned)v.size_x,(unsigned)v.size_y,(unsigned)v.bits);
				video_mode_container_insert(vmc,&v);
     			}

			if (vbe_state.info.VESAVersion >= 0x200) {
				v.driver.vbe.mode |= vbeLinearBuffer;
				if (vbe_mode_import(&v)==0) {
					sprintf(v.name,"bios_vbelinear_%dx%dx%d",(unsigned)v.size_x,(unsigned)v.size_y,(unsigned)v.bits);
					video_mode_container_insert(vmc,&v);
	     			}
			}

			vbe_mode_iterator_next(&vmi);
		}
	}
}

/* Add vbe modes */
void video_mode_container_insert_vbe(video_mode_container* vmc) {
	video_mode_container all;
	video_mode_container_iterator i;

	if (video_option_default_mode_get() == VIDEO_OPTION_DEFAULT_MODE_ALL) {
		video_mode_container_insert_vbe_all(vmc);
		return;
	}

	video_mode_container_init(&all);
	video_mode_container_insert_vbe_all(&all);

	video_mode_container_iterator_begin(&i,&all);
	while (!video_mode_container_iterator_end(&i)) {
		video_mode* ip = video_mode_container_iterator_get(&i);

		video_mode_container_iterator j;
		video_mode_container_iterator_begin(&j,&all);
		while (!video_mode_container_iterator_end(&j)) {
			video_mode* jp = video_mode_container_iterator_get(&j);
			if (video_mode_size_x(ip) == video_mode_size_x(jp) && video_mode_size_y(ip) == video_mode_size_y(jp) && video_mode_bytes_per_pixel(ip) == video_mode_bytes_per_pixel(jp)) {
				if (!(video_mode_flags(ip) & VIDEO_FLAGS_LINEAR) && (video_mode_flags(jp) & VIDEO_FLAGS_LINEAR))
					break;
				if (video_mode_bits_per_pixel(ip) < video_mode_bits_per_pixel(jp))
					break;
			}

			video_mode_container_iterator_next(&j);
		}

		if (video_mode_container_iterator_end(&j)) {
			video_mode_container_insert(vmc,ip);
		}

		video_mode_container_iterator_next(&i);
	}

	video_mode_container_done(&all);
}

static void video_mode_vbe_print(char* s, const video_mode* vm) {
       sprintf(s,"%s vbe(%x)",vm->name,vm->driver.vbe.mode);
}

/* ----------------------------------------------------------------------- */
/* VBELINE */

/* the mode number is set later */
const char* VBELINE_DEFAULT[] = {

/* step 16 - 256 272 288 304 320 336 352 368 384 400 416 432 448 */

/* PC MONITOR AT 31.5 kHz */
"pc_31.5_vbe_256x192 vbeline(0 10.11 256 265 289 321 192 199 201 208 doublescan -hsync -vsync)",
"pc_31.5_vbe_288x216 vbeline(0 11.63 288 297 329 369 216 224 226 233 doublescan -hsync -vsync)",
"pc_31.5_vbe_304x228 vbeline(0 12.13 304 313 345 385 228 236 238 245 doublescan -hsync -vsync)",
"pc_31.5_vbe_320x240 vbeline(0 12.64 320 329 361 401 240 249 251 258 doublescan -hsync -vsync)",
"pc_31.5_vbe_336x252 vbeline(0 13.17 336 345 377 417 252 258 260 270 doublescan -hsync -vsync)",
"pc_31.5_vbe_344x258 vbeline(0 13.87 344 361 393 441 258 267 269 276 doublescan -hsync -vsync)",
"pc_31.5_vbe_352x264 vbeline(0 14.17 352 369 401 449 264 274 276 283 doublescan -hsync -vsync)",
"pc_31.5_vbe_512x384 vbeline(0 20.22 512 530 578 642 384 388 390 407 -hsync +vsync)",
"pc_31.5_vbe_640x480 vbeline(0 25.28 640 658 722 802 480 490 493 516 -hsync -vsync)",

/* PC MONITOR AT 35.16 kHz */
/* "pc_35.16_vbe_352x264 vbeline(0 15.81 352 369 401 449 264 274 276 284 doublescan -hsync -vsync)", */
"pc_35.16_vbe_368x276 vbeline(0 16.41 368 377 417 465 276 286 288 296 doublescan -hsync -vsync)",
"pc_35.16_vbe_384x288 vbeline(0 17.50 384 401 441 497 288 298 300 308 doublescan -hsync -vsync)",
"pc_35.16_vbe_400x300 vbeline(0 18.06 400 417 457 513 300 311 313 321 doublescan -hsync -vsync)",
"pc_35.16_vbe_432x324 vbeline(0	19.20 432 449 489 545 324 336 338 345 doublescan -hsync -vsync)",
"pc_35.16_vbe_800x600 vbeline(0 36.12 800 835 915 1027 600 614 617 642 -hsync -vsync)",

/* PC MONITOR AT 48.0 kHz */
"pc_48.0_vbe_512x384 vbeline(0 31.10 512 520 552 648 384 385 390 400 doublescan -hsync -vsync)",
"pc_48.0_vbe_1024x768 vbeline(0 61.44 1024 1040 1128 1280 768 770 780 800 -hsync -vsync)",

/* PAL TV */
"pal_vbe_288x268 vbeline(0 6.29 288 312 364 400 268 282 283 313 -hsync -vsync)",
"pal_vbe_336x268 vbeline(0 7.08 336 356 404 452 268 278 279 313 -hsync -vsync)",
"pal_vbe_384x288 vbeline(0 8.00 384 412 460 512 288 292 295 313 -hsync -vsync)",
"pal_vbe_768x288 vbeline(0 16.42 768 836 916 1048 288 289 294 313 -hsync -vsync)",
"pal_vbe_768x576 vbeline(0 16.42 768 836 916 1048 576 579 589 626 -hsync -vsync interlaced)",

/* NTSC TV */
"ntsc_vbe_288x224 vbeline(0 6.29 288 312 360 400 224 234 237 263 -hsync -vsync)",
"ntsc_vbe_336x224 vbeline(0 7.08 336 356 404 448 224 234 237 263 -hsync -vsync)",
"ntsc_vbe_320x240 vbeline(0 6.82 320 348 388 432 240 241 243 263 -hsync -vsync)",
"ntsc_vbe_640x240 vbeline(0 13.63 640 688 768 864 240 241 243 263 -hsync -vsync)",
"ntsc_vbe_640x480 vbeline(0 13.66 640 684 764 864 480 484 488 526 -hsync -vsync interlaced)",

/* 24.6 kHz */
"arcade_24.6_512x384 vbeline(0 17.71 512 608 648 720 384 383 387 410 -hsync -vsync)",

0
};

static int vbeline_search_target_mode(unsigned bits, unsigned req_x, unsigned req_y) {
	video_mode best;
	int best_set = 0;

	vbe_mode_iterator vmi;
	vbe_mode_iterator_begin(&vmi);
	while (!vbe_mode_iterator_end(&vmi)) {
		video_mode v;
		int found = 0;
		memset(&v,0,sizeof(v));

		v.driver.vbe.mode = vbe_mode_iterator_get(&vmi) | vbeLinearBuffer;
		found =	vbe_mode_import(&v)==0;
		if (!found) {
			v.driver.vbe.mode = vbe_mode_iterator_get(&vmi);
			found =	vbe_mode_import(&v)==0;
		}

		if (found) {
			if (v.size_x >= req_x && v.size_y >= req_y && v.bits == bits) {
				if (!best_set ||
					(v.size_x <= best.size_x && v.size_y <= best.size_y)) {
					best_set = 1;
					best = v;
				}
			}
		}

		vbe_mode_iterator_next(&vmi);
	}

	if (best_set)
		return best.driver.vbe.mode;
	else
		return -1;
}

video_error vbeline_mode_import_and_insert(video_mode_container* vmc, video_mode* mode) {
	int mode_number;
	if (mode->driver.vbeline.mode != 0) {
		if (video_mode_import(mode)==0) {
			video_mode_container_insert(vmc,mode);
		}
	} else {
	 	video_mode mode8 = *mode;
		video_mode mode16 = *mode;

		strcat(mode8.name,"x8");
		mode_number = vbeline_search_target_mode(8,mode8.driver.vbeline.crtc.hde,mode8.driver.vbeline.crtc.vde);
		if (mode_number >= 0) {
			mode8.driver.vbeline.mode = mode_number;
			if (video_mode_import(&mode8)==0) {
				video_mode_container_insert(vmc,&mode8);
			}
		}

		strcat(mode16.name,"x16");
		mode_number = vbeline_search_target_mode(16,mode16.driver.vbeline.crtc.hde,mode16.driver.vbeline.crtc.vde);
		if (mode_number >= 0) {
			mode16.driver.vbeline.mode = mode_number;
			if (video_mode_import(&mode16)==0) {
				video_mode_container_insert(vmc,&mode16);
			}
		}
	}

	return 0;
}

void video_mode_container_insert_vbeline(video_mode_container* vmc) {
	if (card_active()) {
		const char** p = VBELINE_DEFAULT;
		while (*p) {
			video_mode mode;
			char* buffer = strdup(*p);
			memset(&mode,0,sizeof(mode));
			if (video_mode_parse_line(&mode,buffer)==0) {
				vbeline_mode_import_and_insert(vmc,&mode);
			}
			free(buffer);
			++p;
		}
	}
}

static void video_mode_vbeline_print(char* s, const video_mode* mode) {
	s += sprintf(s,"%s vbeline(",mode->name);
	s += print_modecrtc(s,mode->driver.vbeline.mode,&mode->driver.vbeline.crtc);
	s += sprintf(s,")");
}

/* ----------------------------------------------------------------------- */
/* TWEAK */

/* Default register for vga mode 13h 320x200 256 color */
static tweak_register tweak_vga320x200[] = {
	{ 0x3c2, 0x0, 0x63 },
	{ 0x3d4, 0x0, 0x5f },
	{ 0x3d4, 0x1, 0x4f },
	{ 0x3d4, 0x2, 0x50 },
	{ 0x3d4, 0x3, 0x82 },
	{ 0x3d4, 0x4, 0x54 },
	{ 0x3d4, 0x5, 0x80 },
	{ 0x3d4, 0x6, 0xbf },
	{ 0x3d4, 0x7, 0x1f },
	{ 0x3d4, 0x8, 0x0 },
	{ 0x3d4, 0x9, 0x41 },
	{ 0x3d4, 0x10, 0x9c },
	{ 0x3d4, 0x11, 0x8e },
	{ 0x3d4, 0x12, 0x8f },
	{ 0x3d4, 0x13, 0x28 },
	{ 0x3d4, 0x14, 0x40 },
	{ 0x3d4, 0x15, 0x96 },
	{ 0x3d4, 0x16, 0xb9 },
	{ 0x3d4, 0x17, 0xa3 },
	{ 0x3c4, 0x1, 0x1 },
//	{ 0x3c4, 0x3, 0x0 },
	{ 0x3c4, 0x4, 0xe },
	{ 0x3ce, 0x5, 0x40 },
	{ 0x3ce, 0x6, 0x5 },
	{ 0x3c0, 0x10, 0x41 },
//	{ 0x3c0, 0x11, 0x0 },
//	{ 0x3c0, 0x12, 0xf },
	{ 0x3c0, 0x13, 0x0 },
//	{ 0x3c0, 0x14, 0x0 }
};

/* Default register for vga mode 6Ah 800x600 16 color */
static tweak_register tweak_vga800x600[] = {
	{ 0x3c2, 0x0, 0x2f },
	{ 0x3d4, 0x0, 0x7f },
	{ 0x3d4, 0x1, 0x63 },
	{ 0x3d4, 0x2, 0x64 },
	{ 0x3d4, 0x3, 0x82 },
	{ 0x3d4, 0x4, 0x6b },
	{ 0x3d4, 0x5, 0x1b },
	{ 0x3d4, 0x6, 0x72 },
	{ 0x3d4, 0x7, 0xf0 },
	{ 0x3d4, 0x8, 0x0 },
	{ 0x3d4, 0x9, 0x60 },
	{ 0x3d4, 0x10, 0x58 },
	{ 0x3d4, 0x11, 0x8c },
	{ 0x3d4, 0x12, 0x57 },
	{ 0x3d4, 0x13, 0x32 },
	{ 0x3d4, 0x14, 0x0 },
	{ 0x3d4, 0x15, 0x57 },
	{ 0x3d4, 0x16, 0x0 },
	{ 0x3d4, 0x17, 0xe3 },
	{ 0x3c4, 0x1, 0x1 },
//	{ 0x3c4, 0x3, 0x0 },
	{ 0x3c4, 0x4, 0x6 },
	{ 0x3ce, 0x5, 0x0 },
	{ 0x3ce, 0x6, 0x5 },
	{ 0x3c0, 0x10, 0x1 },
//	{ 0x3c0, 0x11, 0x0 },
//	{ 0x3c0, 0x12, 0xf },
	{ 0x3c0, 0x13, 0x0 },
//	{ 0x3c0, 0x14, 0x0 }
};

unsigned tweak_register_get(unsigned port, int subindex, void* arg) {
	vgatweak_video_mode* tweak = (vgatweak_video_mode*)arg;
	unsigned i;
	tweak_register* r = 0;
	unsigned r_num = 0;

	/* specific register */
	for(i=0;i<tweak->regs_num;++i)
		if (tweak->regs[i].port == port && (subindex<0 || tweak->regs[i].index == subindex))
			return tweak->regs[i].value;

	/* default register */
	if (tweak->mode == 0x13) {
		r = tweak_vga320x200;
		r_num =	sizeof(tweak_vga320x200)/sizeof(tweak_register);
	} else if (tweak->mode == 0x6A) {
		r = tweak_vga800x600;
		r_num =	sizeof(tweak_vga800x600)/sizeof(tweak_register);
	}

	if (r) {
	        for(i=0;i<r_num;++i)
		        if (r[i].port == port && (subindex<0 || r[i].index == subindex))
				return r[i].value;
	}

	return 0;
}

int tweak_register_set(unsigned value, unsigned port, int subindex, void* arg) {
	vgatweak_video_mode* tweak = (vgatweak_video_mode*)arg;
	unsigned i;

	/* register already present */
	for(i=0;i<tweak->regs_num;++i) {
		if (tweak->regs[i].port == port && (subindex<0 || tweak->regs[i].index == subindex)) {
			tweak->regs[i].value = value;
			return 0;
		}
	}

	/* add register */
	if (tweak->regs_num < VIDEO_MODE_REGS_MAX) {
		tweak->regs[tweak->regs_num].index = subindex;
		tweak->regs[tweak->regs_num].port = port;
		tweak->regs[tweak->regs_num].value = value;
		++tweak->regs_num;
		return 0;
	}

	return -1;
}

/* Default tweak mode */
#define TWEAK_MODE(n,r) { n, sizeof(r)/sizeof(tweak_register), r },

struct vgatweak_default_struct {
	const char* name;
	unsigned regs_num;
	tweak_register* regs;
} vgatweak_default[] = {
	TWEAK_MODE( "tweak320x200", tweak_vga320x200 ) // VGA mode 13h
	{ 0, 0, 0 }
};

video_error vgatweak_mode_import_and_insert(video_mode_container* vmc, video_mode* mode) {
	video_mode mode8 = *mode;
	if (video_mode_import(&mode8)==0) {
		video_mode_container_insert(vmc,&mode8);
	}
	return 0;
}

void video_mode_container_insert_vgatweak(video_mode_container* vmc) {
	struct vgatweak_default_struct* p = vgatweak_default;
	while (p->name) {
		unsigned i;
		video_mode v;
		memset(&v,0,sizeof(v));
		assert(	p->regs_num <= VIDEO_MODE_REGS_MAX );
		v.driver.vgatweak.mode = 0x13;
		v.driver.vgatweak.regs_num = p->regs_num;
		for(i=0;i<p->regs_num;++i)
			v.driver.vgatweak.regs[i] = p->regs[i];
		if (vgatweak_mode_import(&v)==0) {
			assert( strlen(p->name) < VIDEO_MODE_NAME_MAX );
			strcpy(v.name,p->name);
			video_mode_container_insert(vmc,&v);
		}
		++p;
	}
}

#if 1

static void video_mode_vgatweak_print(char* s, const video_mode* mode) {
       unsigned i;
       sprintf(s,"%s vgatweak(%x ", mode->name, mode->driver.vgatweak.mode);
       for(i=0;i<mode->driver.vgatweak.regs_num;++i) {
		sprintf(s+strlen(s),"%x[%x]=%x",
                        (unsigned)mode->driver.vgatweak.regs[i].port,
                        (unsigned)mode->driver.vgatweak.regs[i].index,
                        (unsigned)mode->driver.vgatweak.regs[i].value);
                if (i+1!=mode->driver.vgatweak.regs_num)
                        sprintf(s+strlen(s)," ");
       }
       sprintf(s+strlen(s),")");
}

#else

/* Export in vgaline format */
static void video_mode_vgatweak_print(char* s, const video_mode* mode) {
	tweak_mode_info info;
	const char* flag1;
	const char* flag2;
	const char* flag3;
	tweak_detect_mode(&info, tweak_register_get, (void*)&mode->driver.vgatweak);
	flag1 = info.char_size_y>1 ? " doublescan" : "";
	flag2 = info.hsync ? " -hsync" : " +hsync";
	flag3 = info.vsync ? " -vsync" : " +vsync";
	sprintf(s,"%s vgaline(%x %.2f %d %d %d %d %d %d %d %d%s%s%s)",
		mode->name, mode->driver.vgatweak.mode, (double)info.dot_clock/2E6,
		info.horz_display_end*info.horz_pixel_per_clock, info.horz_retrace_start*info.horz_pixel_per_clock, info.horz_retrace_end*info.horz_pixel_per_clock, info.horz_total*info.horz_pixel_per_clock,
		info.vert_display_end/info.char_size_y, info.vert_retrace_start/info.char_size_y, info.vert_retrace_end/info.char_size_y, info.vert_total/info.char_size_y,
		flag1,flag2,flag3
	);
}
#endif

/* ----------------------------------------------------------------------- */
/* VGALINE */

const char* VGALINE_DEFAULT[] = {

/* PC MONITOR AT 31.5 kHz */
"pc_31.5_vga_304x228x8 vgaline(13 12.59 304 324 356 400 228 236 238 245 doublescan -hsync -vsync)",
"pc_31.5_vga_320x240x8 vgaline(13 12.59 320 336 356 400 240 249 254 262 doublescan -hsync -vsync)",
"pc_31.5_vga_336x252x8 vgaline(13 14.16 336 368 416 448 252 266 270 288 doublescan -hsync -vsync)",
"pc_31.5_vga_384x288x8 vgaline(13 14.16 384 392 408 448 288 290 292 298 doublescan -hsync -vsync)",
"pc_31.5_vga_text vgaline(3 28.32 720 765 873 900 400 412 414 449 -hsync +vsync)",

/* ARCADE MONITOR */
"arcade_vga_text vgaline(3 14.16 640 712 792 880 200 225 227 267 -hsync -vsync)",

/* PAL TV */
"pal_vga_288x268x8 vgaline(13 6.29 288 312 364 400 268 282 283 316 -hsync -vsync)",
"pal_vga_336x268x8 vgaline(13 7.08 336 356 404 452 268 278 279 310 -hsync -vsync)",
"pal_vga_576x268x8 vgaline(13 12.59 576 624 728 800 268 282 283 316 -hsync -vsync)",
"pal_vga_672x268x8 vgaline(13 14.16 672 712 808 904 268 278 279 310 -hsync -vsync)",
"pal_vga_text vgaline(3 14.16 643 700 828 896 256 276 280 311 -hsync -vsync)",

/* NTSC TV */
"ntsc_vga_288x224x8 vgaline(13 6.29 288 312 360 400 224 234 237 266 -hsync -vsync)",
"ntsc_vga_336x224x8 vgaline(13 7.08 336 356 404 448 224 234 237 266 -hsync -vsync)",
"ntsc_vga_576x224x8 vgaline(13 12.59 576 624 720 800 224 234 237 266 -hsync -vsync)",
"ntsc_vga_672x224x8 vgaline(13 14.16 672 712 808 900 224 234 237 266 -hsync -vsync)",
"ntsc_vga_text vgaline(3 14.16 640 736 800 904 200 220 229 263 -hsync -vsync)",

0
};

video_error vgaline_mode_import_and_insert(video_mode_container* vmc, video_mode* mode) {
	video_mode mode8 = *mode;
	if (video_mode_import(&mode8)==0) {
		video_mode_container_insert(vmc,&mode8);
	}
	return 0;
}

void video_mode_container_insert_vgaline(video_mode_container* vmc) {
	const char** p = VGALINE_DEFAULT;
	while (*p) {
		video_mode mode;
		char* buffer = strdup(*p);
		memset(&mode,0,sizeof(mode));
		if (video_mode_parse_line(&mode,buffer)==0) {
			vgaline_mode_import_and_insert(vmc,&mode);
		}
		free(buffer);
		++p;
	}
}

static void video_mode_vgaline_print(char* s, const video_mode* mode) {
	s += sprintf(s,"%s vgaline(",mode->name);
	s += print_modecrtc(s,mode->driver.vgaline.mode,&mode->driver.vgaline.crtc);
	s += sprintf(s,")");
}

/* ----------------------------------------------------------------------- */

/* Max number of mode saved in config file */
#define VIDEO_CONFIG_MAX 256

/* Load the list of video mode */
video_error video_mode_container_load(video_mode_container* vmc) {
	unsigned i;
	int error = 0;
	for(i=0;i<VIDEO_CONFIG_MAX;++i) {
		char modeline[32];
                char* s;
                sprintf(modeline,"modeline%d",i);
		s = get_config_string("video",modeline,0);
		if (s) {
			video_mode mode;
			char* buffer = strdup(s);
			memset(&mode,0,sizeof(mode));
			if (video_mode_parse_line(&mode,buffer)) {
				if (!error) {
					error = 1;
					video_error_description_set("%s parsing modeline %s",video_mode_parse_error,s);
				}
			} else {
				if (video_mode_import_and_insert(vmc,&mode)!=0) {
					if (!error) {
						error = 1;
						video_error_description_set("Unsupported video mode %s",mode.name);
					}
				}
			}
			free(buffer);
		}
	}
	return error ? -1 : 0;
}

/* Export the modeline */
void video_mode_print(char* buffer, const video_mode* vm) {
	switch (vm->flags & VIDEO_FLAGS_DRIVER_MASK) {
		case VIDEO_FLAGS_DRIVER_VGA:
			video_mode_vga_print(buffer,vm);
		break;
		case VIDEO_FLAGS_DRIVER_VGALINE:
			video_mode_vgaline_print(buffer,vm);
		break;
		case VIDEO_FLAGS_DRIVER_VGATWEAK:
			video_mode_vgatweak_print(buffer,vm);
		break;
		case VIDEO_FLAGS_DRIVER_VBE :
			video_mode_vbe_print(buffer,vm);
		break;
		case VIDEO_FLAGS_DRIVER_VBELINE :
			video_mode_vbeline_print(buffer,vm);
		break;
	}
}

/* Save the list of video mode */
void video_mode_container_save(video_mode_container* vmc) {
	video_mode_container_iterator j;
	int i;
	video_mode_container_iterator_begin(&j,vmc);
	for(i=0;i<VIDEO_CONFIG_MAX;++i) {
		char modeline[32];
		sprintf(modeline,"modeline%d",i);
		if (!video_mode_container_iterator_end(&j)) {
			char buffer[1024];
			video_mode* vm = video_mode_container_iterator_get(&j);
			switch (vm->flags & VIDEO_FLAGS_DRIVER_MASK) {
				case VIDEO_FLAGS_DRIVER_VGA:
					video_mode_vga_print(buffer,vm);
				break;
				case VIDEO_FLAGS_DRIVER_VGALINE:
					video_mode_vgaline_print(buffer,vm);
				break;
				case VIDEO_FLAGS_DRIVER_VGATWEAK:
					video_mode_vgatweak_print(buffer,vm);
				break;
				case VIDEO_FLAGS_DRIVER_VBE :
					video_mode_vbe_print(buffer,vm);
				break;
				case VIDEO_FLAGS_DRIVER_VBELINE :
					video_mode_vbeline_print(buffer,vm);
				break;
			}
			set_config_string("video",modeline,buffer);
			video_mode_container_iterator_next(&j);
		} else {
			set_config_string("video",modeline,0);
		}
	}
}


void video_mode_container_iterator_begin(video_mode_container_iterator* vmci, video_mode_container* vmc) {
	vmci->base = vmc->base;
}

void video_mode_container_iterator_next(video_mode_container_iterator* vmci) {
	vmci->base = vmci->base->next;
}

int video_mode_container_iterator_end(video_mode_container_iterator* vmci) {
	return vmci->base == 0;
}

video_mode* video_mode_container_iterator_get(video_mode_container_iterator* vmci) {
	return vmci->base;
}

/***************************************************************************/

static video_error video_monitor_parse(video_monitor_range* range, const char* begin, float mult) {
	const double monitor_approsimation_perc = 0.03;
	unsigned i = 0;
	const char* end = begin + strlen(begin);
	parse_separator(" \t",&begin,end);
	while (begin < end) {
		double v0;
		if (i==VIDEO_MONITOR_RANGE_MAX) {
			return -1;
		}
		if (parse_double(&v0,&begin,end))
			return -1;
		parse_separator(" \t",&begin,end);
		if (*begin=='-') {
			double v1;
			++begin;
			if (parse_double(&v1,&begin,end))
				return -1;
			parse_separator(" \t",&begin,end);
			if (v0 < 0 || v0 > v1 || v1 > 300) {
				return -1;
			}
			range[i].low = mult * v0 * (1-monitor_approsimation_perc);
			range[i].hight = mult * v1 * (1+monitor_approsimation_perc);
			++i;
		} else if (*begin==',' || !*begin) {
			++begin;
			if (v0 < 0 || v0 > 300) {
				return -1;
			}
			range[i].low = mult * v0 * (1-monitor_approsimation_perc);
			range[i].hight = mult * v0 * (1+monitor_approsimation_perc);
			++i;
		} else {
			return -1;
		}
		parse_separator(" \t",&begin,end);
	}
	while (i<VIDEO_MONITOR_RANGE_MAX) {
		range[i].low = 0;
		range[i].hight = 0;
		++i;
	}
	return 0;
}

/* Load the video options configuration
 * return:
 *   ==0 success
 */
video_error video_options_load(void) {
	const char* s;

	/* Options must be loaded before initialization */
	assert( !video_active() );

 	s = get_config_string("video","vbeline_tweak","aggressive");
	if (s) {
		if (stricmp(s,"aggressive")==0) {
			video_option_vbeline_tweak_set(VIDEO_OPTION_VBELINE_TWEAK_AGGRESSIVE);
		} else if (stricmp(s,"moderate")==0) {
			video_option_vbeline_tweak_set(VIDEO_OPTION_VBELINE_TWEAK_MODERATE);
		} else {
			video_error_description_set("Invalid option 'vbeline_tweak = %s'",s);
			return -1;
		}
	}

 	s = get_config_string("video","vbeline_driver","auto");
	if (s) {
		if (stricmp(s,"no")==0 || stricmp(s,"none")==0) {
			video_option_vbeline_driver_set(VIDEO_OPTION_VBELINE_DRIVER_NONE);
		} else if (stricmp(s,"auto")==0) {
			video_option_vbeline_driver_set(VIDEO_OPTION_VBELINE_DRIVER_AUTO);
		} else if (stricmp(s,"vbe3")==0) {
			video_option_vbeline_driver_set(VIDEO_OPTION_VBELINE_DRIVER_VBE3);
		} else if (stricmp(s,"laguna")==0) {
			video_option_vbeline_driver_set(VIDEO_OPTION_VBELINE_DRIVER_LAGUNA);
		} else if (stricmp(s,"sis6326")==0) {
			video_option_vbeline_driver_set(VIDEO_OPTION_VBELINE_DRIVER_SIS6326);
		} else if (stricmp(s,"savage")==0) {
			video_option_vbeline_driver_set(VIDEO_OPTION_VBELINE_DRIVER_SAVAGE);
		} else if (stricmp(s,"3dfx")==0) {
			video_option_vbeline_driver_set(VIDEO_OPTION_VBELINE_DRIVER_3DFX);
		} else if (stricmp(s,"ati")==0) {
			video_option_vbeline_driver_set(VIDEO_OPTION_VBELINE_DRIVER_ATI);
		} else if (stricmp(s,"s3")==0) {
			video_option_vbeline_driver_set(VIDEO_OPTION_VBELINE_DRIVER_S3);
		} else if (stricmp(s,"cirrus")==0) {
			video_option_vbeline_driver_set(VIDEO_OPTION_VBELINE_DRIVER_CIRRUS);
		} else if (stricmp(s,"matrox")==0) {
			video_option_vbeline_driver_set(VIDEO_OPTION_VBELINE_DRIVER_MATROX);
		} else if (stricmp(s,"r128")==0) {
			video_option_vbeline_driver_set(VIDEO_OPTION_VBELINE_DRIVER_R128);
		} else {
			video_error_description_set("Invalid option 'vbeline_driver = %s'",s);
			return -1;
		}
	}

 	s = get_config_string("video","default_mode","smart");
	if (s) {
		if (stricmp(s,"smart")==0) {
			video_option_default_mode_set(VIDEO_OPTION_DEFAULT_MODE_SMART);
		} else if (stricmp(s,"all")==0) {
			video_option_default_mode_set(VIDEO_OPTION_DEFAULT_MODE_ALL);
		} else {
			video_error_description_set("Invalid option 'mode_default = %s'",s);
			return -1;
		}
	}

	s = get_config_string("video","crtc_blank","aggressive");
	if (s) {
		if (stricmp(s,"aggressive")==0) {
			video_option_crtc_blank_set(VIDEO_OPTION_CRTC_BLANK_AGGRESSIVE);
		} else if (stricmp(s,"moderate")==0) {
			video_option_crtc_blank_set(VIDEO_OPTION_CRTC_BLANK_MODERATE);
		} else {
			video_error_description_set("Invalid option 'crtc_blank = %s'",s);
			return -1;
		}
	}

	s = get_config_string("video","vbeline_clock_multiplier","1.0");
	if (s) {
		double v = strtod(s,0);
		if (v<=0) {
			video_error_description_set("Invalid option 'vbeline_clock_multiplier = %s'",s);
			return -1;
		}
		video_option_vbeline_clock_multiplier_set(v);
	}

	return 0;
}

/* Load the monitor configuration
 * return:
 *   ==0 success
 *   <0 error
 *   >0 configuration missing
 */
video_error video_monitor_load(video_monitor* monitor) {
	const char* s;
 	s = get_config_string("video","hclock",0);
	if (!s) {
		video_error_description_set("Missing monitor 'hclock' specification");
		return 1;
	}
	if (video_monitor_parse(monitor->hclock,s,1000)) {
		video_error_description_set("Invalid monitor 'hclock' specification");
		return -1;
	}
 	s = get_config_string("video","vclock",0);
	if (!s) {
		video_error_description_set("Missing monitor 'vclock' specification");
		return 1;
	}
	if (video_monitor_parse(monitor->vclock,s,1)) {
		video_error_description_set("Invalid monitor 'vclock' specification");
		return -1;
	}
	return 0;
}

/* Check if a video mode is acceptable
 * return:
 *   !=0 acceptable
 *   ==0 no
 */
int video_monitor_clock_check(const video_monitor* monitor, float hclock, float vclock) {
	unsigned i;
	for(i=0;i<VIDEO_MONITOR_RANGE_MAX;++i)
		if (monitor->hclock[i].low <= hclock &&	hclock <= monitor->hclock[i].hight)
			break;
	if (i==VIDEO_MONITOR_RANGE_MAX) {
		video_error_description_set("Horizontal clock of %.2f kHz is out of range of your monitor", (double)hclock / 1E3);
		return 0;
	}
	for(i=0;i<VIDEO_MONITOR_RANGE_MAX;++i)
		if (monitor->vclock[i].low <= vclock &&	vclock <= monitor->vclock[i].hight)
			break;
	if (i==VIDEO_MONITOR_RANGE_MAX) {
		video_error_description_set("Vertical clock of %.2f Hz is out of range of your monitor", (double)vclock);
		return 0;
	}
	return 1;
}

/* Check if a video mode is acceptable
 * return:
 *   !=0 acceptable
 *   ==0 no
 */
int video_monitor_mode_check(const video_monitor* monitor, const video_mode* mode) {
	switch (mode->flags & VIDEO_FLAGS_DRIVER_MASK) {
		case VIDEO_FLAGS_DRIVER_VGA:
		case VIDEO_FLAGS_DRIVER_VBE :
			return 1;
		case VIDEO_FLAGS_DRIVER_VGALINE:
			return video_monitor_clock_check(monitor,
				mode->horz_freq,
				mode->vert_freq);
		case VIDEO_FLAGS_DRIVER_VGATWEAK: {
			struct tweak_crtc info;
			tweak_crtc_get(&info, tweak_register_get, (void*)&mode->driver.vgatweak);
			return video_monitor_clock_check(monitor,info.horz_clock,info.vert_clock);
		}
		case VIDEO_FLAGS_DRIVER_VBELINE:
			return video_monitor_clock_check(monitor,
				mode->horz_freq,
				mode->vert_freq);
	}
	return 0;
}

