/*
 * This file is part of the AdvanceMAME project
 *
 * Copyright (C) 1999-2000 Andrea Mazzoleni
 *
 * The contents of this file are free for non-commercial use.
 * Any commercial use is forbidden.
 */

#include "scrcard.h"
#include "vbe3.h"

static int vbe3_mode_set(unsigned mode, const video_crtc* crtc) {
	vbe_ModeInfoBlock info;
	vbe_CRTCInfoBlock vd;

	if (vbe_state.info.VESAVersion < 0x0300) {
		video_error_description_set("VBE CRTC programming isn't avaliable on VBE version prior 3.0");
		return -1;
	}

	if (vbe_mode_info_get(&info, mode)!=0) {
		video_error_description_set("VBE report that mode %x isn't supported", mode);
		return -1;
	}

	if ((info.ModeAttributes & vbeMdAvailable) == 0) {
		video_error_description_set("VBE report that mode %x isn't avaliable", mode);
		return -1;
	}

	if (crtc->interlaced && (info.ModeAttributes & vbeMdInterlace)==0) {
		video_error_description_set("VBE report that mode %x doesn't support interlace", mode);
		return -1;
	}

	if (crtc->doublescan && (info.ModeAttributes & vbeMdDoubleScan)==0) {
		video_error_description_set("VBE report that mode %x doesn't support doublescan", mode);
		return -1;
	}

	if (crtc->interlaced &&	crtc->doublescan) {
		video_error_description_set("VBE doesn't support interlaced and doublescan at the same time");
		return -1;
	}

	memset(&vd,0,sizeof(vd));

	/* if the resulting mode is smaller and if later sync values are set */
	/* use fake value to prevent a possible mode set failure if the */
	/* specified values are smaller than the fixed resolution */

	if (crtc->hde < info.XResolution && video_option_vbeline_tweak_get() == VIDEO_OPTION_VBELINE_TWEAK_AGGRESSIVE) {
		/* use fake values, correct values are set later */
		vd.HorizontalSyncStart = info.XResolution * 41 / 40;
		vd.HorizontalSyncEnd = info.XResolution * 10 / 9;
		vd.HorizontalTotal = info.XResolution * 5 / 4;
	} else {
		vd.HorizontalTotal = crtc->ht;
		vd.HorizontalSyncStart = crtc->hrs;
		vd.HorizontalSyncEnd = crtc->hre;
	}
	if (crtc->vde < info.YResolution && video_option_vbeline_tweak_get() == VIDEO_OPTION_VBELINE_TWEAK_AGGRESSIVE) {
		/* use fake values, correct values are set later */
		vd.VerticalSyncStart = info.YResolution * 31 / 30;
		vd.VerticalSyncEnd = info.YResolution * 21 / 20;
		vd.VerticalTotal = info.YResolution * 16 / 15;
	} else {
		vd.VerticalTotal = crtc->vt;
		vd.VerticalSyncStart = crtc->vrs;
		vd.VerticalSyncEnd = crtc->vre;
	}

	/* VBE3 requires doubled values for doublescan mode */
	/* but NOT half values for interlaced modes */
	if (crtc->doublescan) {
		vd.VerticalTotal *= 2;
		vd.VerticalSyncStart *= 2;
		vd.VerticalSyncEnd *= 2;
	}

	vd.Flags = 0;
	if (crtc->doublescan)
		vd.Flags |= vbeCRTCDoubleScan;
	if (crtc->interlaced)
		vd.Flags |= vbeCRTCInterlaced;
	if (crtc->hsync)
		vd.Flags |= vbeCRTCHorizontalSyncNegative;
	if (crtc->hsync)
		vd.Flags |= vbeCRTCVerticalSyncNegative;

	vd.PixelClock = crtc->pixelclock;
	if (vbe_pixelclock_get(&vd.PixelClock,mode)!=0) {
		video_error_description_set("Error in VBE pixelclock get");
		return -1;
	}
	vd.RefreshRate = vd.PixelClock * 100 / (crtc->ht * crtc->vt);
	if (crtc->doublescan)
		vd.RefreshRate /= 2;

	video_log("vbe3: vesa %dx%d (%04x %.2f %d %d %d %d %d %d)\n",(unsigned)info.XResolution,(unsigned)info.YResolution,(unsigned)mode,(double)vd.PixelClock/1E6,(unsigned)vd.HorizontalSyncStart,(unsigned)vd.HorizontalSyncEnd,(unsigned)vd.HorizontalTotal,(unsigned)vd.VerticalSyncStart,(unsigned)vd.VerticalSyncEnd,(unsigned)vd.VerticalTotal);

	if (vbe_mode_set(mode,&vd)!=0) {
		video_error_description_set("Error in VBE mode set");
		return -1;
	}

	/* set only if different */
	if (crtc->hde != vbe_state.mode_info.XResolution || video_option_vbeline_tweak_get() == VIDEO_OPTION_VBELINE_TWEAK_AGGRESSIVE) {
		unsigned x_div;

		if ((vbe_state.info.Capabilities & vbeNonVGA) != 0) {
			video_error_description_set("VBE reports that your hardware isn't VGA compatible at register level");
			return -1;
		} else {
			unsigned de,bs,rs,re,be,tt;
			/* get the active hde */
			unsigned hde = tweak_hde_get(tweak_reg_read, 0);
			if (hde == vbe_state.mode_info.XResolution) {
				x_div = 1;
			} else if (2U*hde == vbe_state.mode_info.XResolution) {
				x_div = 2;
			} else if (4U*hde == vbe_state.mode_info.XResolution) {
				x_div = 4;
			} else if (8U*hde == vbe_state.mode_info.XResolution) {
				x_div = 8;
			} else if (16U*hde == vbe_state.mode_info.XResolution) {
				x_div = 16;
			} else {
				video_error_description_set("Unknow VGA HDE register value of %d for resolution %d",(int)hde,(int)vbe_state.mode_info.XResolution);
				return -1;
			}

			video_log("vbe3: mode horiz div %d\n",x_div);

			if (vbe_scanline_pixel_request((crtc->hde + x_div - 1) & ~(x_div - 1)) != 0) {
				video_error_description_set("Error in VBE scanline request for %d pixel",(crtc->hde + x_div - 1) & ~(x_div - 1));
				return -1;
			}

			tweak_ready_vga_regs();

			video_crtc_realize_h(crtc,&de,&bs,&rs,&re,&be,&tt,x_div);

			if (video_option_vbeline_tweak_get() == VIDEO_OPTION_VBELINE_TWEAK_AGGRESSIVE) {
				tweak_htt_set( tt, tweak_reg_write, tweak_reg_read, 0);
				tweak_hrs_set( rs, tweak_reg_write, tweak_reg_read, 0);
				tweak_hre_set( re, tweak_reg_write, tweak_reg_read, 0);
			}
			tweak_hde_set( de, tweak_reg_write, tweak_reg_read, 0);
			tweak_hbs_set( bs, tweak_reg_write, tweak_reg_read, 0);
			tweak_hbe_set( be, tweak_reg_write, tweak_reg_read, 0);

			video_log("vbe3: mode horiz set (%d %d %d %d %d %d)\n",de,bs,rs,re,be,tt);
		}

		vbe_state.mode_info.XResolution = crtc->hde;
	}

	/* set only if different */
	if (crtc->vde != vbe_state.mode_info.YResolution || video_option_vbeline_tweak_get() == VIDEO_OPTION_VBELINE_TWEAK_AGGRESSIVE) {
		unsigned y_mul;

		if (vbe_state.virtual_y < crtc->vde) {
			video_error_description_set("Too low virtual y size");
			return -1;
		}

		if ((vbe_state.info.Capabilities & vbeNonVGA) != 0) {
			video_error_description_set("VBE reports that your hardware isn't VGA compatible at register level");
			return -1;
		} else {
			unsigned de,bs,rs,re,be,tt;
			/* get the active vde */
			unsigned vde = tweak_vde_get(tweak_reg_read, 0);
			if (vde == vbe_state.mode_info.YResolution) {
				y_mul = 1;
			} else if (vde == 2U*vbe_state.mode_info.YResolution) {
				y_mul = 2;
			} else if (vde == 4U*vbe_state.mode_info.YResolution) {
				y_mul = 4;
			} else {
				video_error_description_set("Unknow VGA VDE register value of %d for resolution %d",vde,vbe_state.mode_info.YResolution);
				return -1;
			}

			video_log("vbe3: mode vert mul %d\n",y_mul);

			tweak_ready_vga_regs();

			video_crtc_realize_v(crtc,&de,&bs,&rs,&re,&be,&tt,y_mul);

			if (video_option_vbeline_tweak_get() == VIDEO_OPTION_VBELINE_TWEAK_AGGRESSIVE) {
				tweak_vtt_set(tt, tweak_reg_write, tweak_reg_read, 0);
				tweak_vrs_set(rs, tweak_reg_write, tweak_reg_read, 0);
				tweak_vre_set(re, tweak_reg_write, tweak_reg_read, 0);
			}
			tweak_vde_set( de, tweak_reg_write, tweak_reg_read, 0);
			tweak_vbs_set( bs, tweak_reg_write, tweak_reg_read, 0);
			tweak_vbe_set( be, tweak_reg_write, tweak_reg_read, 0);

			video_log("vbe3: mode vert set (%d %d %d %d %d %d)\n",de,bs,rs,re,be,tt);
		}

		vbe_state.mode_info.YResolution = crtc->vde;
	}

	if (video_option_vbeline_tweak_get() == VIDEO_OPTION_VBELINE_TWEAK_AGGRESSIVE) {
		tweak_hsync_set(crtc->hsync, tweak_reg_write, tweak_reg_read, 0);
		tweak_vsync_set(crtc->vsync, tweak_reg_write, tweak_reg_read, 0);
	}

	return 0;
}

/***************************************************************************/
/* VBE3 interface */

int detectVBE3(void) {
	if (vbe_state.info.VESAVersion < 0x0300) {
		return 0;
	}
	return 1;
}

int setVBE3(void) {
	if (vbe3_mode_set(card_state.mode_request,&card_state.crtc_request)!=0)
		return 0;
	return 1;
}

void resetVBE3(void) {
}

